/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu
  

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu

  Greg Morrow, Department of Physics,
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu

***************************************************************************/





#include "PLA.h"
#include <math.h>
#include <stdio.h>
#include "PLA_Blas1.h"

/*****************************************************************************/
int PLA_Nrm2_enter(PLA_Obj x, PLA_Obj result)
{
   int check_parameter, instrument;
   char current_routine[20] = "PLA_Nrm2";
   int objtype;
   MPI_Datatype datatype;
   PLA_Template template;
   int  error = 0, original_error = 0;

   PLA_Stack_push_current( current_routine );

   PLA_Get_check_parameters ( &check_parameter );

   if ( PLA_CHECK_PARAMETER == check_parameter )
   {
   /* check objects for validity */
      PLA_Obj_object_valid( x, "x", error, original_error);
      PLA_Obj_object_valid( result, "result", error, original_error);

    /* check datatypes */
      PLA_Obj_datatype_valid( x, "x", error, original_error );
      PLA_Obj_datatype( x, &datatype );
      PLA_Obj_datatype_equal( result, datatype, "result", error, original_error );

    /* Global call result must be a Multi-scalar */
      PLA_Obj_objtype_equal( result, PLA_MSCALAR, "result", error, original_error );

    /* check if templates are all equal (temporary restriction) */
    /* Copy can work with different templates 03/12/97 so commenting out JAG */
/*      PLA_Obj_template( alpha, &template );
      PLA_Obj_template_equal( x, template, "x", error, original_error );
      PLA_Obj_template_equal( y, template, "y", error, original_error ); */

    /* check dimensions */
      PLA_Obj_global_length_equal  ( result, 1, "result", error, original_error );
      PLA_Obj_global_width_equal   ( result, 1, "result", error, original_error );

   }
   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   return(original_error);
}

int PLA_Nrm2_exit( PLA_Obj result, PLA_Obj x )
{
   int instrument;
   char current_routine[20];

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   PLA_Stack_pop_current( current_routine );
   /* Free all of the "in" objects -- not implemented YET!! */
   return(PLA_SUCCESS);
}

/******************************************************************************/

int PLA_Nrm2( PLA_Obj X, PLA_Obj Result )
{
   PLA_Obj     Result_dup = NULL, Extra_dup = NULL;
   void *Result_dup_buffer = NULL, *Extra_dup_buffer = NULL;
   MPI_Datatype datatype;
   PLA_Template template_Result;
   MPI_Comm     Communicator;

   PLA_Nrm2_enter(X, Result);

/* Take care of the multiscalar duplication */

   PLA_Obj_datatype (Result, &datatype);
   PLA_Obj_template (Result, &template_Result);
   PLA_Temp_comm_all( template_Result, &Communicator);

   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                          template_Result,   &Result_dup);

   PLA_Copy(Result, Result_dup);

   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                          template_Result, &Extra_dup);

   PLA_Local_dot(X, X, Result_dup);

   PLA_Obj_local_buffer(Result_dup, &Result_dup_buffer);
   PLA_Obj_local_buffer(Extra_dup , &Extra_dup_buffer );

/* Get local then global answer -- distribute it to all in the communicator */

   MPI_Allreduce(Result_dup_buffer, Extra_dup_buffer, 1, datatype,
                 PLASUM(datatype), Communicator);

/* This needs to be fixed but I don't know if Baker has such a routine or ??? */
   PLA_Take_sqrt( Extra_dup ); 

   PLA_Copy(Extra_dup, Result);

   PLA_Obj_free( &Result_dup );
   PLA_Obj_free( &Extra_dup  );

   PLA_Nrm2_exit(X, Result);

   return PLA_SUCCESS;
}

