/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu
  

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/




#include "PLA.h"
#include <math.h>
#include <stdio.h>
#include "PLA_Blas1.h"

/******************************************************************/
/* PLA_Scal() takes a multi-scalar and an object and scales the    */
/* object by the multi-scalar                                     */
/******************************************************************/
int PLA_Scal(PLA_Obj Alpha,
            PLA_Obj X)
{
   int                Llength, Lwidth, Lstride_X, lld_X, i;
   PLA_Obj           Alpha_dup = NULL;
   PLA_Obj_obj_type  Object_type_Alpha;
   MPI_Datatype       Datatype_X, Datatype_Alpha;
   void               *X_Buffer, *Alpha_Buffer;
   int                PLA_Error = 0, idummy;
   int                owner_row, owner_col;

   PLA_Error += PLA_Obj_objtype(Alpha, &Object_type_Alpha);
   if(Object_type_Alpha != PLA_MSCALAR)
      return PLA_BAD_OBJ_TYPE;
   PLA_Obj_owner_row(Alpha, &owner_row);
   PLA_Obj_owner_col(Alpha, &owner_col);
   if ( owner_row != PLA_ALL_ROWS || owner_col != PLA_ALL_COLS ) 
   {
         PLA_Mscalar_create_conf_to( Alpha, PLA_ALL_ROWS, PLA_ALL_COLS,
                                         &Alpha_dup);
         PLA_Copy( Alpha, Alpha_dup );
   }
   PLA_Error += PLA_Obj_local_all(X, &Datatype_X, &Llength, &Lwidth,
                                  &X_Buffer, &lld_X);
   PLA_Error += PLA_Obj_local_stride(X, &Lstride_X);
   if(NULL == Alpha_dup)
      PLA_Error += PLA_Obj_local_all(Alpha, &Datatype_Alpha, &idummy, &idummy,
                                     &Alpha_Buffer, &idummy);
   else /* Needed to create duplicated Alpha */
      PLA_Error +=PLA_Obj_local_all(Alpha_dup,&Datatype_Alpha, &idummy, &idummy,
                                     &Alpha_Buffer, &idummy);
   if(Datatype_X != Datatype_Alpha)
		return PLA_BAD_DATA_TYPE;

	if(Llength)
		for(i = 0; i < Lwidth ; i++)
		{
			if(Datatype_X == MPI_DOUBLE)
			   PLA_dscal( &Llength, (double*)Alpha_Buffer,
				  ((double*)X_Buffer+(lld_X*i)), &Lstride_X);
			else if(Datatype_X == MPI_DOUBLE_COMPLEX)
			   PLA_zscal( &Llength, (double*)Alpha_Buffer,
			  ((PLA_DOUBLE_COMPLEX*)X_Buffer+(lld_X*i)),&Lstride_X);
			else if(Datatype_X == MPI_FLOAT)
				PLA_sscal( &Llength, (float*)Alpha_Buffer,
				  ((float*) X_Buffer+(lld_X*i)), &Lstride_X);
			else if(Datatype_X == MPI_COMPLEX)
				PLA_cscal( &Llength, (float*)Alpha_Buffer,
				 ((PLA_COMPLEX*)X_Buffer+(lld_X*i)),&Lstride_X);
	}
        PLA_Obj_free(&Alpha_dup);
	return 0;
}
