#include "PLA.h"

extern  MPI_Op PLA_SUM;

int  PLA_Trmv_upper_trans( int diag, PLA_Obj a,
                           PLA_Obj x_source_col,
                           PLA_Obj one,
                           PLA_Obj y_target_row, PLA_Obj y_target_col )
{
   PLA_Obj a_11 = NULL, a_12 = NULL, a_22 = NULL;
   PLA_Obj y_top_target  = NULL,  y_bottom_target = NULL;
   PLA_Obj y_left_target = NULL,  y_right_target  = NULL;
   PLA_Obj x_top_source  = NULL,  x_bottom_source = NULL;

   int     right_size, right_owner, bottom_size, bottom_owner, square_size;
   int     my_proc_row, my_proc_col;
   PLA_Template template;

   PLA_Obj_template(a, &template);
   PLA_Temp_comm_col_rank( template, &my_proc_row );
   PLA_Temp_comm_row_rank( template, &my_proc_col );

   PLA_Obj_view_all(a, &a_11);
   PLA_Obj_view_all(x_source_col, &x_top_source);
   PLA_Obj_view_all(y_target_col, &y_top_target);
   PLA_Obj_view_all(y_target_row, &y_left_target);

   while(TRUE)
   {
      PLA_Obj_split_size( a_11, PLA_SIDE_RIGHT, &right_size, &right_owner);
      PLA_Obj_split_size( a_11, PLA_SIDE_BOTTOM,  &bottom_size,  &bottom_owner );
      if( 0 == (square_size = min(right_size, bottom_size))) break;
      PLA_Obj_split_4(a_11, -square_size,-square_size,  &a_11,      &a_12,
                                                        PLA_DUMMY,  &a_22     );
      PLA_Obj_horz_split_2(y_top_target, -square_size,  &y_top_target,
                                                        &y_bottom_target );
      PLA_Obj_horz_split_2(x_top_source, -square_size,  &x_top_source,
                                                        &x_bottom_source );
      PLA_Obj_vert_split_2(y_left_target,-square_size, &y_left_target, &y_right_target);
      if (my_proc_col == right_owner)
      {
         if( my_proc_row == bottom_owner)
         {
            PLA_Local_copy( x_bottom_source, y_bottom_target );
            PLA_Local_trmv( PLA_UPPER_TRIANGULAR, PLA_TRANSPOSE, diag,
                            a_22, y_bottom_target);
         }
         PLA_Obj_implicit_local_transpose(y_right_target);
         PLA_Local_gemv( PLA_TRANS, one, a_12, x_top_source, one, y_right_target);
         PLA_Obj_implicit_local_transpose(y_right_target); /* not really needed*/
      }
   }
   PLA_Obj_free( &a_11 );   PLA_Obj_free( &x_top_source);
   PLA_Obj_free( &a_12 );   PLA_Obj_free( &x_bottom_source);
   PLA_Obj_free( &a_22 );
   PLA_Obj_free( &y_left_target );   PLA_Obj_free( &y_top_target);
   PLA_Obj_free( &y_right_target);   PLA_Obj_free( &y_bottom_target);
   return(PLA_SUCCESS);
}

int  PLA_Trmv_lower_trans( int diag, PLA_Obj a,
                           PLA_Obj x_source_col,
                           PLA_Obj one,
                           PLA_Obj y_target_row, PLA_Obj y_target_col )
{
   PLA_Obj a_11 = NULL, a_21 = NULL, a_22 = NULL;
   PLA_Obj y_top_target  = NULL,  y_bottom_target = NULL;
   PLA_Obj y_left_target = NULL,  y_right_target  = NULL;
   PLA_Obj x_top_source  = NULL,  x_bottom_source = NULL;

   int     left_size, left_owner, top_size, top_owner, square_size;
   int     my_proc_row, my_proc_col;
   PLA_Template template;

   PLA_Obj_template(a, &template);
   PLA_Temp_comm_col_rank( template, &my_proc_row );
   PLA_Temp_comm_row_rank( template, &my_proc_col );

   PLA_Obj_view_all(a, &a_22);
   PLA_Obj_view_all(x_source_col, &x_bottom_source);
   PLA_Obj_view_all(y_target_col, &y_bottom_target);
   PLA_Obj_view_all(y_target_row, &y_right_target);

   while(TRUE)
   {
      PLA_Obj_split_size( a_22, PLA_SIDE_LEFT, &left_size, &left_owner);
      PLA_Obj_split_size( a_22, PLA_SIDE_TOP,  &top_size,  &top_owner );
      if( 0 == (square_size = min(left_size, top_size))) break;
      PLA_Obj_split_4(a_22, square_size, square_size, &a_11, PLA_DUMMY,
                                                      &a_21, &a_22     );
      PLA_Obj_horz_split_2(y_bottom_target, square_size, &y_top_target,
                                                         &y_bottom_target );
      PLA_Obj_horz_split_2(x_bottom_source, square_size, &x_top_source,
                                                         &x_bottom_source );
      PLA_Obj_vert_split_2(y_right_target, square_size, &y_left_target, &y_right_target);

      if (my_proc_col == left_owner)
      {
         if( my_proc_row == top_owner)
         {
            PLA_Local_copy( x_top_source, y_top_target );
            PLA_Local_trmv( PLA_LOWER_TRIANGULAR, PLA_TRANSPOSE, diag,
                            a_11, y_top_target);
         }
         PLA_Obj_implicit_local_transpose(y_left_target);
         PLA_Local_gemv( PLA_TRANS, one, a_21, x_bottom_source,
                                    one, y_left_target);
         PLA_Obj_implicit_local_transpose(y_left_target); /* unnecessary */
      }
   }
   PLA_Obj_free( &a_11 );   PLA_Obj_free( &x_top_source);
   PLA_Obj_free( &a_21 );   PLA_Obj_free( &x_bottom_source);
   PLA_Obj_free( &a_22 );
   PLA_Obj_free( &y_left_target );   PLA_Obj_free( &y_top_target);
   PLA_Obj_free( &y_right_target);   PLA_Obj_free( &y_bottom_target);
   return(PLA_SUCCESS);
}

int  PLA_Trmv_upper_notrans( int diag, PLA_Obj a, PLA_Obj xtilde,
         PLA_Obj one, PLA_Obj ytilde , PLA_Obj y2tilde)
{
   PLA_Obj a_11 = NULL, a_12 = NULL, a_22 = NULL;
   PLA_Obj x_left = NULL,  x_right  = NULL;
   PLA_Obj y_top  = NULL,  y_bottom = NULL;
   PLA_Obj y2_top  = NULL,  y2_bottom = NULL;

   int     right_size, right_owner, bottom_size, bottom_owner, square_size;
   int     my_proc_row, my_proc_col;
   PLA_Template template;

   PLA_Obj_template(a, &template);
   PLA_Temp_comm_col_rank( template, &my_proc_row );
   PLA_Temp_comm_row_rank( template, &my_proc_col );

   PLA_Obj_view_all(a, &a_11);
   PLA_Obj_view_all(xtilde, &x_left);
   PLA_Obj_view_all(ytilde, &y_top);
   PLA_Obj_view_all(y2tilde, &y2_top);

   while(TRUE)
   {
      PLA_Obj_split_size( a_11, PLA_SIDE_RIGHT, &right_size, &right_owner);
      PLA_Obj_split_size( a_11, PLA_SIDE_BOTTOM,&bottom_size,  &bottom_owner );
      if( 0 == (square_size = min(right_size, bottom_size))) break;
      PLA_Obj_split_4(a_11, -square_size, -square_size, &a_11,     &a_12,
                                                        PLA_DUMMY, &a_22     );
      PLA_Obj_horz_split_2(y_top,   -square_size, &y_top,
                                                  &y_bottom );
      PLA_Obj_horz_split_2(y2_top,   -square_size, &y2_top,
                                                  &y2_bottom );
      PLA_Obj_vert_split_2(x_left,  -square_size, &x_left, &x_right );
      if (my_proc_col == right_owner)
      {
         PLA_Obj_implicit_local_transpose(x_right);
         if( my_proc_row == bottom_owner)
         {
            PLA_Local_copy( x_right, y_bottom);
            PLA_Local_trmv( PLA_UPPER_TRIANGULAR, PLA_NO_TRANSPOSE, diag,
                            a_22, y_bottom);
         }
         PLA_Local_gemv( PLA_NO_TRANS, one, a_12, x_right,   one, y2_top);
         PLA_Obj_implicit_local_transpose(x_right);
      }
   }
   PLA_Obj_free( &a_11 );   PLA_Obj_free( &x_left );   PLA_Obj_free( &y_top);
   PLA_Obj_free( &a_12 );   PLA_Obj_free( &x_right);   PLA_Obj_free( &y_bottom);
   PLA_Obj_free( &a_22 );   PLA_Obj_free( &y2_top); PLA_Obj_free(&y2_bottom);
   return(PLA_SUCCESS);
}


int  PLA_Trmv_lower_notrans( int diag, PLA_Obj a, PLA_Obj xtilde,
                                       PLA_Obj one, 
                                       PLA_Obj ytilde, PLA_Obj y2tilde )
{
   PLA_Obj a_11 = NULL, a_21 = NULL, a_22 = NULL;
   PLA_Obj x_left = NULL,  x_right  = NULL;
   PLA_Obj y_top  = NULL,  y_bottom = NULL;
   PLA_Obj y2_top  = NULL,  y2_bottom = NULL;
   MPI_Datatype datatype;

   int     left_size, left_owner, top_size, top_owner, square_size;
   int     my_proc_row, my_proc_col, nb;
   PLA_Template template;

   PLA_Obj_template(a, &template);
   PLA_Temp_comm_col_rank( template, &my_proc_row );
   PLA_Temp_comm_row_rank( template, &my_proc_col );

   PLA_Obj_view_all(a, &a_22);
   PLA_Obj_view_all(xtilde, &x_right);
   PLA_Obj_view_all(ytilde, &y_bottom);
   PLA_Obj_view_all(y2tilde, &y2_bottom);

   while(TRUE)
   {
      PLA_Obj_split_size( a_22, PLA_SIDE_LEFT, &left_size, &left_owner);
      PLA_Obj_split_size( a_22, PLA_SIDE_TOP,  &top_size,  &top_owner );
      if( 0 == (square_size = min(left_size, top_size))) break;
      PLA_Obj_split_4(a_22, square_size, square_size, &a_11, PLA_DUMMY,
                                                      &a_21, &a_22     );
      PLA_Obj_horz_split_2(y_bottom, square_size, &y_top,
                                                  &y_bottom );
      PLA_Obj_horz_split_2(y2_bottom, square_size, &y2_top,
                                                  &y2_bottom );
      PLA_Obj_vert_split_2(x_right,  square_size, &x_left, &x_right );
      if (my_proc_col == left_owner)
      {
         PLA_Obj_implicit_local_transpose(x_left);
         if( my_proc_row == top_owner)
         {
            PLA_Local_copy( x_left, y_top);
            PLA_Local_trmv( PLA_LOWER_TRIANGULAR, PLA_NO_TRANSPOSE, diag,
                            a_11, y_top);
         }
         PLA_Local_gemv( PLA_NO_TRANS, one, a_21, x_left, one, y2_bottom);
         PLA_Obj_implicit_local_transpose(x_left);
      }
   }
   PLA_Obj_free( &a_11 );   PLA_Obj_free( &x_left );   PLA_Obj_free( &y_top);
   PLA_Obj_free( &a_21 );   PLA_Obj_free( &x_right);   PLA_Obj_free( &y_bottom);
   PLA_Obj_free( &a_22 );   PLA_Obj_free( &y2_top);   PLA_Obj_free(&y2_bottom);
   return(PLA_SUCCESS);
}

int PLA_Trmv(int uplo, int trans, int diag,
             PLA_Obj a, PLA_Obj x)

{

  PLA_Obj      xtilde = NULL, ytilde = NULL, one = NULL;
  PLA_Obj      ytilde2 = NULL; /* for transpose cases */
  MPI_Datatype datatype;
  PLA_Template template;

/*  PLA_Trmv_enter(uplo, trans, diag, a, x); */

  PLA_Obj_datatype(a, &datatype);
  PLA_Obj_template(a, &template);
  PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, &xtilde);
  /* y1tilde will be where the answer goes treated like y <- Ax */
  PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, &ytilde);

  PLA_Copy(x, xtilde);
  PLA_Obj_set_to_zero ( ytilde );
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &one);
  PLA_Obj_set_to_one ( one);
  if(PLA_NO_TRANSPOSE == trans)
  {
     PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, &xtilde);
     /* y1tilde will be where the answer goes treated like y <- Ax */
     PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, &ytilde);
     PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, &ytilde2);
  }
  else
  {
     PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, &xtilde);
     /* y1tilde will be where the answer goes treated like y <- Ax */
     PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, &ytilde);
     PLA_Pvector_create_conf_to( a, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, &ytilde2);
  }
  PLA_Copy(x, xtilde);
  PLA_Obj_set_to_zero ( ytilde );
     PLA_Obj_set_to_zero(ytilde2);
  if( PLA_UPPER_TRIANGULAR == uplo )
  {
     if(PLA_NO_TRANSPOSE == trans)
        PLA_Trmv_upper_notrans( diag, a, xtilde, one, ytilde, ytilde2);
     else
        PLA_Trmv_upper_trans  ( diag, a, xtilde, one, ytilde2, ytilde);
  }
  else
  {
     if(PLA_NO_TRANSPOSE == trans)
        PLA_Trmv_lower_notrans( diag, a, xtilde, one, ytilde, ytilde2);
     else
        PLA_Trmv_lower_trans  ( diag, a, xtilde, one, ytilde2, ytilde);
  }
  PLA_Reduce(ytilde, PLASUM(datatype), x);
  PLA_Reduce_add(ytilde2, MPI_SUM, x);
  PLA_Obj_free( &ytilde);  PLA_Obj_free( &xtilde);
  PLA_Obj_free( &ytilde2); /* for transpose cases */
  PLA_Obj_free( &one);
/*   PLA_Trmv_exit(uplo, trans, diag, a, x); */
  return(0);

}




