/***************************************************************************
  03 April 1996 10:11

  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu

  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu

  Greg Morrow, Department of Physics,
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu

***************************************************************************/
/*
 * $Log: PLA_All_Gemm.c,v $
 * Revision 1.10  1997/03/12 21:12:29  gunnels
 * Used a macro to define betwen reducing a duplicated COMPLEX/DC and
 * built in MPI_SUM handled types (now chooses PLA_SUM or MPI_SUM based
 * on this macro).
 *
 * Revision 1.9  1997/03/11 01:31:41  gunnels
 * Added support for complex/double complex (more involved that I thought).
 *
 * Revision 1.8  1997/03/05 02:38:17  gunnels
 * Removed type dependence in gemm routines.
 * Made panel re-sizing potentially dynamic "back and forth".
 * Tested with Exhaustive.2x2
 *
 * Revision 1.7  1997/02/14 18:11:02  baker
 * Several Modifications resulting from port to Cray.
 * 1)  Added pla_malloc, pla_free routines
 * 2)  combined .h files in UTIL
 *
 * Revision 1.6  1997/02/10 21:15:04  overfelt
 * Had PLA_ALL_ROWS where there should be PLA_ALL_COLS in the
 *
 * Revision 1.5  1997/02/08 20:52:04  overfelt
 * Changed lots of #define statements.  Did not add or reduce functionality.
 * Just trying to make the code a little closer to The Book.
 *
 * A special tag "Before_Changing_define_Constants" was created so these
 * changes could be baked out if needed.
 *
 * Revision 1.4  1997/01/31 01:35:27  overfelt
 * Added PLA_Zero_loc_obj(c1tilde) after the creation to make  sure
 * buffer was initialized to zero.  fix by  Jason Wu .
 *
 * Revision 1.3  1997/01/22 19:05:12  gunnels
 * Updated to have the 4th case (large k, small m,n) included.
 * Note : This also includes some mscalar reduction that will be
 * replaced by PLA_Reduce calls.  That is, the code related to
 * the reduction to a duplicated or non-duplicated mscalar that is
 * in this file should NEVER be called by another function as it is
 * subject to removal at any time.
 *
 * Revision 1.2  1996/10/23  03:30:57  overfelt
 * Changed all LAObj to Obj and LAOBJ to OBJ.
 *
 * Revision 1.1  1996/10/06  22:48:49  overfelt
 * Initial revision
 *
 */

#include "PLA.h"
#include "PLA_Gemm.h"
#include <stdio.h>

#ifndef min3
#define min3(x, y, z)        min(min(x, y), z)
#endif

#ifndef PLASUM
#define PLASUM(datatype) ( datatype == MPI_DOUBLE_COMPLEX ? PLA_SUM : MPI_SUM)
#endif

int PLA_Reduce_dup_mscalar_to_all(PLA_Obj mscalar, MPI_Op operation);
int PLA_Reduce_dup_mscalar_to_matrix(PLA_Obj mscalar, MPI_Op operation, PLA_Obj matrix);


int PLA_Reduce_dup_mscalar_to_all(PLA_Obj mscalar, MPI_Op operation)
{
	int error = 0;
	int  length, width, lda_object, lda_buffer;
	void *send_buffer, *recv_buffer, *object_buffer;
	MPI_Datatype datatype;
  extern  MPI_Op PLA_SUM;
	MPI_Comm  comm1d;
	PLA_Template template;

	PLA_Obj_local_length(mscalar, &length);
	PLA_Obj_local_width(mscalar, &width);
	PLA_Obj_local_ldim(mscalar, &lda_object);
	PLA_Obj_datatype(mscalar, &datatype);
	PLA_Obj_local_buffer(mscalar, &object_buffer);
	PLA_Obj_template(mscalar, &template);
	PLA_Temp_comm1D(template, &comm1d);
	if(MPI_DOUBLE == datatype)
	{
		send_buffer = (void*)pla_malloc(length*width*sizeof(double));
		recv_buffer = (void*)pla_malloc(length*width*sizeof(double));
		PLA_dlacpy("G", &length, &width, object_buffer, &lda_object,
							 send_buffer, &length);
	}
	else if(MPI_FLOAT == datatype)
	{
		send_buffer = (void*)pla_malloc(length*width*sizeof(float));
		recv_buffer = (void*)pla_malloc(length*width*sizeof(float));
		PLA_slacpy("G", &length, &width, object_buffer, &lda_object,
							 send_buffer, &length);
	}
	else if(MPI_COMPLEX == datatype)
	{
		send_buffer = (void*)pla_malloc(2*length*width*sizeof(float));
		recv_buffer = (void*)pla_malloc(2*length*width*sizeof(float));
		PLA_clacpy("G", &length, &width, object_buffer, &lda_object,
							 send_buffer, &length);
	}
	else if(MPI_DOUBLE_COMPLEX == datatype)
	{
		send_buffer = (void*)pla_malloc(2*length*width*sizeof(double));
		recv_buffer = (void*)pla_malloc(2*length*width*sizeof(double));
		PLA_zlacpy("G", &length, &width, object_buffer, &lda_object,
							 send_buffer, &length);
	}

	MPI_Allreduce(send_buffer, recv_buffer, length*width, datatype, operation, comm1d);

	if(MPI_DOUBLE == datatype)
	{
		PLA_dlacpy("G", &length, &width, recv_buffer, &length,
							 object_buffer, &lda_object);
	}
	else if(MPI_FLOAT == datatype)
	{
		PLA_slacpy("G", &length, &width, recv_buffer, &length,
							 object_buffer, &lda_object);
	}
	else if(MPI_COMPLEX == datatype)
	{
		PLA_clacpy("G", &length, &width, recv_buffer, &length,
							 object_buffer, &lda_object);
	}
	else if(MPI_DOUBLE_COMPLEX == datatype)
	{
		PLA_zlacpy("G", &length, &width, recv_buffer, &length,
							 object_buffer, &lda_object);
	}

	if(send_buffer) pla_free(send_buffer);
	if(recv_buffer) pla_free(recv_buffer);
}

int PLA_Reduce_dup_mscalar_to_matrix(PLA_Obj omscalar, MPI_Op operation, PLA_Obj omatrix)
{
	int lda_matrix, lda_mscalar, top_size, left_size;
	int my_col, my_row, left_owner, top_owner, i, j, idummy;
	PLA_Obj matrix = NULL, mscalar = NULL;
	PLA_Obj A1 = NULL, A2 = NULL, A11 = NULL, A21 = NULL;
	PLA_Obj B1 = NULL, B2 = NULL, B11 = NULL, B21 = NULL;
	void *matrix_buffer, *mscalar_buffer;
	MPI_Comm dummy;
	PLA_Template template;
	MPI_Datatype datatype;
  extern  MPI_Op PLA_SUM;

	PLA_Obj_view_all(omatrix, &matrix);
	PLA_Obj_view_all(omscalar, &mscalar);
	PLA_Obj_template(matrix, &template);
	PLA_Temp_comm_row_info(template, &dummy, &my_col, &idummy);
	PLA_Temp_comm_col_info(template, &dummy, &my_row, &idummy);
	PLA_Obj_local_ldim(matrix, &lda_matrix);
	PLA_Obj_local_ldim(mscalar, &lda_mscalar);
	PLA_Obj_datatype(mscalar, &datatype);

	while(TRUE)
	{
		PLA_Obj_split_size(matrix, PLA_SIDE_LEFT, &left_size, &left_owner);
		if(0 == left_size) break;
		PLA_Obj_vert_split_2(matrix, left_size, &A1, &A2);
		PLA_Obj_vert_split_2(mscalar, left_size, &B1, &B2);
		while(TRUE)
		{
			PLA_Obj_split_size(A1, PLA_SIDE_TOP, &top_size, &top_owner);
			if(0 == top_size) break;
			PLA_Obj_horz_split_2(A1, top_size, &A11, &A21);
			PLA_Obj_horz_split_2(B1, top_size, &B11, &B21);
			if( (my_col == left_owner) && (my_row == top_owner) )
			{
				PLA_Obj_local_buffer(A11, &matrix_buffer);
				PLA_Obj_local_buffer(B11, &mscalar_buffer);
				for(i = 0; i < left_size; i++)
				{
					for(j = 0; j < top_size; j++)
					{
					  if(MPI_DOUBLE == datatype)
						  *(((double*)matrix_buffer)+j) = *(((double*)matrix_buffer)+j) +
																	 *(((double*)mscalar_buffer)+j);
					  else if(MPI_FLOAT == datatype)
						  *(((float*)matrix_buffer)+j) = *(((float*)matrix_buffer)+j) +
																	 *(((float*)mscalar_buffer)+j);
					  else if(MPI_COMPLEX == datatype)
					  {
						  *(((float*)matrix_buffer)+j*2) = *(((float*)matrix_buffer)+j*2) +
																	 *(((float*)mscalar_buffer)+j*2);
						  *(((float*)matrix_buffer)+j*2+1) = *(((float*)matrix_buffer)+j*2+1) +
																	 *(((float*)mscalar_buffer)+j*2+1);
					  }
					  else if(MPI_DOUBLE_COMPLEX == datatype)
					  {
						  *(((double*)matrix_buffer)+j*2) = *(((double*)matrix_buffer)+j*2) +
																	 *(((double*)mscalar_buffer)+j*2);
						  *(((double*)matrix_buffer)+j*2+1) = *(((double*)matrix_buffer)+j*2+1) +
																	 *(((double*)mscalar_buffer)+j*2+1);
					  }
					}
					if(MPI_DOUBLE == datatype)
					{
						matrix_buffer = ((double*)matrix_buffer)   + lda_matrix;
						mscalar_buffer = ((double*)mscalar_buffer) + lda_mscalar;
					}
					else if(MPI_FLOAT == datatype)
					{
						matrix_buffer = ((float*)matrix_buffer)   + lda_matrix;
						mscalar_buffer = ((float*)mscalar_buffer) + lda_mscalar;
					}
					else if(MPI_COMPLEX == datatype)
					{
						matrix_buffer = ((float*)matrix_buffer)   + lda_matrix;
						mscalar_buffer = ((float*)mscalar_buffer) + lda_mscalar;
					}
					else if(MPI_DOUBLE_COMPLEX == datatype)
					{
						matrix_buffer = ((double*)matrix_buffer)   + lda_matrix*2;
						mscalar_buffer = ((double*)mscalar_buffer) + lda_mscalar*2;
					}
				 }
			}
			PLA_Obj_view_swap( &A1, &A21);
			PLA_Obj_view_swap( &B1, &B21);
		}
		PLA_Obj_view_swap( &matrix, &A2);
		PLA_Obj_view_swap( &mscalar, &B2);
	}
	PLA_Obj_free( &mscalar ) ; PLA_Obj_free( &matrix);
	PLA_Obj_free( &A1 );PLA_Obj_free( &A2 );PLA_Obj_free( &A11 );PLA_Obj_free( &A21 );
	PLA_Obj_free( &B1 );PLA_Obj_free( &B2 );PLA_Obj_free( &B11 );PLA_Obj_free( &B21 );
}

/* Handles the cases for No Transpose A / No Transpose B for :
	9.1.6.1
	9.1.6.2
	9.1.6.4
	9.1.6.6
	9.1.6.8
*/
int PLA_Gemm_notransa_notransb_1(int nb_alg, int extra_ta, int extra_tb, PLA_Obj alpha, PLA_Obj a,
											  PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
  PLA_Obj  acur = NULL, bcur = NULL,
			  a1 = NULL,   b1t = NULL,   One = NULL,
			  a2 = NULL,   b2t = NULL,
			  a1tilde = NULL, b1tilde = NULL;
  int size, width_a, length_b;
  PLA_Template  template = NULL;
  MPI_Datatype datatype;
  extern  MPI_Op PLA_SUM;

  PLA_Scal( beta, c );
  PLA_Obj_template(c, &template);
  PLA_Obj_datatype(c, &datatype);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &One );
  PLA_Obj_set_to_one(One);
									  /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST,
																  PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST,
												PLA_ALIGN_FIRST, &bcur );
  PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb_alg,
										  &a1tilde );
  PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb_alg,
										  &b1tilde );
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
	 PLA_Obj_global_width ( acur, &width_a );
	 PLA_Obj_global_length( bcur, &length_b );
	 if ( ( size = min3( width_a, length_b, nb_alg )) == 0 ) break;
	 PLA_Obj_vert_split_2 ( acur, size,              &a1,   &a2 );
	 PLA_Obj_horz_split_2 ( bcur, size,              &b1t,  &b2t );
	 PLA_Obj_view_shift(a1tilde,                   0,
										  nb_alg - size,          0,
																  0);
	 PLA_Obj_view_shift(b1tilde,           nb_alg - size,
													0,               0,
																  0);
									  /* annotate the views                 */
	 PLA_Obj_set_orientation( a1, PLA_PROJ_ONTO_COL );
	 PLA_Obj_set_orientation( b1t, PLA_PROJ_ONTO_ROW );
		  /* spread a1 and a2                   */
	 PLA_Copy( a1, a1tilde );
	 PLA_Copy( b1t, b1tilde );
									  /* perform local rank-size update     */
	 PLA_Local_gemm(PLA_NO_TRANS,PLA_NO_TRANS, alpha, a1tilde, b1tilde,
														 One, c);
	 PLA_Obj_swap( &a2,  &acur );
	 PLA_Obj_swap( &b2t, &bcur );
	 PLA_Obj_view_shift(a1tilde,                   0,
										  size - nb_alg,          0,
																  0);
	 PLA_Obj_view_shift(b1tilde,           size - nb_alg,
													0,               0,
																  0);
  }
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1 );  PLA_Obj_free( &a2 );
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1t );  PLA_Obj_free( &b2t );
  PLA_Obj_free( &b1tilde );  PLA_Obj_free( &a1tilde );  PLA_Obj_free( &One);
}
/********************  NEXT CASE **********************************************/
/* Handles the cases for No Transpose A / No Transpose B for :
	9.1.6.3
*/
int PLA_Gemm_notransa_notransb_2(int nb_alg, int extra_ta, int extra_tb,PLA_Obj alpha, PLA_Obj a,
											  PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
  PLA_Obj  bcur = NULL, ccur = NULL,
			  b1 = NULL,   c1 = NULL,   Zero = NULL,
			  b2 = NULL,   c2 = NULL,
			  b1tilde = NULL,  c1tilde = NULL;
  int size, width_b, width_c;
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;
									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Obj_datatype(c, &datatype);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &Zero );
  PLA_Obj_set_to_zero(Zero);

									  /* take a view of all of both A and B */
  PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL,
							 PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &ccur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,&bcur );
/* create dupl. pmvectors for spreading panels of B and collecting into  C */
  PLA_Pmvector_create_conf_to( a, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
											nb_alg, &b1tilde );
  PLA_Pmvector_create_conf_to( a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
											nb_alg, &c1tilde );
  PLA_Obj_set_to_zero(c1tilde);
									  /* Loop until no more of A and B      */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
									  /* determine width of next update     */
	 PLA_Obj_global_width ( bcur, &width_b );
	 PLA_Obj_global_width ( ccur, &width_c );
	 if ( ( size = min3( width_b, width_c, nb_alg )) == 0 ) break;
									  /* split off first column of Bcur     */
	 PLA_Obj_vert_split_2( bcur, size,              &b1, &b2 );
									  /* split off first row of Ccur        */
	 PLA_Obj_vert_split_2( ccur, size,              &c1, &c2 );
									  /* size the workspaces                */
	 PLA_Obj_view_shift(b1tilde,           nb_alg - size,
													0,               0,
																  0);
	 PLA_Obj_view_shift(c1tilde,                   0,
										  nb_alg - size,          0,
																  0);
									  /* annotate the views                 */
	 PLA_Obj_set_orientation( b1, PLA_PROJ_ONTO_COL );
	 PLA_Obj_set_orientation( c1, PLA_PROJ_ONTO_COL );
									  /* spread a1 and a2                   */
	 PLA_Copy( b1, b1tilde );
/* It might make sense to call the following since we are reducing into this */
/* duplicated panel -- however, it is easier just to use 0 for that part of  */
/* the local gemm call -- that is beta == 0                                  */
/* This may be necessary if malloc sets things to NaN for example            */
/*    PLA_Zero_loc_obj(c1tilde);   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!     */
									  /* perform local matrix-panel update     */
/* An implicit transpose has already happened so we need another in the kernel*/
	 PLA_Local_gemm(PLA_NO_TRANS, PLA_TRANS, alpha, a,  b1tilde,
						  Zero, c1tilde);
	 PLA_Reduce_add(c1tilde, PLASUM(datatype), c1);
	 PLA_Obj_swap( &b2, &bcur );
	 PLA_Obj_swap( &c2, &ccur );
	 PLA_Obj_view_shift(b1tilde,           size - nb_alg,
													0,               0,
																  0);
	 PLA_Obj_view_shift(c1tilde,                   0,
										  size - nb_alg,          0,
																  0);
  }
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1 );  PLA_Obj_free( &b2 );
  PLA_Obj_free( &ccur );  PLA_Obj_free( &c1 );  PLA_Obj_free( &c2 );
  PLA_Obj_free( &b1tilde ); PLA_Obj_free( &c1tilde ); PLA_Obj_free(&Zero);
}
/********************  NEXT CASE **********************************************/
/* Handles the cases for No Transpose A / No Transpose B for :
	9.1.6.5
	9.1.6.7
*/
int PLA_Gemm_notransa_notransb_3(int nb_alg, int extra_ta, int extra_tb,PLA_Obj alpha, PLA_Obj a,
											  PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
  PLA_Obj  acur = NULL, ccur = NULL,
			  a1t = NULL,   c1t = NULL,   Zero = NULL,
			  a2t = NULL,   c2t = NULL,   a1tilde = NULL,  c1tilde = NULL;
  int size, length_a, length_c;
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;

									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Obj_datatype(c, &datatype);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &Zero );
  PLA_Obj_set_to_zero(Zero);

									  /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
							 PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,&ccur );
/* create dupl. pmvectors for spreading panels of B and collecting into  C */
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
											nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
											nb_alg, &c1tilde );
  PLA_Obj_set_to_zero(c1tilde);
									  /* Loop until no more of A and B      */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
									  /* determine width of next update     */
	 PLA_Obj_global_length ( acur, &length_a );
	 PLA_Obj_global_length ( ccur, &length_c );
	 if ( ( size = min3( length_a, length_c, nb_alg )) == 0 ) break;
									  /* split off first column of Acur     */
	 PLA_Obj_horz_split_2( acur, size,              &a1t, &a2t );
									  /* split off first row of Bcur        */
	 PLA_Obj_horz_split_2( ccur, size,              &c1t, &c2t );
									  /* size the workspaces                */
	 PLA_Obj_view_shift(a1tilde,                   0,
										  nb_alg - size,          0,
																  0);
	 PLA_Obj_view_shift(c1tilde,           nb_alg - size,
													0,               0,
																  0);
									  /* annotate the views                 */
	 PLA_Obj_set_orientation( a1t, PLA_PROJ_ONTO_ROW );
	 PLA_Obj_set_orientation( c1t, PLA_PROJ_ONTO_ROW );
									  /* spread a1 and a2                   */
	 PLA_Copy( a1t, a1tilde );
/* An implicit transpose has already happened so we need another in the kernel*/
	 PLA_Local_gemm(PLA_TRANS, PLA_NO_TRANS, alpha, a1tilde, b,
						  Zero, c1tilde);
	 PLA_Reduce_add(c1tilde, PLASUM(datatype), c1t);
	 PLA_Obj_swap( &a2t, &acur );
	 PLA_Obj_swap( &c2t, &ccur );
	 PLA_Obj_view_shift(a1tilde,                   0,
										  size - nb_alg,          0,
																  0);
	 PLA_Obj_view_shift(c1tilde,           size - nb_alg,
													0,               0,
																  0);
  }
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1t );  PLA_Obj_free( &a2t );
  PLA_Obj_free( &ccur );  PLA_Obj_free( &c1t );  PLA_Obj_free( &c2t );
  PLA_Obj_free( &a1tilde );  PLA_Obj_free( &c1tilde ); PLA_Obj_free(&Zero);
}

/* Small m, small n, large k */
int PLA_Gemm_notransa_notransb_4(int nb_alg, int extra_ta, int extra_tb,PLA_Obj alpha, PLA_Obj a,
											  PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;
  MPI_Comm      dummy_comm;
  int nb, num_rows, idummy;

  int m, n, size_a, size_b;

  int zero_or_one, msub, nsub, length_mvector, length_c, width_c;

  PLA_Obj  acur = NULL, a_mvector = NULL, a_mvector_cur = NULL,
			  acur_row = NULL, acur_rest_rows = NULL,
			  bcur = NULL, b_mvector = NULL, b_mvector_cur = NULL,
			  bcur_col = NULL, bcur_rest_cols = NULL,
			  ccur = NULL, c_mscalar_cur_outer = NULL, c_mscalar_cur_inner = NULL,
			  ccur_row = NULL, c_square = NULL, c_mscalar = NULL,
			  ccur_rest_rows = NULL, ccur_rest_cols = NULL,
			  Zero = NULL;

									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Temp_zero_or_one(template, &zero_or_one);
  PLA_Temp_nb( template, &nb);
  PLA_Temp_comm_row_info(template, &dummy_comm, &idummy, &num_rows);
  /* since we are going ONLY to multivectors we can use a bit more storage */
  if(nb_alg < (nb * num_rows)) nb_alg = nb * num_rows;
  PLA_Obj_datatype(c, &datatype);

  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template,
							 &Zero );
  PLA_Obj_set_to_zero(Zero);
									  /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
							 PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,&bcur );
  PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,&ccur );
  PLA_Obj_global_width(a, &length_mvector);
  PLA_Obj_global_length(c, &m);
  PLA_Obj_global_width(c, &n);
  size_a = min(nb_alg, m);
  size_b = min(nb_alg, n);
/* create mvectors for spreading panels of A and B  */
  PLA_Mvector_create( datatype, length_mvector, size_a, template, zero_or_one,
							 &a_mvector);
  PLA_Mvector_create( datatype, length_mvector, size_b, template, zero_or_one,
							 &b_mvector);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, size_a, size_b,
							 template, &c_mscalar);

									  /* Loop until no more of A and B      */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
									  /* determine width of next update     */
	 PLA_Obj_global_length ( ccur, &length_c );
	 if ( ( msub = min( length_c, size_a )) == 0 ) break;
	 PLA_Obj_horz_split_2(ccur, msub, &ccur_row, &ccur_rest_rows);
	 PLA_Obj_horz_split_2(acur, msub, &acur_row, &acur_rest_rows);
	 PLA_Obj_set_orientation(acur_row, PLA_PROJ_ONTO_ROW);
	 if(msub < size_a)
	 {
		 PLA_Obj_vert_split_2 ( a_mvector, msub,          &a_mvector_cur, PLA_DUMMY );
		 PLA_Obj_swap(&a_mvector, &a_mvector_cur);
		 PLA_Obj_horz_split_2 ( c_mscalar, msub,          &c_mscalar_cur_outer, PLA_DUMMY );
		 PLA_Obj_swap(&c_mscalar_cur_outer, &c_mscalar);
	 }
	 PLA_Copy(acur_row, a_mvector);
	 PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,&bcur );
	 while(TRUE)
	 {
		 PLA_Obj_global_width(ccur_row, &width_c);
		 if( (nsub = min( width_c, size_b)) == 0) break;
		 PLA_Obj_vert_split_2(ccur_row, nsub, &c_square, &ccur_rest_cols);
		 PLA_Obj_vert_split_2(bcur, nsub, &bcur_col, &bcur_rest_cols);
		 PLA_Obj_set_orientation(bcur_col, PLA_PROJ_ONTO_COL);
		 if(nsub < size_b)
		 {
			 PLA_Obj_vert_split_2 ( b_mvector, nsub,          &b_mvector_cur, PLA_DUMMY );
			 PLA_Obj_view_swap(&b_mvector, &b_mvector_cur);
			 PLA_Obj_vert_split_2 ( c_mscalar, msub,          &c_mscalar_cur_inner, PLA_DUMMY );
			 PLA_Obj_view_swap(&c_mscalar, &c_mscalar_cur_inner);
		 }
		 PLA_Copy(bcur_col, b_mvector);
		 PLA_Local_gemm(PLA_TRANS, PLA_NO_TRANS, alpha, a_mvector, b_mvector, Zero, c_mscalar);
		 PLA_Reduce_dup_mscalar_to_all(c_mscalar, PLASUM(datatype));
		 PLA_Reduce_dup_mscalar_to_matrix(c_mscalar, PLASUM(datatype), c_square);
		 if(nsub < size_b)
		 {
			PLA_Obj_view_swap(&c_mscalar, &c_mscalar_cur_inner);
			PLA_Obj_view_swap(&b_mvector, &b_mvector_cur);
		 }
		 PLA_Obj_view_swap( &ccur_row, &ccur_rest_cols);
		 PLA_Obj_view_swap( &bcur,     &bcur_rest_cols);
	 } /* End inner loop */
	 PLA_Obj_view_swap( &ccur, &ccur_rest_rows);
	 PLA_Obj_view_swap( &acur, &acur_rest_rows);
	 if(msub < size_a)
	 {
		  PLA_Obj_view_swap(&c_mscalar, &c_mscalar_cur_outer);
		  PLA_Obj_view_swap(&a_mvector, &a_mvector_cur);
	 }
  } /* End outer loop */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
  PLA_Obj_free( &acur); PLA_Obj_free( &a_mvector); PLA_Obj_free( &a_mvector_cur );
  PLA_Obj_free( &acur_row ); PLA_Obj_free( &bcur ); PLA_Obj_free( & b_mvector );
  PLA_Obj_free( &b_mvector_cur); PLA_Obj_free( &bcur_col ); PLA_Obj_free( &ccur );
  PLA_Obj_free( &c_mscalar_cur_outer ); PLA_Obj_free( &c_mscalar_cur_inner );
  PLA_Obj_free( &ccur_row ); PLA_Obj_free( &c_square ); PLA_Obj_free( &c_mscalar);
  PLA_Obj_free( &Zero );
  PLA_Obj_free( &ccur_rest_cols); PLA_Obj_free( &ccur_rest_rows);
  PLA_Obj_free( &acur_rest_rows); PLA_Obj_free( &bcur_rest_cols);
}


/************* END NO TRANSPOSE A / NO TRANSPOSE B CASE **********************/

/* Handles the cases for Transpose A /  Transpose B for :
	9.1.6.1
	9.1.6.2
	9.1.6.4
	9.1.6.6
	9.1.6.7 -- probably a better way to do this One
	9.1.6.8
*/

int PLA_Gemm_transa_transb_1(
						  int nb_alg, int extra_ta, int extra_tb, PLA_Obj alpha, PLA_Obj a, PLA_Obj b,
						  PLA_Obj beta, PLA_Obj c )
{
  PLA_Obj  acur = NULL, bcur = NULL,
			  a1t = NULL,   b1 = NULL,   One = NULL,
			  a2t = NULL,   b2 = NULL,   a1tilde = NULL, b1tilde = NULL;
  int size, width_b, length_a;
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;
  int           transa, transb; /* for COMPLEX cases */
									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Obj_datatype(c, &datatype);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &One );
  PLA_Obj_set_to_one(One);
  transa = PLA_NO_TRANSPOSE;
  transb = PLA_NO_TRANSPOSE;
  if( (MPI_COMPLEX == datatype) || (MPI_DOUBLE_COMPLEX == datatype) )
  {
	  if( PLA_CONJ_TRANSPOSE == extra_ta)
		  transa = PLA_CONJ;
	  if( PLA_CONJ_TRANSPOSE == extra_tb)
		  transb = PLA_CONJ;
  }
									  /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &bcur );
 /* create dupl. pmvectors for spreading panels of A and B */
  PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
											nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
											nb_alg, &b1tilde );
									  /* Loop until no more of A and B      */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
									  /* determine width of next update     */
	 PLA_Obj_global_length( acur, &length_a );
	 PLA_Obj_global_width ( bcur, &width_b );
	 if ( ( size = min3( length_a, width_b, nb_alg )) == 0 ) break;
									  /* split off first column of Acur     */
	 PLA_Obj_horz_split_2 ( acur, size,              &a1t, &a2t );
									  /* split off first row of Bcur        */
	 PLA_Obj_vert_split_2( bcur, size,              &b1, &b2 );
									  /* size the workspaces                */
	 PLA_Obj_view_shift(a1tilde,                   0,
										  nb_alg - size,          0,
																  0);
	 PLA_Obj_view_shift(b1tilde,           nb_alg - size,
													0,               0,
																  0);
									  /* annotate the views                 */
	 PLA_Obj_set_orientation( a1t, PLA_PROJ_ONTO_ROW );
	 PLA_Obj_set_orientation( b1,  PLA_PROJ_ONTO_COL );
									  /* spread a1 and a2                   */

	 PLA_Copy( a1t, a1tilde);
	 PLA_Copy( b1,  b1tilde);
	 PLA_Local_gemm(transa, transb, alpha, a1tilde, b1tilde, One, c);
	 PLA_Obj_swap( &a2t, &acur );
	 PLA_Obj_swap( &b2, &bcur );
	 PLA_Obj_view_shift(a1tilde,                   0,
										  size - nb_alg,          0,
																  0);
	 PLA_Obj_view_shift(b1tilde,           size - nb_alg,
													0,               0,
																  0);
  }
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1t );  PLA_Obj_free( &a2t );
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1 );   PLA_Obj_free( &b2 );
  PLA_Obj_free( &b1tilde ); PLA_Obj_free( &a1tilde );  PLA_Obj_free(&One);


}

/* Handles the cases for Transpose A /  Transpose B for :
	9.1.6.3
*/

int PLA_Gemm_transa_transb_2(
						  int nb_alg, int extra_ta, int extra_tb, PLA_Obj alpha, PLA_Obj a, PLA_Obj b,
						  PLA_Obj beta, PLA_Obj c )
{
  PLA_Obj  bcur = NULL, ccur = NULL,
			  b1t = NULL,   c1 = NULL,   Zero = NULL,
			  b2t = NULL,   c2 = NULL,   b1tilde = NULL, c1tilde = NULL;
  int size, length_b, width_c;
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;
  int transa, transb; /* COMPLEX cases */
									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Obj_datatype(c, &datatype);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &Zero );
  PLA_Obj_set_to_zero(Zero);
  transa = extra_ta;
  transb = PLA_NO_TRANSPOSE;
  if( (MPI_COMPLEX == datatype) || (MPI_DOUBLE_COMPLEX == datatype) )
  {
	  if( PLA_CONJ_TRANSPOSE == extra_tb)
		  transb = PLA_CONJ;
  }
									  /* take a view of all of both A and B */
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &bcur );
  PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &ccur );
  PLA_Pmvector_create_conf_to( a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
											nb_alg, &b1tilde );
  PLA_Pmvector_create_conf_to( a, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
											nb_alg, &c1tilde );
  PLA_Obj_set_to_zero(c1tilde);
									  /* Loop until no more of A and B      */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
									  /* determine width of next update     */
/* I don't think that we need two of these -- they both run out at the
	same time or you get the wrong answer anyway */
	 PLA_Obj_global_length( bcur, &length_b );
	 PLA_Obj_global_width ( ccur, &width_c );
	 if ( ( size = min3( length_b, width_c, nb_alg )) == 0 ) break;
									  /* split off first column of Acur     */
	 PLA_Obj_horz_split_2 ( bcur, size,              &b1t, &b2t );
									  /* split off first row of Bcur        */
	 PLA_Obj_vert_split_2 ( ccur, size,              &c1, &c2 );
									  /* size the workspaces                */
	 PLA_Obj_view_shift(b1tilde,                   0,
										  nb_alg - size,          0,
																  0);
	 PLA_Obj_view_shift(c1tilde,           nb_alg - size,
													0,               0,
																  0);
									  /* annotate the views                 */
	 PLA_Obj_set_orientation( b1t, PLA_PROJ_ONTO_ROW );
	 PLA_Obj_set_orientation( c1,  PLA_PROJ_ONTO_COL );
									  /* spread a1 and a2                   */

	 PLA_Copy( b1t, b1tilde);
	 PLA_Local_gemm(transa, transb, alpha, b1tilde, a, Zero, c1tilde);
/* I know what I want to do here -- I am not certain that reduce handles it */
	 PLA_Reduce_add( c1tilde, PLASUM(datatype), c1);
	 PLA_Obj_swap( &b2t, &bcur );
	 PLA_Obj_swap( &c2,  &ccur );
	 PLA_Obj_view_shift(b1tilde,                   0,
										  size - nb_alg,          0,
																  0);
	 PLA_Obj_view_shift(c1tilde,           size - nb_alg,
													0,               0,
																  0);
  }
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1t );  PLA_Obj_free( &b2t );
  PLA_Obj_free( &ccur );  PLA_Obj_free( &c1 );   PLA_Obj_free( &c2 );
  PLA_Obj_free( &c1tilde );   PLA_Obj_free( &b1tilde );  PLA_Obj_free(&Zero);
}

/* Handles the cases for Transpose A /  Transpose B for :
	9.1.6.5
*/

int PLA_Gemm_transa_transb_3(
						  int nb_alg, int extra_ta, int extra_tb, PLA_Obj alpha, PLA_Obj a, PLA_Obj b,
						  PLA_Obj beta, PLA_Obj c )
{
  PLA_Obj  acur = NULL, ccur = NULL,
			  a1 = NULL,   c1t = NULL,   Zero = NULL,
			  a2 = NULL,   c2t = NULL,   a1tilde = NULL, c1tilde = NULL;
  int size, width_a, length_c;
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;
  int           transa, transb; /* COMPLEX cases */
									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Obj_datatype(c, &datatype);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &Zero );
  PLA_Obj_set_to_zero(Zero);
  transa = PLA_NO_TRANSPOSE;
  transb = extra_tb;
  if( (MPI_COMPLEX == datatype) || (MPI_DOUBLE_COMPLEX == datatype) )
  {
	  if( PLA_CONJ_TRANSPOSE == extra_ta)
		  transa = PLA_CONJ;
  }

									  /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
							 PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &ccur );
									  /* create dupl. pmvectors for spreading
										  panels of A and B */
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
											nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
											nb_alg, &c1tilde );
  PLA_Obj_set_to_zero(c1tilde);
									  /* Loop until no more of A and B      */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
									  /* determine width of next update     */
	 PLA_Obj_global_width  ( acur, &width_a );
	 PLA_Obj_global_length ( ccur, &length_c );
	 if ( ( size = min3( width_a, length_c, nb_alg )) == 0 ) break;
									  /* split off first column of Acur     */
	 PLA_Obj_vert_split_2 ( acur, size,              &a1, &a2   );
									  /* split off first row of Bcur        */
	 PLA_Obj_horz_split_2 ( ccur, size,              &c1t, &c2t );
									  /* size the workspaces                */
	 PLA_Obj_view_shift(c1tilde,                   0,
										  nb_alg - size,          0,
																  0);
	 PLA_Obj_view_shift(a1tilde,           nb_alg - size,
													0,               0,
																  0);
									  /* annotate the views                 */
	 PLA_Obj_set_orientation( a1,  PLA_PROJ_ONTO_COL );
	 PLA_Obj_set_orientation( c1t, PLA_PROJ_ONTO_ROW );
									  /* spread a1 and a2                   */
	 PLA_Copy( a1, a1tilde);
	 PLA_Local_gemm(transa, transb, alpha, b, a1tilde, Zero, c1tilde);
	 PLA_Reduce_add( c1tilde, PLASUM(datatype), c1t);
	 PLA_Obj_swap( &a2,  &acur );
	 PLA_Obj_swap( &c2t, &ccur );
	 PLA_Obj_view_shift(c1tilde,                   0,
										  size - nb_alg,          0,
																  0);
	 PLA_Obj_view_shift(a1tilde,           size - nb_alg,
													0,               0,
																  0);
  }
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
/* Problematic placeholder xxxxxxxxxxxxxxxxx */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1 );   PLA_Obj_free( &a2 );
  PLA_Obj_free( &ccur );  PLA_Obj_free( &c1t );  PLA_Obj_free( &c2t );
  PLA_Obj_free( &a1tilde ); PLA_Obj_free( &c1tilde ); PLA_Obj_free(&Zero);
}

/* Small m, small n, large k */
int PLA_Gemm_transa_transb_4(int nb_alg, int extra_ta, int extra_tb, PLA_Obj alpha, PLA_Obj a,
									  PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;
  MPI_Comm      dummy_comm;
  int nb, num_rows, idummy;

  int m, n, size_a, size_b;

  int zero_or_one, msub, nsub, length_mvector, length_c, width_c;

  PLA_Obj  acur = NULL, a_mvector = NULL, a_mvector_cur = NULL,
			  acur_row = NULL, acur_rest_rows = NULL,
			  bcur = NULL, b_mvector = NULL, b_mvector_cur = NULL,
			  bcur_col = NULL, bcur_rest_cols = NULL,
			  ccur = NULL, c_mscalar_cur_outer = NULL, c_mscalar_cur_inner = NULL,
			  ccur_row = NULL, c_square = NULL, c_mscalar = NULL,
			  ccur_rest_rows = NULL, ccur_rest_cols = NULL,
			  Zero = NULL;
  int  transa, transb;

									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Temp_zero_or_one(template, &zero_or_one);
  PLA_Temp_nb( template, &nb);
  PLA_Temp_comm_row_info(template, &dummy_comm, &idummy, &num_rows);
  /* since we are going ONLY to multivectors we can use a bit more storage */
  if(nb_alg < (nb * num_rows)) nb_alg = nb * num_rows;
  PLA_Obj_datatype(c, &datatype);

  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &Zero );
  PLA_Obj_set_to_zero(Zero);
  transa = extra_ta;
  transb = PLA_NO_TRANSPOSE;
  if( (MPI_COMPLEX == datatype) || (MPI_DOUBLE_COMPLEX == datatype) )
  {
	  if( PLA_CONJ_TRANSPOSE == extra_tb)
		  transb = PLA_CONJ;
  }
									  /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
							 PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,&bcur );
  PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,&ccur );
  PLA_Obj_global_length(a, &length_mvector);
  PLA_Obj_global_length(c, &m);
  PLA_Obj_global_width(c, &n);
  size_a = min(nb_alg, m);
  size_b = min(nb_alg, n);
/* create mvectors for spreading panels of A and B  */
  PLA_Mvector_create( datatype, length_mvector, size_a, template, zero_or_one,
							 &a_mvector);
  PLA_Mvector_create( datatype, length_mvector, size_b, template, zero_or_one,
							 &b_mvector);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, size_a, size_b,
							 template, &c_mscalar);

									  /* Loop until no more of A and B      */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
									  /* determine width of next update     */
	 PLA_Obj_global_length ( ccur, &length_c );
	 if ( ( msub = min( length_c, size_a )) == 0 ) break;
	 PLA_Obj_horz_split_2(ccur, msub, &ccur_row, &ccur_rest_rows);
	 PLA_Obj_vert_split_2(acur, msub, &acur_row, &acur_rest_rows);
	 PLA_Obj_set_orientation(acur_row, PLA_PROJ_ONTO_COL);
	 if(msub < size_a)
	 {
		 PLA_Obj_vert_split_2 ( a_mvector, msub,          &a_mvector_cur, PLA_DUMMY );
		 PLA_Obj_swap(&a_mvector, &a_mvector_cur);
		 PLA_Obj_horz_split_2 ( c_mscalar, msub,          &c_mscalar_cur_outer, PLA_DUMMY );
		 PLA_Obj_swap(&c_mscalar_cur_outer, &c_mscalar);
	 }
	 PLA_Copy(acur_row, a_mvector);
	 PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,&bcur );
	 while(TRUE)
	 {
		 PLA_Obj_global_width(ccur_row, &width_c);
		 if( (nsub = min( width_c, size_b)) == 0) break;
		 PLA_Obj_vert_split_2(ccur_row, nsub, &c_square, &ccur_rest_cols);
		 PLA_Obj_horz_split_2(bcur, nsub, &bcur_col, &bcur_rest_cols);
		 PLA_Obj_set_orientation(bcur_col, PLA_PROJ_ONTO_ROW);
		 if(nsub < size_b)
		 {
			 PLA_Obj_vert_split_2 ( b_mvector, nsub,          &b_mvector_cur, PLA_DUMMY );
			 PLA_Obj_view_swap(&b_mvector, &b_mvector_cur);
			 PLA_Obj_vert_split_2 ( c_mscalar, msub,          &c_mscalar_cur_inner, PLA_DUMMY );
			 PLA_Obj_view_swap(&c_mscalar, &c_mscalar_cur_inner);
		 }
		 PLA_Copy(bcur_col, b_mvector);
		 PLA_Local_gemm(transa, transb, alpha, a_mvector, b_mvector, Zero, c_mscalar);
		 PLA_Reduce_dup_mscalar_to_all(c_mscalar, PLASUM(datatype));
		 PLA_Reduce_dup_mscalar_to_matrix(c_mscalar, PLASUM(datatype), c_square);
		 if(nsub < size_b)
		 {
			PLA_Obj_view_swap(&c_mscalar, &c_mscalar_cur_inner);
			PLA_Obj_view_swap(&b_mvector, &b_mvector_cur);
		 }
		 PLA_Obj_view_swap( &ccur_row, &ccur_rest_cols);
		 PLA_Obj_view_swap( &bcur,     &bcur_rest_cols);
	 } /* End inner loop */
	 PLA_Obj_view_swap( &ccur, &ccur_rest_rows);
	 PLA_Obj_view_swap( &acur, &acur_rest_rows);
	 if(msub < size_a)
	 {
		  PLA_Obj_view_swap(&c_mscalar, &c_mscalar_cur_outer);
		  PLA_Obj_view_swap(&a_mvector, &a_mvector_cur);
	 }
  } /* End outer loop */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
  PLA_Obj_free( &acur); PLA_Obj_free( &a_mvector); PLA_Obj_free( &a_mvector_cur );
  PLA_Obj_free( &acur_row ); PLA_Obj_free( &bcur ); PLA_Obj_free( & b_mvector );
  PLA_Obj_free( &b_mvector_cur); PLA_Obj_free( &bcur_col ); PLA_Obj_free( &ccur );
  PLA_Obj_free( &c_mscalar_cur_outer ); PLA_Obj_free( &c_mscalar_cur_inner );
  PLA_Obj_free( &ccur_row ); PLA_Obj_free( &c_square ); PLA_Obj_free( &c_mscalar);
  PLA_Obj_free( &Zero );
  PLA_Obj_free( &ccur_rest_cols); PLA_Obj_free( &ccur_rest_rows);
  PLA_Obj_free( &acur_rest_rows); PLA_Obj_free( &bcur_rest_cols);
}

/*************  END OF TRANPOSE A / TRANSPOSE B SECTION **********************/
/************** BEGIN  NO TRANSPOSE A / TRANSPOSE B SECTION ******************/
/* Handles the cases for No Transpose A / Transpose B for :
	9.1.6.1
	9.1.6.3
	9.1.6.4
	9.1.6.6
	9.1.6.7
	9.1.6.8
*/
int PLA_Gemm_notransa_transb_1(int nb_alg, int extra_ta, int extra_tb, PLA_Obj alpha, PLA_Obj a,
										  PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
  PLA_Obj  bcur = NULL, ccur = NULL,
			  b1t = NULL,   c1 = NULL,   Zero = NULL,
			  b2t = NULL,   c2 = NULL,   b1tilde = NULL, c1tilde = NULL;
  int size, length_b, width_c;
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;
  int           transa, transb;
									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Obj_datatype(c, &datatype);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &Zero );
  PLA_Obj_set_to_zero(Zero);
  transa = extra_ta;
  transb = extra_tb;
									  /* take a view of all of both A and B */
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &bcur );
  PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &ccur );
  PLA_Pmvector_create_conf_to( a, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
											nb_alg, &b1tilde );
  PLA_Pmvector_create_conf_to( a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
											nb_alg, &c1tilde );
  PLA_Obj_set_to_zero(c1tilde);
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
									  /* determine width of next update     */
	 PLA_Obj_global_length( bcur, &length_b );
	 PLA_Obj_global_width ( ccur, &width_c );
	 if ( ( size = min3( length_b, width_c, nb_alg )) == 0 ) break;
	 PLA_Obj_horz_split_2 ( bcur, size,              &b1t, &b2t );
	 PLA_Obj_vert_split_2 ( ccur, size,              &c1, &c2 );
									  /* size the workspaces                */
	 PLA_Obj_view_shift(b1tilde,           nb_alg - size,
													0,               0,
																  0);
	 PLA_Obj_view_shift(c1tilde,                   0,
										  nb_alg - size,          0,
																  0);
									  /* annotate the views                 */
	 PLA_Obj_set_orientation( b1t, PLA_PROJ_ONTO_ROW );
	 PLA_Obj_set_orientation( c1,  PLA_PROJ_ONTO_COL );

	 PLA_Copy( b1t, b1tilde);
									  /* perform local matrix-panel update     */
	 PLA_Local_gemm(transa, transb, alpha, a, b1tilde, Zero, c1tilde);
	 PLA_Reduce_add(c1tilde, PLASUM(datatype), c1);
	 PLA_Obj_swap( &b2t, &bcur );
	 PLA_Obj_swap( &c2,  &ccur );
	 PLA_Obj_view_shift(b1tilde,           size - nb_alg,
													0,               0,
																  0);
	 PLA_Obj_view_shift(c1tilde,                   0,
										  size - nb_alg,          0,
																  0);
  }
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1t ); PLA_Obj_free( &b2t );
  PLA_Obj_free( &ccur );  PLA_Obj_free( &c1 );  PLA_Obj_free( &c2 );
  PLA_Obj_free( &b1tilde );  PLA_Obj_free( &c1tilde );  PLA_Obj_free( &Zero);

}
/*****************************************************************************/
/* Handles the cases for No Transpose A / Transpose B for :
	9.1.6.2
*/
int PLA_Gemm_notransa_transb_2(int nb_alg, int extra_ta, int extra_tb, PLA_Obj alpha, PLA_Obj a,
											PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
  PLA_Obj  acur = NULL, bcur = NULL,
			  a1 = NULL,   b1 = NULL,   One = NULL,
			  a2 = NULL,   b2 = NULL,   a1tilde = NULL, b1tilde = NULL;
  int size, width_a, width_b;
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;
  int           transa, transb;
									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Obj_datatype(c, &datatype);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &One );
  PLA_Obj_set_to_one(One);
  transa = extra_ta;
  transb = PLA_NO_TRANSPOSE;
  if( (MPI_COMPLEX == datatype) || (MPI_DOUBLE_COMPLEX == datatype) )
  {
	  if( PLA_CONJ_TRANSPOSE == extra_tb)
		  transb = PLA_CONJ;
  }

									  /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &bcur );
									  /* create dupl. pmvectors for spreading
										  panels of A and B */
  PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
											nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
											nb_alg, &b1tilde );
									  /* Loop until no more of A and B      */ 
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
									  /* determine width of next update     */
	 PLA_Obj_global_width( acur, &width_a );
	 PLA_Obj_global_width ( bcur, &width_b );
	 if ( ( size = min3( width_a, width_b, nb_alg )) == 0 ) break;
									  /* split off first column of Acur     */
	 PLA_Obj_vert_split_2 ( acur, size,              &a1, &a2 );
									  /* split off first row of Bcur        */
	 PLA_Obj_vert_split_2 ( bcur, size,              &b1, &b2 );
									  /* size the workspaces                */
	 PLA_Obj_view_shift(a1tilde,                   0,
										  nb_alg - size,          0,
																  0);
	 PLA_Obj_view_shift(b1tilde,           nb_alg - size,
													0,               0,
																  0);
									  /* annotate the views                 */
	 PLA_Obj_set_orientation( a1,   PLA_PROJ_ONTO_COL );
	 PLA_Obj_set_orientation( b1,   PLA_PROJ_ONTO_COL );
									  /* spread a1 and a2                   */
	 PLA_Copy( a1,  a1tilde);
	 PLA_Copy( b1,  b1tilde);
									  /* perform local rank-size update     */
	 PLA_Local_gemm(transa, transb, alpha, a1tilde, b1tilde, One, c);
	 PLA_Obj_swap( &a2,  &acur );
	 PLA_Obj_swap( &b2,  &bcur );
	 PLA_Obj_view_shift(a1tilde,                   0,
										  size - nb_alg,          0,
																  0);
	 PLA_Obj_view_shift(b1tilde,           size - nb_alg,
													0,               0,
																  0);
  }
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1 );  PLA_Obj_free( &a2 );
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1 );  PLA_Obj_free( &b2 );
  PLA_Obj_free( &b1tilde );  PLA_Obj_free( &a1tilde ); PLA_Obj_free( &One);
}
/*****************************************************************************/
/* Handles the cases for No Transpose A / Transpose B for :
	9.1.6.5
*/
int PLA_Gemm_notransa_transb_3(int nb_alg, int extra_ta, int extra_tb, PLA_Obj alpha, PLA_Obj a,
											PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
  PLA_Obj  acur = NULL, ccur = NULL,
			  a1t = NULL,   c1t = NULL,   Zero = NULL,
			  a2t = NULL,   c2t = NULL,   a1tilde = NULL, c1tilde = NULL;
  int size, length_a, length_c;
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;
  int           transa = extra_ta, transb = extra_tb;

									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Obj_datatype(c, &datatype);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &Zero );
  PLA_Obj_set_to_zero(Zero);

									  /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &ccur );
									  /* create dupl. pmvectors for spreading
										  panels of A and B */
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
											nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
											nb_alg, &c1tilde );
  PLA_Obj_set_to_zero(c1tilde);
									  /* Loop until no more of A and B      */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
									  /* determine width of next update     */
	 PLA_Obj_global_length  ( acur, &length_a );
	 PLA_Obj_global_length  ( ccur, &length_c );
	 if ( ( size = min3( length_a, length_c, nb_alg )) == 0 ) break;
									  /* split off first column of Acur     */
	 PLA_Obj_horz_split_2 ( acur, size,              &a1t, &a2t );
									  /* split off first row of Bcur        */
	 PLA_Obj_horz_split_2 ( ccur, size,              &c1t, &c2t );
									  /* size the workspaces                */
	 PLA_Obj_view_shift(a1tilde,           nb_alg - size,
													0,               0,
																  0);
	 PLA_Obj_view_shift(c1tilde,                   0,
										  nb_alg - size,          0,
																  0);
									  /* annotate the views                 */
	 PLA_Obj_set_orientation( a1t,    PLA_PROJ_ONTO_ROW );
	 PLA_Obj_set_orientation( c1t,    PLA_PROJ_ONTO_ROW );
									  /* spread a1 and a2                   */
	 PLA_Copy( a1t,  a1tilde);
									  /* perform local rank-size update     */
	 PLA_Local_gemm(transa, transb , alpha, b, a1tilde, Zero, c1tilde);
	 PLA_Reduce_add( c1tilde, PLASUM(datatype), c1t);
	 PLA_Obj_swap( &a2t,  &acur );
	 PLA_Obj_swap( &c2t,  &ccur );
	 PLA_Obj_view_shift(a1tilde,           size - nb_alg,
													0,               0,
																  0);
	 PLA_Obj_view_shift(c1tilde,                   0,
										  size - nb_alg,          0,
																  0);
  }
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1t );  PLA_Obj_free( &a2t );
  PLA_Obj_free( &ccur );  PLA_Obj_free( &c1t );  PLA_Obj_free( &c2t );
  PLA_Obj_free( &a1tilde ); PLA_Obj_free( &c1tilde ); PLA_Obj_free( &Zero);
}

/* Small m, small n, large k */
int PLA_Gemm_notransa_transb_4(int nb_alg, int extra_ta, int extra_tb, PLA_Obj alpha, PLA_Obj a,
											  PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;
  MPI_Comm      dummy_comm;
  int nb, num_rows, idummy;

  int m, n, size_a, size_b;

  int zero_or_one, msub, nsub, length_mvector, length_c, width_c;

  PLA_Obj  acur = NULL, a_mvector = NULL, a_mvector_cur = NULL,
			  acur_row = NULL, acur_rest_rows = NULL,
			  bcur = NULL, b_mvector = NULL, b_mvector_cur = NULL,
			  bcur_col = NULL, bcur_rest_cols = NULL,
			  ccur = NULL, c_mscalar_cur_outer = NULL, c_mscalar_cur_inner = NULL,
			  ccur_row = NULL, c_square = NULL, c_mscalar = NULL,
			  ccur_rest_rows = NULL, ccur_rest_cols = NULL,
			  Zero = NULL;
  int      transa, transb;

									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Temp_zero_or_one(template, &zero_or_one);
  PLA_Temp_nb( template, &nb);
  PLA_Temp_comm_row_info(template, &dummy_comm, &idummy, &num_rows);
  /* since we are going ONLY to multivectors we can use a bit more storage */
  if(nb_alg < (nb * num_rows)) nb_alg = nb * num_rows;
  PLA_Obj_datatype(c, &datatype);

  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template,
							 &Zero );
  PLA_Obj_set_to_zero(Zero);
  transa = PLA_TRANSPOSE;
  transb = PLA_NO_TRANSPOSE;
  if( (MPI_COMPLEX == datatype) || (MPI_DOUBLE_COMPLEX == datatype) )
  {
	  if( PLA_CONJ_TRANSPOSE == extra_tb)
		  transb = PLA_CONJ;
  }
									  /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
							 PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,&bcur );
  PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,&ccur );
  PLA_Obj_global_width(a, &length_mvector);
  PLA_Obj_global_length(c, &m);
  PLA_Obj_global_width(c, &n);
  size_a = min(nb_alg, m);
  size_b = min(nb_alg, n);
/* create mvectors for spreading panels of A and B  */
  PLA_Mvector_create( datatype, length_mvector, size_a, template, zero_or_one,
							 &a_mvector);
  PLA_Mvector_create( datatype, length_mvector, size_b, template, zero_or_one,
							 &b_mvector);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, size_a, size_b,
							 template, &c_mscalar);

									  /* Loop until no more of A and B      */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
									  /* determine width of next update     */
	 PLA_Obj_global_length ( ccur, &length_c );
	 if ( ( msub = min( length_c, size_a )) == 0 ) break;
	 PLA_Obj_horz_split_2(ccur, msub, &ccur_row, &ccur_rest_rows);
	 PLA_Obj_horz_split_2(acur, msub, &acur_row, &acur_rest_rows);
	 PLA_Obj_set_orientation(acur_row, PLA_PROJ_ONTO_ROW);
	 if(msub < size_a)
	 {
		 PLA_Obj_vert_split_2 ( a_mvector, msub,          &a_mvector_cur, PLA_DUMMY );
		 PLA_Obj_swap(&a_mvector, &a_mvector_cur);
		 PLA_Obj_horz_split_2 ( c_mscalar, msub,          &c_mscalar_cur_outer, PLA_DUMMY );
		 PLA_Obj_swap(&c_mscalar_cur_outer, &c_mscalar);
	 }
	 PLA_Copy(acur_row, a_mvector);
	 PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,&bcur );
	 while(TRUE)
	 {
		 PLA_Obj_global_width(ccur_row, &width_c);
		 if( (nsub = min( width_c, size_b)) == 0) break;
		 PLA_Obj_vert_split_2(ccur_row, nsub, &c_square, &ccur_rest_cols);
		 PLA_Obj_horz_split_2(bcur, nsub, &bcur_col, &bcur_rest_cols);
		 PLA_Obj_set_orientation(bcur_col, PLA_PROJ_ONTO_ROW);
		 if(nsub < size_b)
		 {
			 PLA_Obj_vert_split_2 ( b_mvector, nsub,          &b_mvector_cur, PLA_DUMMY );
			 PLA_Obj_view_swap(&b_mvector, &b_mvector_cur);
			 PLA_Obj_vert_split_2 ( c_mscalar, msub,          &c_mscalar_cur_inner, PLA_DUMMY );
			 PLA_Obj_view_swap(&c_mscalar, &c_mscalar_cur_inner);
		 }
		 PLA_Copy(bcur_col, b_mvector);
		 PLA_Local_gemm(PLA_TRANS, PLA_NO_TRANS, alpha, a_mvector, b_mvector, Zero, c_mscalar);
		 PLA_Reduce_dup_mscalar_to_all(c_mscalar, PLASUM(datatype));
		 PLA_Reduce_dup_mscalar_to_matrix(c_mscalar, PLASUM(datatype), c_square);
		 if(nsub < size_b)
		 {
			PLA_Obj_view_swap(&c_mscalar, &c_mscalar_cur_inner);
			PLA_Obj_view_swap(&b_mvector, &b_mvector_cur);
		 }
		 PLA_Obj_view_swap( &ccur_row, &ccur_rest_cols);
		 PLA_Obj_view_swap( &bcur,     &bcur_rest_cols);
	 } /* End inner loop */
	 PLA_Obj_view_swap( &ccur, &ccur_rest_rows);
	 PLA_Obj_view_swap( &acur, &acur_rest_rows);
	 if(msub < size_a)
	 {
		  PLA_Obj_view_swap(&c_mscalar, &c_mscalar_cur_outer);
		  PLA_Obj_view_swap(&a_mvector, &a_mvector_cur);
	 }
  } /* End outer loop */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
  PLA_Obj_free( &acur); PLA_Obj_free( &a_mvector); PLA_Obj_free( &a_mvector_cur );
  PLA_Obj_free( &acur_row ); PLA_Obj_free( &bcur ); PLA_Obj_free( & b_mvector );
  PLA_Obj_free( &b_mvector_cur); PLA_Obj_free( &bcur_col ); PLA_Obj_free( &ccur );
  PLA_Obj_free( &c_mscalar_cur_outer ); PLA_Obj_free( &c_mscalar_cur_inner );
  PLA_Obj_free( &ccur_row ); PLA_Obj_free( &c_square ); PLA_Obj_free( &c_mscalar);
  PLA_Obj_free( &Zero );
  PLA_Obj_free( &ccur_rest_cols); PLA_Obj_free( &ccur_rest_rows);
  PLA_Obj_free( &acur_rest_rows); PLA_Obj_free( &bcur_rest_cols);
}

/************** END  NO TRANSPOSE A / TRANSPOSE B SECTION ******************/
/************** BEGIN  TRANSPOSE A / NO TRANSPOSE B SECTION ******************/
/* Handles the cases for Transpose A / No Transpose B for :
	9.1.6.1
	9.1.6.4
	9.1.6.5
	9.1.6.6
	9.1.6.7
	9.1.6.8
*/
int PLA_Gemm_transa_notransb_1(int nb_alg, int extra_ta, int extra_tb, PLA_Obj alpha, PLA_Obj a,
											PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
  PLA_Obj  acur = NULL, ccur = NULL,
			  a1 = NULL,   c1t = NULL,   Zero = NULL,
			  a2 = NULL,   c2t = NULL,   a1tilde = NULL, c1tilde = NULL;
  int size, width_a, length_c;
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;
  int           transa, transb;
									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Obj_datatype(c, &datatype);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &Zero );
  PLA_Obj_set_to_zero(Zero);

  transa = extra_ta;
  transb = extra_tb;

									  /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &ccur );
  /* create dupl. pmvectors for spreading panels of A and B */
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
											nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( b, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
											nb_alg, &c1tilde );
  PLA_Obj_set_to_zero(c1tilde);
									  /* Loop until no more of A and B      */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
									  /* determine width of next update     */
	 PLA_Obj_global_width ( acur, &width_a );
	 PLA_Obj_global_length( ccur, &length_c );
	 if ( ( size = min3( width_a, length_c, nb_alg )) == 0 ) break;
									  /* split off first column of Acur     */
	 PLA_Obj_vert_split_2 ( acur, size,              &a1, &a2 );
									  /* split off first row of Bcur        */
	 PLA_Obj_horz_split_2 ( ccur, size,              &c1t, &c2t );
									  /* size the workspaces                */
	 PLA_Obj_view_shift(a1tilde,                   0,
										  nb_alg - size,          0,
																  0);
	 PLA_Obj_view_shift(c1tilde,           nb_alg - size,
													0,               0,
																  0);
									  /* annotate the views                 */
	 PLA_Obj_set_orientation( a1,   PLA_PROJ_ONTO_COL );
	 PLA_Obj_set_orientation( c1t,  PLA_PROJ_ONTO_ROW );
									  /* spread a1 and a2                   */
	 PLA_Copy( a1,  a1tilde);
									  /* perform local rank-size update     */
	 PLA_Local_gemm(transa, transb, alpha, a1tilde, b, Zero, c1tilde);
	 PLA_Reduce_add(c1tilde, PLASUM(datatype), c1t);
	 PLA_Obj_swap( &a2,  &acur );
	 PLA_Obj_swap( &c2t, &ccur );
	 PLA_Obj_view_shift(a1tilde,                   0,
										  size - nb_alg,          0,
																  0);
	 PLA_Obj_view_shift(c1tilde,           size - nb_alg,
													0,               0,
																  0);
  }
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1 );  PLA_Obj_free( &a2 );
  PLA_Obj_free( &ccur );  PLA_Obj_free( &c1t ); PLA_Obj_free( &c2t );
  PLA_Obj_free( &a1tilde );  PLA_Obj_free( &c1tilde );  PLA_Obj_free( &Zero);
}
/*****************************************************************************/
/* Handles the cases for Transpose A / No Transpose B for :
	9.1.6.2
*/
int PLA_Gemm_transa_notransb_2(int nb_alg, int extra_ta, int extra_tb, PLA_Obj alpha, PLA_Obj a,
											PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
  PLA_Obj acur = NULL, bcur = NULL,
			  a1t = NULL,   b1t = NULL,   One = NULL,
			  a2t = NULL,   b2t = NULL,   a1tilde = NULL, b1tilde = NULL;
  int size, length_a, length_b;
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;
  int           transa, transb;

									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Obj_datatype(c, &datatype);
  PLA_Mscalar_create(datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &One );
  PLA_Obj_set_to_one(One);
  transa = PLA_NO_TRANSPOSE;
  transb = extra_tb;
  if( (MPI_COMPLEX == datatype) || (MPI_DOUBLE_COMPLEX == datatype) )
  {
	  if( PLA_CONJ_TRANSPOSE == extra_ta)
		  transa = PLA_CONJ;
  }
									  /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &bcur );
  PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
											nb_alg, &a1tilde );
  PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
											nb_alg, &b1tilde );
									  /* Loop until no more of A and B      */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
									  /* determine width of next update     */
	 PLA_Obj_global_length ( acur, &length_a );
	 PLA_Obj_global_length ( bcur, &length_b );
	 if ( ( size = min3( length_a, length_b, nb_alg )) == 0 ) break;
									  /* split off first column of Acur     */
	 PLA_Obj_horz_split_2 ( acur, size,              &a1t, &a2t );
									  /* split off first row of Bcur        */
	 PLA_Obj_horz_split_2 ( bcur, size,              &b1t, &b2t );
									  /* size the workspaces                */
	 PLA_Obj_view_shift(a1tilde,                   0,
										  nb_alg - size,          0,
																  0);
	 PLA_Obj_view_shift(b1tilde,           nb_alg - size,
													0,               0,
																  0);
									  /* annotate the views                 */
	 PLA_Obj_set_orientation( a1t,   PLA_PROJ_ONTO_ROW );
	 PLA_Obj_set_orientation( b1t,   PLA_PROJ_ONTO_ROW );
									  /* spread a1 and a2                   */
	 PLA_Copy( a1t,  a1tilde);
	 PLA_Copy( b1t,  b1tilde);
									  /* perform local rank-size update     */
	 PLA_Local_gemm(PLA_NO_TRANS,PLA_NO_TRANS, alpha, a1tilde, b1tilde,
														 One, c);
	 PLA_Obj_swap( &a2t, &acur );
	 PLA_Obj_swap( &b2t, &bcur );
	 PLA_Obj_view_shift(a1tilde,                   0,
										  size - nb_alg,          0,
																  0);
	 PLA_Obj_view_shift(b1tilde,           size - nb_alg,
													0,               0,
																  0);
  }
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
  PLA_Obj_free( &acur );  PLA_Obj_free( &a1t );  PLA_Obj_free( &a2t );
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1t );  PLA_Obj_free( &b2t );
  PLA_Obj_free( &b1tilde );  PLA_Obj_free( &a1tilde );  PLA_Obj_free( &One);
}
/*****************************************************************************/
/* Handles the cases for Transpose A / No Transpose B for :
	9.1.6.3
*/
int PLA_Gemm_transa_notransb_3(int nb_alg, int extra_ta, int extra_tb, PLA_Obj alpha, PLA_Obj a,
											PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
  PLA_Obj bcur = NULL, ccur = NULL,
			  b1 = NULL,   c1 = NULL,   Zero = NULL,
			  b2 = NULL,   c2 = NULL,   b1tilde = NULL, c1tilde = NULL;
  int size, width_b, width_c;
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;
  int  transa, transb;
									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Obj_datatype(c, &datatype);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &Zero );
  PLA_Obj_set_to_zero(Zero);
  transa = extra_ta;
  transb = extra_tb;
									  /* take a view of all of both A and B */
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &bcur );
  PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL,
							  PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &ccur );
									  /* create dupl. pmvectors for spreading
										  panels of A and B */
  PLA_Pmvector_create_conf_to( a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
											nb_alg, &b1tilde );
  PLA_Pmvector_create_conf_to( a, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
											nb_alg, &c1tilde );
  PLA_Obj_set_to_zero(c1tilde);
									  /* Loop until no more of A and B      */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
									  /* determine width of next update     */
	 PLA_Obj_global_width ( bcur,  &width_b );
	 PLA_Obj_global_width  ( ccur, &width_c );
	 if ( ( size = min3( width_b, width_c, nb_alg )) == 0 ) break;
									  /* split off first column of Acur     */
	 PLA_Obj_vert_split_2 ( bcur, size,              &b1, &b2 );
									  /* split off first row of Bcur        */
	 PLA_Obj_vert_split_2 ( ccur, size,              &c1, &c2 );
									  /* size the workspaces                */
	 PLA_Obj_view_shift(b1tilde,                   0,
										  nb_alg - size,          0,
																  0);
	 PLA_Obj_view_shift(c1tilde,           nb_alg - size,
													0,               0,
																  0);
									  /* annotate the views                 */
	 PLA_Obj_set_orientation( b1,    PLA_PROJ_ONTO_COL );
	 PLA_Obj_set_orientation( c1,    PLA_PROJ_ONTO_COL );
									  /* spread a1 and a2                   */
	 PLA_Copy( b1,  b1tilde);
									  /* perform local rank-size update     */
	 PLA_Local_gemm( transa, transb, alpha,  b1tilde, a, Zero, c1tilde);
	 PLA_Reduce_add( c1tilde, PLASUM(datatype), c1);
	 PLA_Obj_swap( &b2,  &bcur );
	 PLA_Obj_swap( &c2,  &ccur );
	 PLA_Obj_view_shift(b1tilde,                   0,
										  size - nb_alg,          0,
																  0);
	 PLA_Obj_view_shift(c1tilde,           size - nb_alg,
													0,               0,
																  0);
  }
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
									  /* Free the views                     */
  PLA_Obj_free( &bcur );  PLA_Obj_free( &b1 );  PLA_Obj_free( &b2 );
  PLA_Obj_free( &ccur );  PLA_Obj_free( &c1 );  PLA_Obj_free( &c2 );
  PLA_Obj_free( &b1tilde );  PLA_Obj_free( &c1tilde );  PLA_Obj_free( &Zero);
}

/* Small m, small n, large k */
int PLA_Gemm_transa_notransb_4(int nb_alg, int extra_ta, int extra_tb, PLA_Obj alpha, PLA_Obj a,
											  PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
  PLA_Template  template = NULL;
  MPI_Datatype  datatype;
  extern  MPI_Op PLA_SUM;
  MPI_Comm      dummy_comm;
  int nb, num_rows, idummy;

  int m, n, size_a, size_b;

  int zero_or_one, msub, nsub, length_mvector, length_c, width_c;

  PLA_Obj  acur = NULL, a_mvector = NULL, a_mvector_cur = NULL,
			  acur_row = NULL, acur_rest_rows = NULL,
			  bcur = NULL, b_mvector = NULL, b_mvector_cur = NULL,
			  bcur_col = NULL, bcur_rest_cols = NULL,
			  ccur = NULL, c_mscalar_cur_outer = NULL, c_mscalar_cur_inner = NULL,
			  ccur_row = NULL, c_square = NULL, c_mscalar = NULL,
			  ccur_rest_rows = NULL, ccur_rest_cols = NULL,
			  Zero = NULL;
  int      transa, transb;

									  /* scale C by beta                    */
  PLA_Scal( beta, c );
									  /* create constant mscalar 1          */
  PLA_Obj_template(c, &template);
  PLA_Temp_zero_or_one(template, &zero_or_one);
  PLA_Temp_nb( template, &nb);
  PLA_Temp_comm_row_info(template, &dummy_comm, &idummy, &num_rows);
  /* since we are going ONLY to multivectors we can use a bit more storage */
  if(nb_alg < (nb * num_rows)) nb_alg = nb * num_rows;
  PLA_Obj_datatype(c, &datatype);
  transa = extra_ta;
  transb = extra_tb;

  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template,
							 &Zero );
  PLA_Obj_set_to_zero(Zero);
									  /* take a view of all of both A and B */
  PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL,
							 PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &acur );
  PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,&bcur );
  PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,&ccur );
  PLA_Obj_global_length(a, &length_mvector);
  PLA_Obj_global_length(c, &m);
  PLA_Obj_global_width(c, &n);
  size_a = min(nb_alg, m);
  size_b = min(nb_alg, n);
/* create mvectors for spreading panels of A and B  */
  PLA_Mvector_create( datatype, length_mvector, size_a, template, zero_or_one,
                      &a_mvector);
  PLA_Mvector_create( datatype, length_mvector, size_b, template, zero_or_one,
                      &b_mvector);
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, size_a, size_b,
                      template, &c_mscalar);

                             /* Loop until no more of A and B      */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
  while ( TRUE ) {
                             /* determine width of next update     */
    PLA_Obj_global_length ( ccur, &length_c );
    if ( ( msub = min( length_c, size_a )) == 0 ) break;
    PLA_Obj_horz_split_2(ccur, msub, &ccur_row, &ccur_rest_rows);
    PLA_Obj_vert_split_2(acur, msub, &acur_row, &acur_rest_rows);
    PLA_Obj_set_orientation(acur_row, PLA_PROJ_ONTO_COL);
    if(msub < size_a)
    {
       PLA_Obj_vert_split_2 ( a_mvector, msub,          &a_mvector_cur, PLA_DUMMY );
       PLA_Obj_swap(&a_mvector, &a_mvector_cur);
       PLA_Obj_horz_split_2 ( c_mscalar, msub,          &c_mscalar_cur_outer, PLA_DUMMY );
       PLA_Obj_swap(&c_mscalar_cur_outer, &c_mscalar);
    }
    PLA_Copy(acur_row, a_mvector);
    PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL,
                       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,&bcur );
    while(TRUE)
    {
    	 PLA_Obj_global_width(ccur_row, &width_c);
       if( (nsub = min( width_c, size_b)) == 0) break;
       PLA_Obj_vert_split_2(ccur_row, nsub, &c_square, &ccur_rest_cols);
       PLA_Obj_vert_split_2(bcur, nsub, &bcur_col, &bcur_rest_cols);
       PLA_Obj_set_orientation(bcur_col, PLA_PROJ_ONTO_COL);
       if(nsub < size_b)
       {
          PLA_Obj_vert_split_2 ( b_mvector, nsub,          &b_mvector_cur, PLA_DUMMY );
          PLA_Obj_view_swap(&b_mvector, &b_mvector_cur);
          PLA_Obj_vert_split_2 ( c_mscalar, msub,          &c_mscalar_cur_inner, PLA_DUMMY );
          PLA_Obj_view_swap(&c_mscalar, &c_mscalar_cur_inner);
       }
       PLA_Copy(bcur_col, b_mvector);
       PLA_Local_gemm(PLA_TRANS, PLA_NO_TRANS, alpha, a_mvector, b_mvector, Zero, c_mscalar);
       PLA_Reduce_dup_mscalar_to_all(c_mscalar, PLASUM(datatype));
       PLA_Reduce_dup_mscalar_to_matrix(c_mscalar, PLASUM(datatype), c_square);
       if(nsub < size_b)
       {
       	PLA_Obj_view_swap(&c_mscalar, &c_mscalar_cur_inner);
         PLA_Obj_view_swap(&b_mvector, &b_mvector_cur);
       }
       PLA_Obj_view_swap( &ccur_row, &ccur_rest_cols);
       PLA_Obj_view_swap( &bcur,     &bcur_rest_cols);
    } /* End inner loop */
    PLA_Obj_view_swap( &ccur, &ccur_rest_rows);
    PLA_Obj_view_swap( &acur, &acur_rest_rows);
    if(msub < size_a)
    {
  	     PLA_Obj_view_swap(&c_mscalar, &c_mscalar_cur_outer);
        PLA_Obj_view_swap(&a_mvector, &a_mvector_cur);
    }
  } /* End outer loop */
  PLA_Set_Flow(SYNCHRONOUS, PLA_SIDE_RIGHT, PLA_DOWN);
                             /* Free the views                     */
  PLA_Obj_free( &acur); PLA_Obj_free( &a_mvector); PLA_Obj_free( &a_mvector_cur );
  PLA_Obj_free( &acur_row ); PLA_Obj_free( &bcur ); PLA_Obj_free( & b_mvector );
  PLA_Obj_free( &b_mvector_cur); PLA_Obj_free( &bcur_col ); PLA_Obj_free( &ccur );
  PLA_Obj_free( &c_mscalar_cur_outer ); PLA_Obj_free( &c_mscalar_cur_inner );
  PLA_Obj_free( &ccur_row ); PLA_Obj_free( &c_square ); PLA_Obj_free( &c_mscalar);
  PLA_Obj_free( &Zero );
  PLA_Obj_free( &ccur_rest_cols); PLA_Obj_free( &ccur_rest_rows);
  PLA_Obj_free( &acur_rest_rows); PLA_Obj_free( &bcur_rest_cols);
}


