#include "PLA.h"

int PLA_Local_syr2k(
   int uplo,
   int trans,
   PLA_Obj alpha,
   PLA_Obj a,
   PLA_Obj b,
   PLA_Obj beta,
   PLA_Obj c)
{
   char uplo_char[1], trans_char[1];
   void *alpha_buffer, *a_buffer, *b_buffer, *beta_buffer, *c_buffer;
   int  n, k, lda, ldb, ldc;
   MPI_Datatype datatype;

   PLA_Local_syr2k_enter( uplo, trans, alpha, a, b, beta, c);

   PLA_Obj_datatype( a, &datatype);

   if(PLA_UPPER_TRIANGULAR == uplo)
      uplo_char[0] = 'U';
   else uplo_char[0] = 'L';

   if(PLA_NO_TRANSPOSE == trans)
      trans_char[0] = 'N';
   else if(PLA_TRANSPOSE == trans)
      trans_char[0] = 'T';
   else trans_char[0] = 'C';

   if(PLA_NO_TRANSPOSE == trans)
   {
      PLA_Obj_local_length( a, &n );
      PLA_Obj_local_width ( a, &k );
   }
   else
   {
      PLA_Obj_local_length( a, &k );
      PLA_Obj_local_width ( a, &n );
   }

   PLA_Obj_local_buffer( alpha, &alpha_buffer );
   PLA_Obj_local_buffer( a,     &a_buffer );
   PLA_Obj_local_buffer( b,     &b_buffer );
   PLA_Obj_local_buffer( beta,  &beta_buffer );
   PLA_Obj_local_buffer( c,     &c_buffer );

   PLA_Obj_local_ldim( a, &lda );
   PLA_Obj_local_ldim( b, &ldb );
   PLA_Obj_local_ldim( c, &ldc );


   if( (n*k) > 0)
   {
      if(MPI_DOUBLE == datatype)
         PLA_dsyr2k( uplo_char, trans_char, &n, &k,
                       alpha_buffer, a_buffer, &lda, b_buffer, &ldb,
                       beta_buffer, c_buffer, &ldc);
      if(MPI_FLOAT == datatype)
         PLA_ssyr2k( uplo_char, trans_char, &n, &k,
                       alpha_buffer, a_buffer, &lda, b_buffer, &ldb,
                       beta_buffer, c_buffer, &ldc);
/*    if(MPI_COMPLEX == datatype)
         PLA_csyr2k( uplo_char, trans_char, &n, &k,
                       alpha_buffer, a_buffer, &lda, b_buffer, &ldb,
                       beta_buffer, c_buffer, &ldc);
      if(MPI_DOUBLE_COMPLEX == datatype)
         PLA_zsyr2k( uplo_char, trans_char, &n, &k,
                       alpha_buffer, a_buffer, &lda, b_buffer, &ldb,
                       beta_buffer, c_buffer, &ldc); */
   }

   PLA_Local_syr2k_exit ( uplo, trans, alpha, a, b, beta, c);
   return(PLA_SUCCESS);
}
/******************************************************************************/
int pla_trap_scal(PLA_Obj alpha, PLA_Obj a, int upper_or_lower)
{
  PLA_Obj a_remainder = NULL, a_strip = NULL;
  int     size;

  PLA_Obj_view_all ( a, & a_remainder );
  PLA_Obj_global_length ( a_remainder, & size );
  if(PLA_LOWER_TRIANGULAR == upper_or_lower)
     for( ; ; )
     {
        PLA_Obj_global_length ( a_remainder, & size );
        if(size == 0) break;
        PLA_Obj_vert_split_2 ( a_remainder, 1, & a_strip, & a_remainder );
        PLA_Scal ( alpha, a_strip );
        PLA_Obj_view_shift ( a_remainder,    1,
                                         0,      0,
                                             0 );
        PLA_Obj_global_length ( a_remainder, & size );
     }
   else /* PLA_UPPER_TRIANGULAR -- note, internal routine "pla" - no error checking */
     for( ; ; )
     {
        PLA_Obj_global_length ( a_remainder, & size );
        if(size == 0) break;
        PLA_Obj_vert_split_2 ( a_remainder, -1, & a_remainder, & a_strip );
        PLA_Scal ( alpha, a_strip );
        PLA_Obj_view_shift ( a_remainder,    0,
                                          0,     0,
                                            -1 );
     }
  PLA_Obj_free( &a_remainder);
  PLA_Obj_free( &a_strip );
  return(PLA_SUCCESS); /* hmm, pla_success? */
}
/******************************************************************************/
int pla_syr2k_ln_rankk(PLA_Obj alpha, PLA_Obj a, PLA_Obj b, PLA_Obj beta, PLA_Obj c)
{
  int     nb_alg, top_size, top_owner, size ;
  int     a_left_size, a_left_owner, b_left_size, b_left_owner;
  int     left_size, left_owner, my_row, my_col;
  MPI_Datatype datatype;
  PLA_Template template;
  PLA_Obj      a_left = NULL,    a_right = NULL, one = NULL,
               b_left = NULL,    b_right = NULL,
               c_topleft = NULL, c_bottomleft = NULL, c_bottomright = NULL,
               work_col = NULL,  work_col_temp = NULL,
               a_col_top = NULL,  a_col_bottom = NULL,
               b_col_top = NULL,  b_col_bottom = NULL,
               work_row = NULL,  work_row_temp = NULL,
               at_row_left = NULL, at_row_right = NULL,
               bt_row_left = NULL, bt_row_right = NULL;

  PLA_Obj_template ( a, & template );
  PLA_Environ_nb_alg ( PLA_OP_PAN_PAN, template, & nb_alg );
  PLA_Temp_comm_row_rank( template, &my_col);
  PLA_Temp_comm_col_rank( template, &my_row);

  pla_trap_scal(beta, c, PLA_LOWER_TRIANGULAR);


  PLA_Obj_datatype ( c, & datatype );
  PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, & one );
  PLA_Obj_set_to_one ( one );

/* 2*nb alg -- okay, maybe this isn't too smart but we are taking a pretty */
/*             big hit in doing SO many gemm updates -- I'm trying to      */
/*             minimize that here by stacking A|B and B^T                  */
/*                                                    A^T  panels          */
  PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb_alg*2, & work_row );
  PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb_alg*2, & work_col );

  PLA_Obj_view_all ( a, & a_right );
  PLA_Obj_view_all ( b, & b_right );

  for ( ; ; )
  {
/*    PLA_Obj_split_size ( a_right, PLA_SIDE_LEFT, & a_left_size, & a_left_owner );
    PLA_Obj_split_size ( b_right, PLA_SIDE_LEFT, & b_left_size, & b_left_owner ); */
    PLA_Obj_global_width(a_right, &a_left_size);
    PLA_Obj_global_width(b_right, &b_left_size);

    if ( (size = min3( nb_alg, a_left_size, b_left_size)) == 0) break;
    PLA_Obj_vert_split_2 ( a_right, size, & a_left, & a_right );
    PLA_Obj_vert_split_2 ( b_right, size, & b_left, & b_right );

    PLA_Obj_vert_split_2 ( work_col,       size, & a_col_bottom, &work_col_temp );
    PLA_Obj_vert_split_2 ( work_col_temp, size , & b_col_bottom, PLA_DUMMY);

    PLA_Obj_horz_split_2 ( work_row,       size, & bt_row_right, &work_row_temp );
    PLA_Obj_horz_split_2 ( work_row_temp,  size, & at_row_right, PLA_DUMMY);

    PLA_Obj_set_orientation ( a_left, PLA_PROJ_ONTO_COL );
    PLA_Obj_set_orientation ( b_left, PLA_PROJ_ONTO_COL );

/* The order of the copies is unimportant -- interlaced for programmer sanity */
/* If we strictly used an mvector intermediate this could be sped up by using */
/* a temp_mvector for both a and b -- however, since spreads in the oriented  */
/* direction employ a "short-circuting shortcut" I don't think that there is  */
/* much point.                                                                */

    PLA_Copy ( a_left, a_col_bottom );
    PLA_Copy ( b_left, b_col_bottom );

    PLA_Copy ( b_left, bt_row_right );
    PLA_Copy ( a_left, at_row_right );

    PLA_Obj_view_all ( c, & c_bottomright );


    for( ; ; )
    {
      PLA_Obj_split_size ( c_bottomright, PLA_SIDE_TOP,  & top_size,  & top_owner );
      PLA_Obj_split_size ( c_bottomright, PLA_SIDE_LEFT, & left_size, & left_owner );
      if ( ( size = min3 ( nb_alg, top_size, left_size) ) == 0) break;
      PLA_Obj_vert_split_2 ( at_row_right, size, & at_row_left, & at_row_right );
      PLA_Obj_vert_split_2 ( bt_row_right, size, & bt_row_left, & bt_row_right );

      PLA_Obj_horz_split_2 ( a_col_bottom, size, & a_col_top, & a_col_bottom );
      PLA_Obj_horz_split_2 ( b_col_bottom, size, & b_col_top, & b_col_bottom );

      PLA_Obj_split_4 ( c_bottomright, size, size, & c_topleft,    PLA_DUMMY,
                                                   & c_bottomleft, & c_bottomright );
      
	  if( (my_row == top_owner) && (my_col == left_owner) )
	       PLA_Local_syr2k ( PLA_LOWER_TRIANGULAR, PLA_NO_TRANSPOSE,
                           alpha, a_col_top, b_col_top, one, c_topleft );
      
      if( (my_col == left_owner) )
	  {
         PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
                          alpha, a_col_bottom, bt_row_left, one, c_bottomleft );

         PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
                          alpha, b_col_bottom, at_row_left, one, c_bottomleft );
	  }
    }
  }

  PLA_Obj_free ( & a_left );
  PLA_Obj_free ( & a_right );
  PLA_Obj_free ( & one );
  PLA_Obj_free ( & b_left  );
  PLA_Obj_free ( & b_right );
  PLA_Obj_free ( & c_topleft );
  PLA_Obj_free ( & c_bottomleft );
  PLA_Obj_free ( & c_bottomright );
  PLA_Obj_free ( & work_row );
  PLA_Obj_free ( & work_row_temp );
  PLA_Obj_free ( & work_col );
  PLA_Obj_free ( & work_col_temp );
  PLA_Obj_free ( & at_row_left );
  PLA_Obj_free ( & at_row_right );
  PLA_Obj_free ( & a_col_top );
  PLA_Obj_free ( & a_col_bottom );
  PLA_Obj_free ( & bt_row_left );
  PLA_Obj_free ( & bt_row_right );
  PLA_Obj_free ( & b_col_top );
  PLA_Obj_free ( & b_col_bottom );
  
  return ( PLA_SUCCESS );
}



/******************************************************************************/
/******************************************************************************/
int pla_syr2k_un_rankk(PLA_Obj alpha, PLA_Obj a, PLA_Obj b, PLA_Obj beta, PLA_Obj c)
{
  int     nb_alg, bottom_size, bottom_owner, size ;
  int     a_right_size, a_right_owner, b_right_size, b_right_owner;
  int     right_size, right_owner, my_row, my_col;
  MPI_Datatype datatype;
  PLA_Template template;
  PLA_Obj      a_left = NULL,    a_right = NULL, one = NULL,
               b_left = NULL,    b_right = NULL,
               c_topleft = NULL, c_bottomright = NULL, c_topright = NULL,
               work_col = NULL,  work_col_temp = NULL,
               a_col_top = NULL,  a_col_bottom = NULL,
               b_col_top = NULL,  b_col_bottom = NULL,
               work_row = NULL,  work_row_temp = NULL,
               at_row_left = NULL, at_row_right = NULL,
               bt_row_left = NULL, bt_row_right = NULL;

  PLA_Obj_template ( a, & template );
  PLA_Environ_nb_alg ( PLA_OP_PAN_PAN, template, & nb_alg );
  PLA_Temp_comm_row_rank( template, &my_col);
  PLA_Temp_comm_col_rank( template, &my_row);

  pla_trap_scal(beta, c, PLA_UPPER_TRIANGULAR);


  PLA_Obj_datatype ( c, & datatype );
  PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, & one );
  PLA_Obj_set_to_one ( one );

/* 2*nb alg -- okay, maybe this isn't too smart but we are taking a pretty */
/*             big hit in doing SO many gemm updates -- I'm trying to      */
/*             minimize that here by stacking A|B and B^T                  */
/*                                                    A^T  panels          */
/* I can use a single gemm call by doing this -- okay, big deal!           */

  PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb_alg*2, & work_row );
  PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb_alg*2, & work_col );

  PLA_Obj_view_all ( a, & a_left );
  PLA_Obj_view_all ( b, & b_left );

  for ( ; ; )
  {
/*    PLA_Obj_split_size ( a_left, PLA_SIDE_RIGHT, & a_right_size, & a_right_owner );
    PLA_Obj_split_size ( b_left, PLA_SIDE_RIGHT, & b_right_size, & b_right_owner ); */
    PLA_Obj_global_width(a_left, &a_right_size);
    PLA_Obj_global_width(b_left, &b_right_size);
    if ( (size = min3( nb_alg, a_right_size, b_right_size)) == 0) break;
    PLA_Obj_vert_split_2 ( a_left, -size, & a_left, & a_right );
    PLA_Obj_vert_split_2 ( b_left, -size, & b_left, & b_right );
    PLA_Obj_vert_split_2 ( work_col,       size,  & a_col_top, &work_col_temp );
    PLA_Obj_vert_split_2 ( work_col_temp,  size , & b_col_top, PLA_DUMMY);
    PLA_Obj_horz_split_2 ( work_row,       size,  & bt_row_left, &work_row_temp );
    PLA_Obj_horz_split_2 ( work_row_temp,  size,  & at_row_left, PLA_DUMMY);
    PLA_Obj_set_orientation ( a_right, PLA_PROJ_ONTO_COL );
    PLA_Obj_set_orientation ( b_right, PLA_PROJ_ONTO_COL );
/* The order of the copies is unimportant -- interlaced for programmer sanity */
/* If we strictly used an mvector intermediate this could be sped up by using */
/* a temp_mvector for both a and b -- however, since spreads in the oriented  */
/* direction employ a "short-circuting shortcut" I don't think that there is  */
/* much point.                                                                */
    PLA_Copy ( a_right, a_col_top );
    PLA_Copy ( b_right, b_col_top );
    PLA_Copy ( b_right, bt_row_left );
    PLA_Copy ( a_right, at_row_left );
    PLA_Obj_view_all ( c, & c_topleft );
    for( ; ; )
    {
      PLA_Obj_split_size ( c_topleft, PLA_SIDE_BOTTOM, & bottom_size,  & bottom_owner );
      PLA_Obj_split_size ( c_topleft, PLA_SIDE_RIGHT,  & right_size,   & right_owner );
      if ( ( size = min3 ( nb_alg, bottom_size, right_size) ) == 0) break;
      PLA_Obj_vert_split_2 ( at_row_left, -size, & at_row_left, & at_row_right );
      PLA_Obj_vert_split_2 ( bt_row_left, -size, & bt_row_left, & bt_row_right );
      PLA_Obj_horz_split_2 ( a_col_top, -size, & a_col_top, & a_col_bottom );
      PLA_Obj_horz_split_2 ( b_col_top, -size, & b_col_top, & b_col_bottom );
      PLA_Obj_split_4 ( c_topleft, -size, -size, & c_topleft,    & c_topright,
                                                 PLA_DUMMY,      & c_bottomright );
	  if( (my_row == bottom_owner) && (my_col == right_owner) )
	       PLA_Local_syr2k ( PLA_UPPER_TRIANGULAR, PLA_NO_TRANSPOSE,
                           alpha, a_col_bottom, b_col_bottom, one, c_bottomright );
      if( (my_col == right_owner) )
	  {
         PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
                          alpha, a_col_top, bt_row_right, one, c_topright );

         PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
                          alpha, b_col_top, at_row_right, one, c_topright );
	  }
    }
  }

  PLA_Obj_free ( & a_left );
  PLA_Obj_free ( & a_right );
  PLA_Obj_free ( & one );
  PLA_Obj_free ( & b_left  );
  PLA_Obj_free ( & b_right );
  PLA_Obj_free ( & c_topleft );
  PLA_Obj_free ( & c_topright );
  PLA_Obj_free ( & c_bottomright );
  PLA_Obj_free ( & work_row );
  PLA_Obj_free ( & work_row_temp );
  PLA_Obj_free ( & work_col );
  PLA_Obj_free ( & work_col_temp );
  PLA_Obj_free ( & at_row_left );
  PLA_Obj_free ( & at_row_right );
  PLA_Obj_free ( & a_col_top );
  PLA_Obj_free ( & a_col_bottom );
  PLA_Obj_free ( & bt_row_left );
  PLA_Obj_free ( & bt_row_right );
  PLA_Obj_free ( & b_col_top );
  PLA_Obj_free ( & b_col_bottom );
  
  return ( PLA_SUCCESS );
}



/******************************************************************************/

/******************************************************************************/
int pla_syr2k_lt_rankk(PLA_Obj alpha, PLA_Obj a, PLA_Obj b, PLA_Obj beta, PLA_Obj c)
{
  int     nb_alg, top_size, top_owner, size ;
  int     a_top_size, a_top_owner, b_top_size, b_top_owner;
  int     left_size, left_owner, my_row, my_col;
  MPI_Datatype datatype;
  PLA_Template template;
  PLA_Obj      a_top = NULL,    a_bottom = NULL, one = NULL,
               b_top = NULL,    b_bottom = NULL,
               c_topleft = NULL, c_bottomleft = NULL, c_bottomright = NULL,
               work_col = NULL,  work_col_temp = NULL,
               at_col_top = NULL,  at_col_bottom = NULL,
               bt_col_top = NULL,  bt_col_bottom = NULL,
               work_row = NULL,  work_row_temp = NULL,
               a_row_left = NULL, a_row_right = NULL,
               b_row_left = NULL, b_row_right = NULL;

  PLA_Obj_template ( a, & template );
  PLA_Environ_nb_alg ( PLA_OP_PAN_PAN, template, & nb_alg );
  PLA_Temp_comm_row_rank( template, &my_col);
  PLA_Temp_comm_col_rank( template, &my_row);

  pla_trap_scal(beta, c, PLA_LOWER_TRIANGULAR);


  PLA_Obj_datatype ( c, & datatype );
  PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, & one );
  PLA_Obj_set_to_one ( one );

/* 2*nb alg -- okay, maybe this isn't too smart but we are taking a pretty */
/*             big hit in doing SO many gemm updates -- I'm trying to      */
/*             minimize that here by stacking A|B and B^T                  */
/*                                                    A^T  panels          */
  PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb_alg*2, & work_row );
  PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb_alg*2, & work_col );

  PLA_Obj_view_all ( a, & a_bottom );
  PLA_Obj_view_all ( b, & b_bottom );

  for ( ; ; )
  {
/*    PLA_Obj_split_size ( a_bottom, PLA_SIDE_TOP, & a_top_size, & a_top_owner );
    PLA_Obj_split_size ( b_bottom, PLA_SIDE_TOP, & b_top_size, & b_top_owner );*/
    PLA_Obj_global_length(a_bottom, &a_top_size);
    PLA_Obj_global_length(b_bottom, &b_top_size);

    if ( (size = min3( nb_alg, a_top_size, b_top_size)) == 0) break;
    PLA_Obj_horz_split_2 ( a_bottom, size, & a_top, & a_bottom );
    PLA_Obj_horz_split_2 ( b_bottom, size, & b_top, & b_bottom );

    PLA_Obj_vert_split_2 ( work_col,       size, & at_col_bottom, &work_col_temp );
    PLA_Obj_vert_split_2 ( work_col_temp, size , & bt_col_bottom, PLA_DUMMY);

    PLA_Obj_horz_split_2 ( work_row,       size, & b_row_right, &work_row_temp );
    PLA_Obj_horz_split_2 ( work_row_temp,  size, & a_row_right, PLA_DUMMY);

    PLA_Obj_set_orientation ( a_top, PLA_PROJ_ONTO_ROW );
    PLA_Obj_set_orientation ( b_top, PLA_PROJ_ONTO_ROW );

/* The order of the copies is unimportant -- interlaced for programmer sanity */
/* If we strictly used an mvector intermediate this could be sped up by using */
/* a temp_mvector for both a and b -- however, since spreads in the oriented  */
/* direction employ a "short-circuting shortcut" I don't think that there is  */
/* much point.                                                                */

    PLA_Copy ( a_top, at_col_bottom );
    PLA_Copy ( b_top, bt_col_bottom );

    PLA_Copy ( b_top, b_row_right );
    PLA_Copy ( a_top, a_row_right );

    PLA_Obj_view_all ( c, & c_bottomright );

    for( ; ; )
    {
      PLA_Obj_split_size ( c_bottomright, PLA_SIDE_TOP,  & top_size,  & top_owner );
      PLA_Obj_split_size ( c_bottomright, PLA_SIDE_LEFT, & left_size, & left_owner );
      if ( ( size = min3 ( nb_alg, top_size, left_size) ) == 0) break;
      PLA_Obj_vert_split_2 ( a_row_right, size, & a_row_left, & a_row_right );
      PLA_Obj_vert_split_2 ( b_row_right, size, & b_row_left, & b_row_right );

      PLA_Obj_horz_split_2 ( at_col_bottom, size, & at_col_top, & at_col_bottom );
      PLA_Obj_horz_split_2 ( bt_col_bottom, size, & bt_col_top, & bt_col_bottom );

      PLA_Obj_split_4 ( c_bottomright, size, size, & c_topleft,    PLA_DUMMY,
                                                   & c_bottomleft, & c_bottomright );
      
	  if( (my_row == top_owner) && (my_col == left_owner) )
	       PLA_Local_syr2k ( PLA_LOWER_TRIANGULAR, PLA_NO_TRANSPOSE,
                           alpha, at_col_top, bt_col_top, one, c_topleft );
      
      if( (my_col == left_owner))
	  {
         PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
                          alpha, at_col_bottom, b_row_left, one, c_bottomleft );

         PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
                          alpha, bt_col_bottom, a_row_left, one, c_bottomleft );
	  }
    }
  }

  PLA_Obj_free ( & a_top );
  PLA_Obj_free ( & a_bottom );
  PLA_Obj_free ( & one );
  PLA_Obj_free ( & b_top  );
  PLA_Obj_free ( & b_bottom );
  PLA_Obj_free ( & c_topleft );
  PLA_Obj_free ( & c_bottomleft );
  PLA_Obj_free ( & c_bottomright );
  PLA_Obj_free ( & work_row );
  PLA_Obj_free ( & work_row_temp );
  PLA_Obj_free ( & work_col );
  PLA_Obj_free ( & work_col_temp );
  PLA_Obj_free ( & a_row_left );
  PLA_Obj_free ( & a_row_right );
  PLA_Obj_free ( & at_col_top );
  PLA_Obj_free ( & at_col_bottom );
  PLA_Obj_free ( & b_row_left );
  PLA_Obj_free ( & b_row_right );
  PLA_Obj_free ( & bt_col_top );
  PLA_Obj_free ( & bt_col_bottom );
  
  return ( PLA_SUCCESS );
}



/******************************************************************************/
/******************************************************************************/
int pla_syr2k_ut_rankk(PLA_Obj alpha, PLA_Obj a, PLA_Obj b, PLA_Obj beta, PLA_Obj c)
{
  int     nb_alg, bottom_size, bottom_owner, size ;
  int     a_bottom_size, a_bottom_owner, b_bottom_size, b_bottom_owner;
  int     right_size, right_owner, my_row, my_col;
  MPI_Datatype datatype;
  PLA_Template template;
  PLA_Obj      a_top = NULL,    a_bottom = NULL, one = NULL,
               b_top = NULL,    b_bottom = NULL,
               c_topleft = NULL, c_bottomright = NULL, c_topright = NULL,
               work_col = NULL,  work_col_temp = NULL,
               at_col_top = NULL,  at_col_bottom = NULL,
               bt_col_top = NULL,  bt_col_bottom = NULL,
               work_row = NULL,  work_row_temp = NULL,
               a_row_left = NULL, a_row_right = NULL,
               b_row_left = NULL, b_row_right = NULL;

  PLA_Obj_template ( a, & template );
  PLA_Environ_nb_alg ( PLA_OP_PAN_PAN, template, & nb_alg );
  PLA_Temp_comm_row_rank( template, &my_col);
  PLA_Temp_comm_col_rank( template, &my_row);

  pla_trap_scal(beta, c, PLA_UPPER_TRIANGULAR);


  PLA_Obj_datatype ( c, & datatype );
  PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, & one );
  PLA_Obj_set_to_one ( one );

/* 2*nb alg -- okay, maybe this isn't too smart but we are taking a pretty */
/*             big hit in doing SO many gemm updates -- I'm trying to      */
/*             minimize that here by stacking A|B and B^T                  */
/*                                                    A^T  panels          */
/* I can use a single gemm call by doing this -- okay, big deal!           */

  PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb_alg*2, & work_row );
  PLA_Pmvector_create_conf_to ( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb_alg*2, & work_col );

  PLA_Obj_view_all ( a, & a_top );
  PLA_Obj_view_all ( b, & b_top );

  for ( ; ; )
  {
/*    PLA_Obj_split_size ( a_top, PLA_SIDE_BOTTOM, & a_bottom_size, & a_bottom_owner );
    PLA_Obj_split_size ( b_top, PLA_SIDE_BOTTOM, & b_bottom_size, & b_bottom_owner ); */
    PLA_Obj_global_length(a_top, &a_bottom_size);
    PLA_Obj_global_length(b_top, &b_bottom_size);
    if ( (size = min3( nb_alg, a_bottom_size, b_bottom_size)) == 0) break;
    PLA_Obj_horz_split_2 ( a_top, -size, & a_top, & a_bottom );
    PLA_Obj_horz_split_2 ( b_top, -size, & b_top, & b_bottom );
    PLA_Obj_vert_split_2 ( work_col,       size,  & at_col_top, &work_col_temp );
    PLA_Obj_vert_split_2 ( work_col_temp,  size , & bt_col_top, PLA_DUMMY);
    PLA_Obj_horz_split_2 ( work_row,       size,  & b_row_left, &work_row_temp );
    PLA_Obj_horz_split_2 ( work_row_temp,  size,  & a_row_left, PLA_DUMMY);
    PLA_Obj_set_orientation ( a_bottom, PLA_PROJ_ONTO_ROW );
    PLA_Obj_set_orientation ( b_bottom, PLA_PROJ_ONTO_ROW );
/* The order of the copies is unimportant -- interlaced for programmer sanity */
/* If we strictly used an mvector intermediate this could be sped up by using */
/* a temp_mvector for both a and b -- however, since spreads in the oriented  */
/* direction employ a "short-circuting shortcut" I don't think that there is  */
/* much point.                                                                */
    PLA_Copy ( a_bottom, at_col_top );
    PLA_Copy ( b_bottom, bt_col_top );
    PLA_Copy ( b_bottom, b_row_left );
    PLA_Copy ( a_bottom, a_row_left );
    PLA_Obj_view_all ( c, & c_topleft );
    for( ; ; )
    {
      PLA_Obj_split_size ( c_topleft, PLA_SIDE_BOTTOM, & bottom_size,  & bottom_owner );
      PLA_Obj_split_size ( c_topleft, PLA_SIDE_RIGHT,  & right_size,   & right_owner );
      if ( ( size = min3 ( nb_alg, bottom_size, right_size) ) == 0) break;
      PLA_Obj_vert_split_2 ( a_row_left, -size, & a_row_left, & a_row_right );
      PLA_Obj_vert_split_2 ( b_row_left, -size, & b_row_left, & b_row_right );
      PLA_Obj_horz_split_2 ( at_col_top, -size, & at_col_top, & at_col_bottom );
      PLA_Obj_horz_split_2 ( bt_col_top, -size, & bt_col_top, & bt_col_bottom );
      PLA_Obj_split_4 ( c_topleft, -size, -size, & c_topleft,    & c_topright,
                                                 PLA_DUMMY,      & c_bottomright );
	  if( (my_row == bottom_owner) && (my_col == right_owner) )
	       PLA_Local_syr2k ( PLA_UPPER_TRIANGULAR, PLA_NO_TRANSPOSE,
                           alpha, at_col_bottom, bt_col_bottom, one, c_bottomright );
      if( (my_col == right_owner) )
	  {
         PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
                          alpha, at_col_top, b_row_right, one, c_topright );

         PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
                          alpha, bt_col_top, a_row_right, one, c_topright );
	  }
    }
  }

  PLA_Obj_free ( & a_top );
  PLA_Obj_free ( & a_bottom );
  PLA_Obj_free ( & one );
  PLA_Obj_free ( & b_top  );
  PLA_Obj_free ( & b_bottom );
  PLA_Obj_free ( & c_topleft );
  PLA_Obj_free ( & c_topright );
  PLA_Obj_free ( & c_bottomright );
  PLA_Obj_free ( & work_row );
  PLA_Obj_free ( & work_row_temp );
  PLA_Obj_free ( & work_col );
  PLA_Obj_free ( & work_col_temp );
  PLA_Obj_free ( & a_row_left );
  PLA_Obj_free ( & a_row_right );
  PLA_Obj_free ( & at_col_top );
  PLA_Obj_free ( & at_col_bottom );
  PLA_Obj_free ( & b_row_left );
  PLA_Obj_free ( & b_row_right );
  PLA_Obj_free ( & bt_col_top );
  PLA_Obj_free ( & bt_col_bottom );
  
  return ( PLA_SUCCESS );
}



/******************************************************************************/
int PLA_Syr2k(
   int uplo,
   int trans,
   PLA_Obj alpha,
   PLA_Obj a,
   PLA_Obj b,
   PLA_Obj beta,
   PLA_Obj c)
{
   PLA_Syr2k_enter( uplo, trans, alpha, a, b, beta, c);
   if(PLA_UPPER_TRIANGULAR == uplo)
   {
      if(PLA_NO_TRANSPOSE == trans)
         pla_syr2k_un_rankk(alpha, a, b, beta, c);
      else
         pla_syr2k_ut_rankk(alpha, a, b, beta, c); 
   }
   else /* if(PLA_LOWER_TRIANGULAR == uplo) */
   {
      if(PLA_NO_TRANSPOSE == trans)
         pla_syr2k_ln_rankk(alpha, a, b, beta, c);
      else 
         pla_syr2k_lt_rankk(alpha, a, b, beta, c); 
   }


   PLA_Syr2k_exit ( uplo, trans, alpha, a, b, beta, c);
   return(PLA_SUCCESS);
}
