#include "PLA.h"

int PLA_Copy_vector_row_opt( PLA_Obj vector, PLA_Obj matrix)
{
  PLA_Obj
    sendbuf = NULL,
    recvbuf = NULL;
  
  void 
    *sendbuf_ptr = NULL,
    *recvbuf_ptr = NULL,
    *vector_ptr = NULL;
  
  PLA_Template template;
  
  MPI_Datatype datatype;
  
  MPI_Comm comm_col;
  
  int 
    i,
    ione = 1,
    idummy,
    total,
    npcols,
    nprows,
    g_length_vector,
    g_align_row_vector,
    nb,
    np,
    local_length_vector,
    local_max_length_vector,
    sendbuf_length, sendbuf_width,
    size,
    myrow,
    mycol,
    owner_top, owner_bottom,
    owner_left, owner_right, owner_1D,
    vector_lld, sendbuf_lld;

  int 
    *displs,
    *recvcounts;

  PLA_Obj_obj_type dest_obj_type;
  

  
  
  /* Inquiries */

  PLA_Obj_template( vector, &template);
  PLA_Temp_comm_col( template, &comm_col);
  PLA_Temp_comm_row_rank( template, &mycol);
  PLA_Temp_comm_col_rank( template, &myrow);
  PLA_Temp_comm_row_size( template, &npcols);
  PLA_Temp_comm_col_size( template, &nprows);
  PLA_Temp_nb( template, &nb);
  PLA_Obj_objtype( matrix, &dest_obj_type);

  np = nprows*npcols;
   
  PLA_Obj_datatype( vector, &datatype);
  PLA_Obj_local_length( vector, &local_length_vector);
  PLA_Obj_global_length( vector, &g_length_vector);
  PLA_Obj_global_align_row( vector, &g_align_row_vector);
  PLA_Obj_local_buffer(vector, &vector_ptr);

  PLA_Obj_top(vector, &size, &owner_1D);
  owner_top = owner_1D%nprows;
  owner_left = owner_1D/nprows;
  PLA_Obj_left(matrix, &size, &idummy);

  PLA_Pmvector_create_conf_to(matrix, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
			      1, &sendbuf);

  PLA_Obj_local_ldim(vector,&vector_lld);
  PLA_Obj_local_ldim(sendbuf, &sendbuf_lld);
  PLA_Obj_local_length(sendbuf,&sendbuf_length);
  PLA_Obj_local_width(sendbuf, &sendbuf_width);

  PLA_Obj_local_buffer(sendbuf, &sendbuf_ptr);
  PLA_Obj_set_to_zero(sendbuf);

/*
  PLA_Show(vector);
  PLA_Show(sendbuf);
*/
  /* set the pointer to the beginning */

  if (myrow == owner_top && mycol == owner_left)
    {

 /* if i own the possible first partial block */
      PLA_dcopy( &size, vector_ptr, &ione, 
	     sendbuf_ptr, &ione );
      sendbuf_ptr = (void *)((char *)sendbuf_ptr + 
			     (size+nb*(nprows-1)) * sizeof(double));
      vector_ptr = (void *)((char *)vector_ptr + size * sizeof(double));
      total = size;
    }
  else if(mycol == owner_left )
    {
      /* i am on the same column with possibly partial block */
      /* skip to where non_zero's are */
      sendbuf_ptr = (void *)((char *)sendbuf_ptr + 
			     size*sizeof(double));
      sendbuf_ptr = (void *)((char *)sendbuf_ptr +
			     ((myrow + nprows - owner_top-1)%nprows)
			     *nb*sizeof(double));
      total = 0;
/*      printf("myrow = %d, total = %d\n",myrow,total);  */

    }
  else
    {
      /* i am not the column owner of the first vector
	 element--possible partial block */

      sendbuf_ptr = (void *)((char *)sendbuf_ptr +
			     (myrow%nprows)*nb*sizeof(double));
      total = 0;
    }

  /* now jump right into copying */
  /*  PLA_Show(sendbuf); */

  for(i=total; i<=local_length_vector-nb; i=i+nb){
    PLA_dcopy(&nb, vector_ptr, &ione, 
	       sendbuf_ptr, &ione );
    sendbuf_ptr = (void *)((char *)sendbuf_ptr +
			   (nprows)*nb*sizeof(double));
    vector_ptr = (void *)((char *)vector_ptr +
			  nb*sizeof(double));
  }
  /*take care of the possible last partial block */
  PLA_Obj_right(sendbuf, &size, &idummy);

  if ( i != local_length_vector ) 
    {
      /*  
   printf("size = %d, row = %d, col = %d, i = %d, llv = %d \n",
	     size, owner_bottom, owner_right, i, local_length_vector);
	     */    
      PLA_dcopy(&size, vector_ptr, &ione, 
	    sendbuf_ptr, &ione );      
    }

/*  PLA_Show(vector); */
/*    PLA_Show(sendbuf); */
  
  PLA_Reduce(sendbuf, MPI_SUM, matrix);
  PLA_Obj_free(&sendbuf);


  return 0;
}






int PLA_Copy_colmatrix_mvector_opt( PLA_Obj matrix, PLA_Obj mvector)
{
  int
    idummy,
    size;

  PLA_Obj
    matrix_view = NULL, mvector_view = NULL,
    current_matrix = NULL, remainder_matrix = NULL,
    current_mvector = NULL, remainder_mvector = NULL,
    current_dupl_mvector;
  

  PLA_Obj_view_all(matrix, &matrix_view);
  PLA_Obj_view_all(mvector, &mvector_view);

  for(;;){
    PLA_Obj_left(matrix_view, &size, &idummy);
    if (size == 0) break;
    PLA_Obj_vert_split_2( matrix_view, size, 
			 &current_matrix, &remainder_matrix); 
    PLA_Pmvector_create_conf_to( current_matrix, PLA_PROJ_ONTO_COL,
			      PLA_ALL_COLS, size,
			      &current_dupl_mvector);
    PLA_Obj_vert_split_2(mvector_view, size, 
			 &current_mvector, &remainder_mvector);

    PLA_Copy( current_matrix, current_dupl_mvector );
    PLA_Unpack_colpanel_mvector_opt( current_dupl_mvector, 
				    current_mvector);
    PLA_Obj_swap(&remainder_matrix, &matrix_view);
    PLA_Obj_swap(&remainder_mvector, &mvector_view);
    PLA_Obj_free(&current_dupl_mvector);
    PLA_Obj_free(&remainder_matrix);
    PLA_Obj_free(&remainder_mvector);
  }
  PLA_Obj_free(&current_mvector); PLA_Obj_free(&remainder_mvector);
  PLA_Obj_free(&current_matrix); PLA_Obj_free(&remainder_matrix);
  return 0;
}  
