#include <stdio.h>
#include "mpi.h"
#include "mmpi.h"
#include "SL.h"
#include "PLA.MMPI.h"
#include "MISC_DEFS.h"
#include "PLA_LAObj_set_matrix.h"

int PLA0_LAObj_set_submatrices_in_global_obj( 
           SL_LAObj     laobj,
           int           nsub_row,
           int           nsub_col,
           int          *size_row,
           int          *size_col,
           int          *displ_row,
           int          *displ_col,
           void         *local_matrix,
           int           local_matrix_ld )
/* 
 *  PLA_LAObj_set_submatrices_in_global_obj
 *
 *  Purpose: set contents of multiple local submatrices in global matrix
 */
{
  void             *mp;
  int              i,j,typesize, cur_row, cur_col;
  MPI_Datatype     datatype;
  struct PLA0_Fill_info *laobj_fill;

  /* check if object is fill-locked */
  if ( (laobj_fill = (struct PLA0_Fill_info *) laobj->fill ) == NULL ){
    printf("Attempt to fill object that is not FILL LOCKED\n");
    exit( 0 );
  }
  SL_LAObj_datatype( laobj, &datatype );
  MPI_Type_size( datatype, &typesize );

  mp = local_matrix;
  cur_col = 0;
  for ( j=0; j<nsub_col; j++ ) {
    /* mp = local_matrix + cur_col * local_matrix_ld * typesize;*/
    mp = (void*) (((char*)local_matrix) + cur_col * local_matrix_ld * typesize);
    for ( i=0; i<nsub_row; i++ ) {
      PLA0_LAObj_set_matrix_in_global_obj( laobj, size_row[i], size_col[i],
   				    displ_row[i], displ_col[i], 
				    mp, local_matrix_ld );
      /* mp = mp + size_row[i] * typesize; */
      mp = (void*) (((char*)mp) + size_row[i] * typesize);
    }
  }
}

int PLA0_LAObj_set_matrix_in_global_obj( 
           SL_LAObj      laobj,
	   int           m,
           int           n,
           int           displ_row,
           int           displ_col,
           void         *local_matrix,
           int           local_matrix_ld )
/* 
 *  PLA_LAObj_set_matrix_in_global_obj
 *
 *  Purpose: add contents of local matrix to global matrix
 *           MMPI producer routine
 */
{
  void         *mp, *mpp;
  int          current_pos_row, 
               current_pos_col, 
               objtype, 
               typesize,
               coor[2], dest, nprows,
               cur_row, cur_col,
               curpos_row, curpos_col,
               size_top, size_left,
               owner_row, owner_col,
               myrow, mycol, local_lld, i, curcol;
  MPI_Datatype datatype;
  struct PLA0_Fill_info *laobj_fill;
  SL_Template  template;
  MPI_Comm   comm;
  SL_LAObj     cur      = NULL,   cur1 = NULL,
               curleft   = NULL,  curleft1 = NULL,
               curtop   = NULL;

  /* check if object is fill-locked */
  if ( ( laobj_fill = (struct PLA0_Fill_info * ) laobj->fill ) == NULL ) {
    printf("Attempt to fill object that is not FILL LOCKED\n");
    exit( 0 );
  }

  SL_LAObj_objtype( laobj, &objtype );
  SL_LAObj_datatype( laobj, &datatype );
  SL_LAObj_template( laobj, &template );
  SL_LAObj_local_lld( laobj, &local_lld );
  SL_Temp_myrow( template, &myrow );
  SL_Temp_mycol( template, &mycol );
  
  SL_Temp_comm1D( template, &comm);
  
  MPI_Type_size( datatype, &typesize );
  if ( objtype == SL_MATRIX ) {
    /* let cur = laobj( displ_row:displ_row+m-1, displ_col:displ_col+n-1 ) */
    SL_Matrix_submatrix( laobj, m, n, displ_row, displ_col,   &cur );
    curpos_col = displ_col;
    cur_col = 0;
    while( TRUE ) {     /* loop through global matrix column panel at a time */
      /* mp = local_matrix + cur_col*local_matrix_ld*typesize;  */
      mp = (void*) (((char*)local_matrix)+ cur_col*local_matrix_ld*typesize); 
      /* split off left panel owned by one column of nodes */
      SL_LAObj_left( cur, &size_left, &owner_col );
      if ( size_left == 0 ) break;
      SL_LAObj_vert_split_2( cur, size_left,       &curleft, &cur1 );
      curpos_row = displ_row;
      while( TRUE ) {    /* loop through column panel block at a time */
	/* split off top block owned by one node */
	SL_LAObj_top( curleft, &size_top, &owner_row );
	if ( size_top == 0 ) break;
	SL_LAObj_horz_split_2( curleft, size_top,       &curtop, 
                                                        &curleft1 );
	if ( owner_row == myrow && owner_col == mycol ) {
	  /* local add */
/*	  PLA0_LAObj_set_local_contents( SL_NOTRANS, size_top, size_left,
 				  mp, local_matrix_ld, 1, curtop );  */
          PLA0_LAObj_set_local_contents1( SL_NOTRANS, datatype, 
                                          mp, local_matrix_ld, 1, curtop ); 
	}
	else {
	  /* initiate message to owner of curtop */
	  MMPI_Con_init( laobj_fill->consumer_receive, &laobj_fill->buffer );
	  MMPI_Buf_pack( &curpos_row, 1, MPI_INT, &laobj_fill->buffer );
	  MMPI_Buf_pack( &curpos_col, 1, MPI_INT, &laobj_fill->buffer );
	  MMPI_Buf_pack( &size_top, 1, MPI_INT, &laobj_fill->buffer );
	  MMPI_Buf_pack( &size_left, 1, MPI_INT, &laobj_fill->buffer );
	  mpp = mp;
	  for (i=0; i<size_left; i++) {
 	    MMPI_Buf_pack( mpp, size_top, datatype, &laobj_fill->buffer );
	    /* mpp += local_matrix_ld*typesize; */
	    mpp = (void*)(((char*)mpp) + local_matrix_ld*typesize);
	  } 
	  SL_Temp_num_proc_row( template, &nprows );
	  dest = owner_col*nprows + owner_row;
	  MMPI_Con_send( laobj_fill->buffer, dest,
			 laobj_fill->consumer_receive ); 
	}
	curpos_row = curpos_row + size_top;
	/* mp = mp + size_top * typesize; */
	mp = (void*) (((char*) mp) + size_top * typesize);
	SL_LAObj_swap( &curleft, &curleft1 );
      }
      curpos_col = curpos_col + size_left;
      cur_col += size_left;
      SL_LAObj_swap( &cur, &cur1 );
    }
  }
  else {
    printf("Attempt to fill an object that is not a matrix with a matrix\n");
    exit( 0 );
  }
  SL_LAObj_free( &cur ); 
  SL_LAObj_free( &curleft ); 
  SL_LAObj_free( &cur1 ); 
  SL_LAObj_free( &curleft1 ); 
  SL_LAObj_free( &curtop ); 

  MMPI_Serve_handler();     /* Opportunity for local consumer handler */
}

