#include "mpi.h"
#include "mmpi.h"
#include "SL.h"
#include "MISC_DEFS.h"
#include "PLA.MMPI.h"
#include "PLA_LAObj_set_vector.h"

#define PLA0_FILL_UNLOCKED NULL


int PLA0_LAObj_set_vector_in_global_obj( 
           SL_LAObj     laobj,
           int           n,
           int           displ,
           void         *local_vector )
/* 
 *  PLA_LAObj_set_vector_in_global_obj
 *
 *  Purpose: set contents of local vector in global vector
 *           MMPI producer routine
 */
{
  void         *vp;
  int          current_pos, 
               objtype, 
               curpos, size, owner, me, np, typesize;
  MPI_Datatype datatype;
  void *fillstate;
  SL_LAObj     cur      = NULL,  cur1 = NULL,
               curtop   = NULL;
  struct PLA0_Fill_info *laobj_fill;
  SL_Template template;
  MPI_Comm    comm;

  /* check if object is fill-locked */
  if ( ( laobj_fill = (struct PLA0_Fill_info *) laobj->fill ) == NULL ) {
    printf("Attempt to fill object that is not FILL LOCKED\n");
    exit( 0 );
  }; 

  SL_LAObj_objtype( laobj, &objtype );
  SL_LAObj_datatype( laobj, &datatype );
  SL_LAObj_template( laobj, &template );
  SL_Temp_me1D( template, &me );

  vp = local_vector;
  curpos = displ;
  MPI_Type_size( datatype, &typesize );
  switch (objtype) {
  case SL_VECTOR:
    /* let cur = laobj( displ:displ+n-1 ) */
    SL_Vector_subvector( laobj, n, displ,         &cur );

    while( TRUE ) {      /* loop through global vector until done */
      /* split off subvector owned by one node */
      SL_LAObj_top( cur, &size, &owner );
      if ( size == 0 ) break;
      SL_LAObj_horz_split_2( cur, size,       &curtop,
                                               &cur1 );
      if ( owner == me ) 
        /* local set */
/*	PLA0_LAObj_set_local_contents( SL_NOTRANS, size, 1, vp, 1, 1,
                                        curtop ); */
	PLA0_LAObj_set_local_contents1( SL_NOTRANS, datatype, vp, 1, 1,
                                        curtop ); 
      else {
	/* initiate message to owner of curtop */
	MMPI_Con_init( laobj_fill->consumer_receive, &laobj_fill->buffer );
        MMPI_Buf_pack( &curpos, 1, MPI_INT, &laobj_fill->buffer );
        MMPI_Buf_pack( &size, 1, MPI_INT, &laobj_fill->buffer );
        MMPI_Buf_pack( vp, size, datatype, &laobj_fill->buffer );
        MMPI_Con_send( laobj_fill->buffer, owner, laobj_fill->consumer_receive );
      };
      curpos = curpos + size;
      /* vp = vp + typesize * size; */
      vp = (void*) (((char*) vp) + typesize * size);
      SL_LAObj_swap( &cur, &cur1 );
    }
    break;
  default:
    printf("Attempt to fill an object that is not a vector with a vector\n");
    exit( 0 );
  }
  SL_LAObj_free( &cur ); 
  SL_LAObj_free( &cur1 ); 
  SL_LAObj_free( &curtop ); 

  MMPI_Serve_handler();     /* Opportunity for local consumer handler */
}


int PLA0_LAObj_set_subvectors_in_global_obj( 
           SL_LAObj     laobj,
           int           nsub,
           int          *size,
           int          *displ,
           void         *local_vector )
/* 
 *  PLA_LAObj_set_subvectors_in_global_obj
 *
 *  Purpose: set contents of multiple local subvectors in global vector
 */
{
  void            *fillstate,
                  *vp;
  int              i, typesize;
  MPI_Datatype     datatype;

  /* check if object is fill-locked */
  if ( laobj->fill  == NULL ) {
    printf("Attempt to fill object that is not FILL LOCKED\n");
    exit( 0 );
  }
  vp = local_vector;
  for ( i = 0; i<nsub; i++ ) {
    PLA0_LAObj_set_vector_in_global_obj( laobj, size[i], displ[i], vp );
    MPI_Type_size( datatype, &typesize );
    /* vp = vp + typesize; */
    vp = (void*) (((char*) vp) + typesize);
  }
}

