


/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/





#ifndef PLA_PANEL_H
#define PLA_PANEL_H


#ifndef NULL
#define NULL 0
#endif

/* PLA_Panel_dupl_conformto
 Creates a panel with size conformal to a given PLA_Obj, usually a matrix.
  Subsequently will be freed by PLA_Obj_free.

 This routine sill allocate a buffer large enough to hold the local part of the
 panel of length global_length and datatype.  The template describes how the
 matrix is layed out on the processors, hence how much is on the local processor.
*/
int PLA_Panel_dupl_conformto(PLA_Obj parent_obj,
                            int      size,
                            PLA_Panel_rowcol_flag rowcol,
                            PLA_Obj *dup_obj);

 
/* PLA_Panel_create
 Creates a panel.  Subsequently will be freed by PLA_Obj_free.

 This routine sill allocate a buffer large enough to hold the local part of the
 matrix of length global_length and datatype.  The template describes how the
 matrix is layed out on the processors, hence how much is on the local processor.
*/
int PLA_Panel_create (MPI_Datatype          datatype, 
                     PLA_Panel_rowcol_flag  rowcol,
                     PLA_Duplicate_flag     duplicate,
		      int                  length,
		      int                  width,
		      PLA_Template          template,
		      int                  global_align_idx,
		      PLA_Obj             *new_matrix);

/* PLA_Panel_subpanel
 This routines allocates a new matrix object, but does not allocate any memory
 to hold the data.  Instead it just contains a pointer to the buffer allocated
 in PLA_Matrix_create.  Hence changing this buffer will change the origional
 buffer.

 Also once a submatrix is created, freeing the origional matrix will NOT result
 in freeing the submatrix or the buffer origionally allocated.  Instead all memory
 will be preserved untill all submatrixs and the origional matrix have been
 freed with PLA_Obj_free, at which time the master buffer will be freed.
*/
int PLA_Panel_subpanel (PLA_Obj old_panel, 
			 int      length, 
			 int      width,
			 int      matrix_align_row,
			 int      matrix_align_col,
			 PLA_Obj *new_matrix);


int PLA_Panel_local_all (PLA_Obj          panel,
			 MPI_Datatype *    datatype,
			 int *             local_length,
			 int *             local_width,
			 void **           buffer,
			 int  *            local_lld);

int PLA_Panel_global_all (PLA_Obj          panel,
                         PLA_Panel_rowcol_flag * rowcol,
			  int *             length,
			  int *             width,
			  int  *            global_align_row,
			  int  *            global_align_col,
			  PLA_Template *     template);


#endif



