


/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/



#include "PLA.h"
#include "math.h"
#include "PLA_Dhrd2.h"

#define PLA_TRANSPOSE 1
#define PLA_NO_TRANSPOSE 0
 

int PLA_Dhrd2(PLA_Obj a)
{
  int 
    mycol,
    ncols,
    idummy,
    size,
    m,n,
    im,in,
    k,
    local_a_lld;

  PLA_Obj 
    v = NULL, 
    u = NULL, 
    y = NULL;

  PLA_Obj vcur=NULL, vcur2=NULL, ucur=NULL, ycur=NULL, ycur2=NULL,
           A_bottom=NULL, A_left=NULL, A_right=NULL, ucur2=NULL, 
           acur = NULL;  
  
  PLA_Obj 
    min_beta, 
    alpha,
    PLA_ONE,
    PLA_ZERO;
  
  MPI_Datatype
    datatype;
  
  PLA_Template
    template;
  
  void
    *alpha_buffer = NULL,
    *min_beta_buffer = NULL,
    *local_a_buffer = NULL,
    *one_buf,
    *zero_buf;


  PLA_Matrix_global_all( a, &m, &n, &im, &in, &template);
  PLA_Matrix_local_all( a, &datatype,
		      &idummy, &idummy,
		      &local_a_buffer, &local_a_lld);
  PLA_Temp_comm_row_size(template, &ncols);
  PLA_Temp_comm_row_rank(template, &mycol);

  if( m!=n )
    {
      printf("matrix A in PLA_Dhrd2.c is of illegal dimentions\n");
      return 1;
    }
  
/*create multiscalars PLA_ONE, PLA_ZERO) */
  PLA_Mscalar_create(MPI_DOUBLE,
                    PLA_ALL_ROWS, PLA_ALL_COLS,
		    1,1,
		    template,
		    &PLA_ONE);
  PLA_Obj_local_buffer(PLA_ONE, &one_buf);
  *((double *)one_buf) = 1.0;

  PLA_Mscalar_create(MPI_DOUBLE,
                    PLA_ALL_ROWS, PLA_ALL_COLS,
		    1,1,
		    template,
		    &PLA_ZERO);
  PLA_Obj_local_buffer(PLA_ZERO, &zero_buf);
  *((double *)zero_buf) = 0.0;




/*  create mscalars alpha, min_beta */
  PLA_Mscalar_create(datatype, 
                    PLA_ALL_ROWS, PLA_ALL_COLS,
		    1,1,
		    template,
		    &alpha);
  PLA_Obj_local_buffer(alpha, &alpha_buffer);
  

  PLA_Mscalar_create(datatype,
                    PLA_ALL_ROWS, PLA_ALL_COLS,
		    1,1,
		    template,
		    &min_beta);
  PLA_Obj_local_buffer(min_beta, &min_beta_buffer);
  




  PLA_Vector_create(datatype,		 
		    n,
		    template,
		    im,
		    &v);

  PLA_Vector_create(datatype,
		    n,
		    template,
		    im,
		    &u);

  PLA_Vector_create(datatype,
		    n,
		    template,
		    im,
		    &y);
  



  PLA_Matrix_submatrix( a, 
		      PLA_DIM_ALL, PLA_DIM_ALL,
		      PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
		      &acur );
  PLA_Vector_subvector( u, 
			PLA_DIM_ALL,
			PLA_ALIGN_FIRST,
			&ucur );
  PLA_Vector_subvector( v, 
			PLA_DIM_ALL,
			PLA_ALIGN_FIRST,
			&vcur );
  PLA_Vector_subvector( y, 
		      PLA_DIM_ALL,
		      PLA_ALIGN_FIRST,
		      &ycur );

  for(k=0;k<n-2;k++)
    {

      PLA_Matrix_submatrix( acur, n, n-k-1, 1, 2, &A_right); 
 

      PLA_Matrix_submatrix( acur, n-k-1, n-k, k+2, 1, &A_bottom);
      PLA_Matrix_submatrix(A_bottom, n-k-1, 1, 1, 1, &A_left);

      
      PLA_Vector_subvector( ucur, n-k-1, 2, &ucur2);

      PLA_Copy( A_left, ucur2 );

      PLA_Vector_subvector( ycur, n-k-1, 2, &ycur2);
      
      compute_household_vector( ucur2, min_beta );


      PLA_Gemv( PLA_TRANSPOSE, PLA_ONE, A_bottom, ucur2, PLA_ZERO, vcur );  
      PLA_Gemv( PLA_NO_TRANSPOSE, PLA_ONE, A_right, ucur2, PLA_ZERO, y );  

      PLA_Vector_subvector( vcur, n-k-1, 2, &vcur2);
      PLA_Dot( vcur2, ucur2, alpha );

      /* the line below should be implemented via PLA_axpy */

      *((double *)alpha_buffer) = *((double *)alpha_buffer) *
                                       *((double *)min_beta_buffer); 


      PLA_vaxpy( alpha, ucur2, ycur2 );

      PLA_Ger( min_beta, ucur2, vcur, A_bottom );
      PLA_Ger( min_beta, y, ucur2, A_right);

      PLA_Obj_swap(&ucur2, &ucur);
      PLA_Obj_swap(&vcur2, &vcur);
      PLA_Obj_swap( &A_right, &acur );
      PLA_Obj_swap(&ycur2, &ycur);
    }


  PLA_Obj_free(&vcur2);  
  PLA_Obj_free(&vcur);
  PLA_Obj_free(&v);

  PLA_Obj_free(&ycur2);
  PLA_Obj_free(&ycur);
  PLA_Obj_free(&y);

  PLA_Obj_free(&ucur2);
  PLA_Obj_free(&ucur);
  PLA_Obj_free(&u);

  PLA_Obj_free(&A_left);
  PLA_Obj_free(&A_bottom);
  PLA_Obj_free(&A_right);
  PLA_Obj_free(&acur);
  
}

int i_am_the_owner(PLA_Obj x)
{
  int 
    me, 
    align,
    x_owner;
  
  PLA_Template 
    template;
  
  PLA_Obj_template( x, &template );  
  PLA_Obj_global_align_row( x, &align);
  PLA_Temp_vector_owner_rank(template, align, &x_owner);
  PLA_Temp_me1D( template, &me );
  
   
  if (me == x_owner)
    return 1;
  else
    return 0;
}


int compute_household_vector(  PLA_Obj u, PLA_Obj min_beta )
{
  int
    idummy,
    size;

  double 
    NORM, 
    x;
  
  PLA_Obj 
    utop = NULL, 
    norm = NULL;

  void
    *x_buf = NULL,
    *norm_buf = NULL,
    *min_beta_buffer = NULL;
  
  PLA_Template
    template;


  PLA_Obj_template(u, &template);
  PLA_Mscalar_create(MPI_DOUBLE,
                    PLA_ALL_ROWS, PLA_ALL_COLS,
		    1,1,
		    template,
		    &norm);
  PLA_Obj_local_buffer(norm, &norm_buf);  
  PLA_Obj_local_buffer(min_beta, &min_beta_buffer);



  PLA_Vector_subvector(u, 1, 1, &utop);

  PLA_Nrm2(u, norm);

/* Do the fallowing part on the processor that contains utop */
  if (i_am_the_owner(utop)) 
      {

	PLA_Obj_local_buffer(utop, &x_buf);

	x = *((double *)x_buf);
	NORM = *((double *)norm_buf);

      /*compute the top element of a householder vector */
	if (x >= 0.0)  
	  x=x+sqrt(NORM);
	else
	  x=x-sqrt(NORM);
	
	*((double *)x_buf) = x;
      }


  PLA_Nrm2( u, norm);
  NORM = *((double *)norm_buf);

  /* compute minus beta */
  NORM=-2.0/NORM;
  *((double*)min_beta_buffer) = NORM;
  

  PLA_Obj_free(&utop);
  PLA_Obj_free(&norm);
  return 0;
}

  


  





















