
#include "PLA_Local_BLAS.h"

int PLA_Local_syrk ( int uplo, int trans,
                      PLA_Obj ALPHAobj, PLA_Obj Aobj,
                      PLA_Obj BETAobj, PLA_Obj Cobj )
{
   MPI_Datatype        Adatatype;
   int                 ALPHAloc_len,  Aloc_len,  BETAloc_len,  Cloc_len,
                       ALPHAloc_wid,  Aloc_wid,  BETAloc_wid,  Cloc_wid,
                       ALPHAstride,   Astride,   BETAstride,   Cstride,
                       ALPHAldim,     Aldim,     BETAldim,     Cldim;
   void              * ALPHAbuffer, * Abuffer, * BETAbuffer, * Cbuffer;
   char c_uplo, c_trans;

   PLA_Obj_local_info ( ALPHAobj, & ALPHAloc_len, & ALPHAloc_wid,
                              & ALPHAbuffer, & ALPHAstride, & ALPHAldim );

   PLA_Obj_datatype (Aobj, & Adatatype) ;
   PLA_Obj_local_info ( Aobj, & Aloc_len, & Aloc_wid,
                                 & Abuffer, & Astride, & Aldim );

   PLA_Obj_local_info ( BETAobj, & BETAloc_len, & BETAloc_wid,
                                    & BETAbuffer, & BETAstride, & BETAldim );

   PLA_Obj_local_info ( Cobj, & Cloc_len, & Cloc_wid,
                                 & Cbuffer, & Cstride, & Cldim );


   if ( Aloc_len > 0 && Aloc_wid > 0 && Cloc_len > 0 && Cloc_wid > 0 )
    {
      if ( Cloc_len != Cloc_wid )
      {
         printf ( "Non-square dimensions in PLA_Local_syrk " );
         printf ( ": length = %d, width = %d\n", Aloc_len, Aloc_wid );
         return ( -1 );
      }

/* there should be a ton more error checking here as far as lengths
   go, but the BLAS call will bomb if there is a problem, so what's
   the point??? :-)
*/

      c_uplo = 'L';
      c_trans = 'N';
      if ( uplo == PLA_UPPER_TRIANGULAR ) c_uplo = 'U';
      if ( trans == PLA_TRANSPOSE )
      {
         int tmp;
         c_trans = 'T';
         tmp = Aloc_len;
         Aloc_len = Aloc_wid;
         Aloc_wid = tmp;
      }

      if ( Adatatype == MPI_DOUBLE )
         PLA_dsyrk ( & c_uplo, & c_trans,
                    & Aloc_len, & Aloc_wid,
                    ( double * ) ALPHAbuffer, ( double * ) Abuffer, & Aldim,
                    ( double * ) BETAbuffer,  ( double * ) Cbuffer, & Cldim );
      else if ( Adatatype == MPI_FLOAT )
         PLA_ssyrk ( & c_uplo, & c_trans,
                    & Aloc_len, & Aloc_wid,
                    ( float * ) ALPHAbuffer, ( float * ) Abuffer, & Aldim,
                    ( float * ) BETAbuffer,  ( float * ) Cbuffer, & Cldim );
      else
      {
         printf ( "Invalid datatype in PLA_Local_syrk\n" );
         return ( -1 );
      }
   }
   return ( 0 );
}
