/* ---------------------------------------------------------------------
Author:     H. Carter Edwards
		http://www.ticam.utexas.edu/~carter
		carter@ticam.utexas.edu

Copyright:  Copyright (C) 1996   H. Carter Edwards

Purpose:    Asynchronous message management for MPI

License:    See 'License.Gnu'

Mail:
	University of Texas at Austin
	TICAM, TAY 2.400
	2313 Speedway
	Austin, TX  78712

Disclaimer:

    MMPI comes with ABSOLUTELY NO WARRANTY; for details read
    the file License.Gnu that came with this library.
    This is free software, and you are welcome to redistribute it
    under certain conditions, i.e. read the file 'License.Gnu'.
--------------------------------------------------------------------- */

#include <stdlib.h>
#include "mmpi.h"

static int con_handler(
  void   * extra_state , /* Extra state of handler */
  int      source ,      /* Message source */
  MMPI_Buf buffer )
{
  int Len ;
  char * name = (char *) extra_state ;
  char message[512] ;

  fprintf(MMPI_Log_file(),"con_handler{%s}( source = %d )\n", name, source );

  MMPI_Buf_unpack( buffer , &Len , 1 , MPI_INT );
  MMPI_Buf_unpack( buffer , message , Len , MPI_CHAR );

  fprintf(MMPI_Log_file(),"\t(%d)%s\n",Len,message);

  fflush(MMPI_Log_file());

  free(message);

  return MPI_SUCCESS ;
}


char name1[] = "CON#1" ;
char name2[] = "CON#2" ;

int main_test( MPI_Comm comm )
{
  int R ;
  int proc ;
  int nproc ;
  MMPI_Con con1 = MMPI_CON_NULL ;
  MMPI_Con con2 = MMPI_CON_NULL ;
  MMPI_Buf buf1 = MMPI_BUF_NULL ;
  MMPI_Buf buf2 = MMPI_BUF_NULL ;

  fprintf(MMPI_Log_file(),"main_test: MMPI_Enable\n" );
  fflush(MMPI_Log_file());

  if ( MPI_SUCCESS != ( R = MMPI_Enable( comm ) ) ) {
    MPI_Abort( comm , R );
  }

  if ( MPI_SUCCESS != ( R = MPI_Comm_size( comm , &nproc ) ) ) {
    MPI_Abort( comm , R );
  }
  if ( MPI_SUCCESS != ( R = MPI_Comm_rank( comm , &proc ) ) ) {
    MPI_Abort( comm , R );
  }

  fprintf(MMPI_Log_file(),"main_test: MMPI_Con_Create\n" );
  fflush(MMPI_Log_file());

  if ( MPI_SUCCESS != ( R =
    MMPI_Con_create( comm , name1 , con_handler , &con1 ) ) )
      MPI_Abort( comm , R );

  fprintf(MMPI_Log_file(),"main_test: MMPI_Con_Create #1 - ok\n" );
  fflush(MMPI_Log_file());

  if ( MPI_SUCCESS != ( R =
    MMPI_Con_create( comm , name2 , con_handler , &con2 ) ) )
      MPI_Abort( comm , R );

  fprintf(MMPI_Log_file(),"main_test: MMPI_Con_Create #2 - ok\n" );
  fflush(MMPI_Log_file());

  if ( MPI_SUCCESS != ( R = MMPI_Con_init( con1 , &buf1 ) ) ) {
    return R ;
  }
  if ( MPI_SUCCESS != ( R = MMPI_Con_init( con2 , &buf2 ) ) ) {
    return R ;
  }

/*
  fprintf(MMPI_Log_file(),"main_test: MMPI_Con_Init #1&2 - ok\n" );
  fflush(MMPI_Log_file());
*/

  MMPI_Con_verify( con1 );
  MMPI_Con_verify( con2 );

  {
    static char message1[] = "message #1" ;
    static char message2[] = "message #2 is longer than message #1" ;
    static char message3[] = "message #3 is between" ;
    static char message4[] = "message #4 is different" ;
    int len1 = strlen( message1 ) + 1 ;
    int len2 = strlen( message2 ) + 1 ;
    int len3 = strlen( message3 ) + 1 ;
    int len4 = strlen( message4 ) + 1 ;
    int i ;

    if ( MPI_SUCCESS != (R = MMPI_Buf_pack( &len1, 1, MPI_INT, &buf1))){
      MPI_Abort( comm , R );
    }
    if ( MPI_SUCCESS != (R = MMPI_Buf_pack(message1,len1,MPI_CHAR,&buf1))){
      MPI_Abort( comm , R );
    }
    if ( MPI_SUCCESS != (R = MMPI_Buf_pack( &len1, 1, MPI_INT, &buf2))){
      MPI_Abort( comm , R );
    }
    if ( MPI_SUCCESS != (R = MMPI_Buf_pack(message1,len1,MPI_CHAR,&buf2))){
      MPI_Abort( comm , R );
    }
/*
  fprintf(MMPI_Log_file(),"main_test: MMPI_Buf_pack #1&2 - ok\n" );
  fflush(MMPI_Log_file());
*/
  MMPI_Con_verify( con1 );
  MMPI_Con_verify( con2 );

    for ( i = 0 ; i < nproc ; ++i ) if ( i != proc ) {
      if ( MPI_SUCCESS != ( R = MMPI_Con_send( buf1, i, con1 ) ) ) {
          MPI_Abort( comm , R );
      }
      fprintf(MMPI_Log_file(),"main_test: MMPI_Con_send#1->%d\n",i);
      fflush(MMPI_Log_file());
      if ( MPI_SUCCESS != ( R = MMPI_Con_send( buf2, i, con2 ) ) ) {
          MPI_Abort( comm , R );
      }
      fprintf(MMPI_Log_file(),"main_test: MMPI_Con_send#2->%d\n",i);
      fflush(MMPI_Log_file());
    }

    if ( MPI_SUCCESS != ( R = MMPI_Con_init( con1 , &buf1 ) ) ) {
      return R ;
    }
    if ( MPI_SUCCESS != ( R = MMPI_Con_init( con2 , &buf2 ) ) ) {
      return R ;
    }
/*
  fprintf(MMPI_Log_file(),"main_test: MMPI_Con_init #1&2 - ok\n" );
  fflush(MMPI_Log_file());
*/
  MMPI_Con_verify( con1 );
  MMPI_Con_verify( con2 );


    if ( MPI_SUCCESS != (R = MMPI_Buf_pack( &len2, 1, MPI_INT, &buf1))){
      MPI_Abort( comm , R );
    }
    if ( MPI_SUCCESS != (R = MMPI_Buf_pack(message2,len2,MPI_CHAR,&buf1))){
      MPI_Abort( comm , R );
    }
    if ( MPI_SUCCESS != (R = MMPI_Buf_pack( &len2, 1, MPI_INT, &buf2))){
      MPI_Abort( comm , R );
    }
    if ( MPI_SUCCESS != (R = MMPI_Buf_pack(message2,len2,MPI_CHAR,&buf2))){
      MPI_Abort( comm , R );
    }
/*
  fprintf(MMPI_Log_file(),"main_test: MMPI_Buf_pack #1&2 - ok\n" );
  fflush(MMPI_Log_file());
*/
    for ( i = 0 ; i < nproc ; ++i ) if ( i != proc ) {
      if ( MPI_SUCCESS != ( R = MMPI_Con_send( buf1, i, con1 ) ) ) {
          MPI_Abort( comm , R );
      }
      fprintf(MMPI_Log_file(),"main_test: MMPI_Con_send#1->%d\n",i);
      fflush(MMPI_Log_file());
      if ( MPI_SUCCESS != ( R = MMPI_Con_send( buf2, i, con2 ) ) ) {
          MPI_Abort( comm , R );
      }
      fprintf(MMPI_Log_file(),"main_test: MMPI_Con_send#2->%d\n",i);
      fflush(MMPI_Log_file());
    }

  MMPI_Con_verify( con1 );
  MMPI_Con_verify( con2 );

    if ( MPI_SUCCESS != ( R = MMPI_Barrier( comm ) ) ) {
      MPI_Abort( comm , R );
    }

    if ( MPI_SUCCESS != ( R = MMPI_Con_init( con1 , &buf1 ) ) ) {
      MPI_Abort( comm , R );
    }
    if ( MPI_SUCCESS != (R = MMPI_Buf_pack( &len3, 1, MPI_INT, &buf1))){
      MPI_Abort( comm , R );
    }
    if ( MPI_SUCCESS != (R = MMPI_Buf_pack(message3,len3,MPI_CHAR,&buf1))){
      MPI_Abort( comm , R );
    }
    if ( MPI_SUCCESS != ( R = MMPI_Con_init( con2 , &buf2 ) ) ) {
      return R ;
    }
    if ( MPI_SUCCESS != (R = MMPI_Buf_pack( &len4, 1, MPI_INT, &buf2))){
      MPI_Abort( comm , R );
    }
    if ( MPI_SUCCESS != (R = MMPI_Buf_pack(message4,len4,MPI_CHAR,&buf2))){
      MPI_Abort( comm , R );
    }

    if ( proc == 0 ) {
    for ( i = 0 ; i < nproc ; ++i ) if ( i != proc ) {
      if ( MPI_SUCCESS != ( R = MMPI_Con_send( buf1, i, con1 ) ) ) {
          MPI_Abort( comm , R );
      }
      fprintf(MMPI_Log_file(),"main_test: MMPI_Con_send#1->%d\n",i);
      fflush(MMPI_Log_file());
      if ( MPI_SUCCESS != ( R = MMPI_Con_send( buf2, i, con2 ) ) ) {
          MPI_Abort( comm , R );
      }
      fprintf(MMPI_Log_file(),"main_test: MMPI_Con_send#2->%d\n",i);
      fflush(MMPI_Log_file());
    }
   }

/*
    for ( i = 0 ; i < nproc ; ++i ) if ( i != proc ) {
      if ( MPI_SUCCESS != ( R = MMPI_Con_wait( con , i ) ) )
        MPI_Abort( comm , R );
      fprintf(MMPI_Log_file(),"main_test: MMPI_Con_wait->%d\n",i);
      fflush(MMPI_Log_file());
    }
*/
  }

  MMPI_Con_verify( con1 );
  MMPI_Con_verify( con2 );

  if ( MPI_SUCCESS != ( R = MMPI_Buf_free( &buf1 ) ))
    MPI_Abort( comm , R );

  fprintf(MMPI_Log_file(),"main_test: MMPI_Buf_free #1\n");
  fflush(MMPI_Log_file());

  if ( MPI_SUCCESS != ( R = MMPI_Buf_free( &buf2 ) ))
    MPI_Abort( comm , R );

  fprintf(MMPI_Log_file(),"main_test: MMPI_Buf_free #2\n");
  fflush(MMPI_Log_file());

  fprintf(MMPI_Log_file(),"main_test: MMPI_Con_free #2 - BEGIN\n");
  fflush(MMPI_Log_file());

  if ( MPI_SUCCESS != ( R = MMPI_Con_free( &con2 ) ) )
      MPI_Abort( comm , R );

  fprintf(MMPI_Log_file(),"main_test: MMPI_Con_free #2 - DONE\n");
  fflush(MMPI_Log_file());

  fprintf(MMPI_Log_file(),"main_test: MMPI_Con_free #1 - BEGIN\n");
  fflush(MMPI_Log_file());

  if ( MPI_SUCCESS != ( R = MMPI_Con_free( &con1 ) ) ) {
      MPI_Abort( comm , R );
  }

  fprintf(MMPI_Log_file(),"main_test: MMPI_Con_free #1 - DONE\n");
  fflush(MMPI_Log_file());

  fprintf(MMPI_Log_file(),"main_test: MMPI_Disable\n");
  fflush(MMPI_Log_file());

  MMPI_Disable( comm );

  fprintf(MMPI_Log_file(),"main_test: - DONE\n");
  fflush(MMPI_Log_file());

  return 0 ;
}


main( int argc , char * argv[] )
{
  int N = 5 ;
  int R , i ;

  if ( MPI_SUCCESS != ( R = MPI_Init( &argc , &argv ) ) ) {
    MPI_Abort( MPI_COMM_WORLD , R );
  }
  if ( MPI_SUCCESS != ( R = MMPI_Log_init( "LogProCon." ) ) ) {
    MPI_Abort( MPI_COMM_WORLD , R );
  }

  for ( i = 0 ; i < N ; ++i ) {
    fprintf( MMPI_Log_file() , "Test Pass #%d\n" , i );
    main_test( MPI_COMM_WORLD );
  }

  MPI_Finalize();

  return 0 ;
}

