/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/
#include "PLA.h"
/* SHOULD the buffer size be a long???? */
int PLA_Mvector_buf_size (MPI_Datatype      datatype,
		      int               global_length,
		      int               global_width,
		      PLA_Template      template,
		      int               global_align,
		      int           *bytes_required)
{
   int local_length, local_width, data_size;
   int zero_or_one;

   MPI_Type_size(datatype, &data_size);
   PLA_Temp_zero_or_one(template, &zero_or_one);
   if(PLA_ALIGN_FIRST == global_align)
      global_align = zero_or_one;
   PLA_Temp_vector_local_length (template, global_length, 
                       global_align, &local_length);
   local_width = global_width;
   *bytes_required = local_length * local_width * data_size;
   return(PLA_SUCCESS);
   
}


/* Buffer flags
1 : A buffer is passed for use but the object does not exist
0 : A buffer is not passed and one should be malloced
-1: A buffer is not passed and one should NOT be malloced
*/

int PLA_Mvector_create (MPI_Datatype      datatype,
		      int               global_length,
		      int               global_width,
		      PLA_Template      template,
		      int               global_align,
		      PLA_Obj          *new_mvector)
{
   void *data_buffer = NULL;
   int  buffer_flag =  0; /* indicates that no buffer is passed but one should be made */
   int  buffer_size =  0; /* unimportant in this case -- may be "rolled into" the flag */
   return PLA_Mvector_create_base(datatype, global_length,global_width,template,
	                               global_align,
                                  buffer_flag, buffer_size, data_buffer,
								          new_mvector);
   /* One of my new fields */
   (*new_mvector)->allocation_type = PLA_HEAP;       /* vs. PLA_USER_BUFFER */
   (*new_mvector)->has_data_space  = PLA_HAS_SPACE;  /* vs. PLA_NO_SPACE */
}

int PLA_Mvector_create_buf (MPI_Datatype      datatype,
		      int               global_length,
		      int               global_width,
		      PLA_Template      template,
		      int               global_align,
			   int               buffer_size,
			   void             *data_buffer,
		      PLA_Obj          *new_mvector)
{
   int  buffer_flag = 1; /* indicates that a buffer is passed for use */
   return PLA_Mvector_create_base(datatype, global_length,global_width,template,
	                               global_align,
                                	 buffer_flag, buffer_size, data_buffer,
								          new_mvector);
   (*new_mvector)->allocation_type = PLA_USER_BUFFER;    /* vs. PLA_HEAP */
   (*new_mvector)->has_data_space  = PLA_HAS_SPACE;      /* vs. PLA_NO_SPACE */
}

int PLA_Mvector_create_no_buf (MPI_Datatype      datatype,
		      int               global_length,
		      int               global_width,
		      PLA_Template      template,
		      int               global_align,
		      PLA_Obj          *new_mvector)
{
   int  buffer_flag = -1; /* indicates that a buffer is passed for use */
   void *data_buffer = NULL;
   int  buffer_size =  0; /* unimportant in this case -- may be "rolled into" the flag */
   return PLA_Mvector_create_base(datatype, global_length,global_width,template,
	                               global_align,
                                  buffer_flag, buffer_size, data_buffer,
								          new_mvector);
   (*new_mvector)->allocation_type = PLA_UNDEFINED;    /* not defined until space attached  */
   (*new_mvector)->has_data_space  = PLA_NO_SPACE;    /* vs. PLA_HAS_SPACE */
}

int PLA_Mvector_attach_buf(PLA_Obj mvector, int buffer_size, void *data_buffer)
{
   int data_size, buf_size, local_buffer_size_required;
   MPI_Datatype datatype;

   PLA_Obj_datatype(mvector, &datatype);
	/* Check for size correctness,
	   Set the flag that indicates that space is passed in */
   MPI_Type_size (datatype, &data_size);
   buf_size = mvector->local_length * mvector->local_width;
   local_buffer_size_required = data_size * buf_size;
   if(buffer_size < local_buffer_size_required)
   {
      printf("PLA_Mvector_attach_buf was called with an attach buffer that was too small\n");
      return(1);
   }
   else mvector->master_buffer   = data_buffer;
   mvector->allocation_type = PLA_USER_SUPPLIED;
   mvector->has_data_space  = PLA_HAS_SPACE;
}

int PLA_Mvector_attach(PLA_Obj mvector)
{
   /* Need to modularize the matrix create routine */
   /* This routine allocs buffer space for a passed in object -- actually,
      I am considering making them all just call PLA_Object_attach
      because they all do the same thing */
   int data_size, buf_size;
   MPI_Datatype datatype;

   PLA_Obj_datatype(mvector, &datatype);
	/* Check for size correctness,
	   Set the flag that indicates that space is passed in */
   MPI_Type_size (datatype, &data_size);
   buf_size = mvector->local_length * mvector->local_width;
   if (0 != buf_size)
   {
      MPI_Type_size (datatype, &data_size);
      mvector->master_buffer = (void *) pla_calloc (buf_size,(int)data_size);
      if (mvector->master_buffer == NULL)
      {
         printf (" Could not allocate buffer of %d bytes in PLA_Mvector_attach\n",
	      buf_size*(int)data_size);
         return(1);
      }
   }
   else mvector->master_buffer = NULL;

   mvector->allocation_type    = PLA_HEAP;
   mvector->has_data_space     = PLA_HAS_SPACE;
}

int PLA_Mvector_create_base
            (MPI_Datatype      datatype,
		       int               global_length,
		       int               global_width,
		       PLA_Template      template,
		       int               global_align,
             int               buffer_flag,
             int               external_buffer_size,
             void             *external_buffer,
		       PLA_Obj          *new_mvector)
{

  PLA_Obj       mvector = NULL;
  int           buf_size;
  int           data_size = 0;
  int           zero_or_one; /* was PLA_Start_num 05/08/97 JAG */
  int           error = 0;
  int           index; /* for getting Obj_struct and view space */
  int           local_buffer_size_required;


  if(! PLA_Initialized_object_structs())
     PLA_Init_object_structs(PLA_DEFAULT_OBJ_COUNT); /* Some arbitrary number - mark/copy */

  if (NULL != *new_mvector) error += PLA_Obj_free (new_mvector);

  index = PLA_Get_free_object_struct();
  if(-1 == index) /* One could not be found */
  {
     printf("Could not get an allocated object structure\n");
     printf("either the user did not set aside any object structures\n");
     printf("or they didn't set aside any and the system number is low\n");
     /* Note : in the future this will call for doing a mark-and-copy  */
     /*        type deal -- IF the user did NOT do their own set-aside */
     /*        and just FAILING if they did.                           */
  }
/*printf("Mvector : Index is %5d and place is %ld\n", index, &PLA_GLOBAL_OBJ_STRUCT[index]);*/
  mvector = (PLA_Obj) (&PLA_GLOBAL_OBJ_STRUCT[index]);
  error += PLA_Obj_init (mvector);
  mvector->obj_type         = PLA_MVECTOR;
  mvector->datatype         = datatype;
  mvector->global_length    = mvector->master_length = global_length;
  mvector->global_width     = mvector->master_width  = global_width;
  mvector->local_width      = global_width;
  mvector->local_stride     = 1;
  mvector->template         = template;
  error += PLA_Temp_zero_or_one(template, &zero_or_one);
  if (PLA_ALIGN_FIRST == global_align)  global_align = zero_or_one;
  mvector->global_align_row = mvector->master_align_row = global_align;
  mvector->global_align_col = mvector->master_align_col = (int) zero_or_one;

  /*
     Determine the local length;
  */
  error += PLA_Temp_vector_local_length (template,
					global_length,
					global_align,
					&(mvector->local_length));
  mvector->local_lld = mvector->local_length;

  /*
     Might as well allocate the buffer;
  */
  if(-1 == buffer_flag) /* Create an object with no buffer */
     buf_size = -1;
  else buf_size = mvector->local_length * global_width;
  if(0 == buffer_flag)
  {
     if (0 != buf_size)
     {
        MPI_Type_size (datatype, &data_size);
        mvector->master_buffer = (void *) pla_calloc (buf_size,(int)data_size);
        if (mvector->master_buffer == NULL)
        {
           printf ("Could not allocate buffer of %d bytes in PLA_Mvector_create\n",
	                 buf_size*(int)data_size);
           return(1);
        }
     }
     else
     {
       mvector->master_buffer = NULL;    /* JAG NOTE : Matrix only does 1st line*/
       mvector->local_length     = 0;
       mvector->local_width      = 0;
       mvector->local_stride     = 0;
     }
  }
  else if(1 == buffer_flag)
  {
     MPI_Type_size (datatype, &data_size);
     local_buffer_size_required = data_size * buf_size;
     if(external_buffer_size < local_buffer_size_required)
     {
        printf("PLA_Mvectpr_create_buf was called with an attached buffer that was too small\n");
        return(1);
     }
     mvector->master_buffer = external_buffer;
  }
  mvector->length_master_buffer_byte = buf_size*(int)data_size;
  mvector->buffer = mvector->master_buffer;

  /* The views will also have to be taken from some set-aside space */
  index = PLA_Get_free_views_struct();
  if(-1 == index)
  {
     printf("Could not get an allocated views structure in Mvector creation\n");
     printf("either the user did not set aside any object structures\n");
     printf("or they didn't set aside any and the system number is low\n");
     /* Note : in the future this will call for doing a mark-and-copy  */
     /*        type deal -- IF the user did NOT do their own set-aside */
     /*        and just FAILING if they did.                           */
  }
  mvector->views  = (int *) (&PLA_GLOBAL_VIEWS_STRUCT[index]);
  (*(mvector->views))= 1;

  /* Set the first and last block size in both dimentions */

  error += PLA_Temp_vector_first_row_block  (template,
					    global_length,
					    global_align,
					    &(mvector->owner_top_block),
					    &(mvector->length_top_block));
  error += PLA_Temp_vector_last_row_block   (template,
					    global_length,
					    global_align,
					    &(mvector->owner_bottom_block),
					    &(mvector->length_bottom_block));
  mvector->width_left_block = 0;
  mvector->width_right_block  = 0;
  mvector->owner_left_block = 0;
  mvector->owner_right_block  = 0;

  *new_mvector = mvector;
  return(error);
}

int PLA_Mvector_submvector (PLA_Obj old_mvector, 
			   int      global_length, 
			   int      global_width,
			   int      matrix_align_row,
			   int      matrix_align_col,
			   PLA_Obj *new_mvector)
{
  
  
  PLA_Obj      mvector;
  PLA_Template   template;
  int           global_align_row, global_align_col; 
  int           local_len;
  int           local_align, offset;
  int           error=0;
  int           data_size;
  int           index;
  PLA_Start_num  zero_or_one;

  if ( old_mvector->obj_type != PLA_MVECTOR) {
    printf ("Invalid object passed to PLA_Mvector_subvector: NonMvector\n");
    return (1);
  }  

  if (*new_mvector != NULL) error += PLA_Obj_free (new_mvector);

  template = old_mvector->template;
  error += PLA_Temp_zero_or_one (template, &zero_or_one);

  index = PLA_Get_free_object_struct();
  if(-1 == index) /* One could not be found */
  {
     printf("Could not get an allocated object structure\n");
     printf("either the user did not set aside any object structures\n");
     printf("or they didn't set aside any and the system number is low\n");
     return(1);
     /* Note : in the future this will call for doing a mark-and-copy  */
     /*        type deal -- IF the user did NOT do their own set-aside */
     /*        and just FAILING if they did.                           */
  }
/*printf("subMvector : Index is %5d and place is %ld\n", index, &PLA_GLOBAL_OBJ_STRUCT[index]); */
  mvector = (PLA_Obj) (&PLA_GLOBAL_OBJ_STRUCT[index]);
  
  /* 
     By default copy all the fields in one blow.  Then modify
     the ones that need modifying.
     */
  *mvector = *old_mvector;

  if (global_length == PLA_DIM_ALL)  global_length = mvector->global_length;
  else                              mvector->global_length = global_length;
  
  if (global_width == PLA_DIM_ALL)   global_width = mvector->global_width;
  else                              mvector->global_width = global_width;
  
  if (matrix_align_row == PLA_ALIGN_FIRST)  matrix_align_row = zero_or_one;
  if (matrix_align_col == PLA_ALIGN_FIRST)  matrix_align_col = zero_or_one;


  global_align_row          = 
    mvector->global_align_row + matrix_align_row - (int)zero_or_one;
  mvector->global_align_row = global_align_row;
  
  global_align_col          = 
    mvector->global_align_col + matrix_align_col - (int)zero_or_one;
  mvector->global_align_col = global_align_col;

  /* 
    Set the local width and length.  There is no check that the 
    submultivector is actually contained within the original
    multivector.  It could be that the calling program messed up.
    For now assume that the calling program knows what it is doing.
    */

  error += PLA_Temp_vector_local_length (template, 
					global_length, 
					global_align_row, 
					&local_len);
  if (error) return(error);

  mvector->local_length = local_len;
  mvector->local_width  = global_width;
  /*
     Determine how much of the local vector to skip at the beginning.
     
     There is a detailed description of this in PLA_Matrix_submatrix.
     */
  if (local_len != 0) {
    matrix_align_row -= (int) zero_or_one;
    matrix_align_col -= (int) zero_or_one;

    error += PLA_Temp_vector_local_length (template, 
					  matrix_align_row,
					  old_mvector->global_align_row,
					  &local_align);
    /* 
       Now have to compute the offset into the buffer.  We want to find
       row local_align and column align_col on a buffer of leading 
       dimention mvector->lld.
       */
    offset =  matrix_align_col * mvector->local_lld + local_align;
    
    MPI_Type_size (mvector->datatype, &data_size);
    offset *= (int) data_size;

    mvector->buffer = (void *) ( (char *)(mvector->buffer) + offset );
  }
  else 
    mvector->buffer = NULL;

  (*(mvector->views))++;


  /* Set the first and last block size in both dimentions */
  
  error += PLA_Temp_vector_first_row_block   (template,
					     global_length,
					     global_align_row,
					     &(mvector->owner_top_block),
					     &(mvector->length_top_block));
  error += PLA_Temp_vector_last_row_block    (template,
					     global_length,
					     global_align_row,
					     &(mvector->owner_bottom_block),
					     &(mvector->length_bottom_block));
  mvector->width_left_block = 0;
  mvector->width_right_block  = 0;
  mvector->owner_left_block = 0;
  mvector->owner_right_block  = 0;


  *new_mvector = mvector;

  return(error);
}



int PLA_Mvector_subvector (PLA_Obj old_mvector, 
			  int      global_length, 
			  int      matrix_align_row,
			  int      matrix_align_col,
			  PLA_Obj *new_vector)
{
  int           error=0;
  
  error += PLA_Mvector_submvector (old_mvector, 
				  global_length, 
				  1,
				  matrix_align_row,
				  matrix_align_col,
				  new_vector);
 
  if (error!=0) (*new_vector)->obj_type = PLA_VECTOR;
  
  return(error);
}




int PLA_Mvector_local_all (PLA_Obj          mvector,
			  MPI_Datatype *    datatype,
			  int *             local_length,
			  int *             local_width,
			  void **           buffer,
			  int  *            local_stride,
			  int  *            local_lld)
{
  int error=0;

  error += PLA_Obj_datatype       (mvector,  datatype);
  error += PLA_Obj_local_length   (mvector,  local_length);
  error += PLA_Obj_local_width    (mvector,  local_width);
  error += PLA_Obj_local_buffer         (mvector,  buffer); 
  error += PLA_Obj_local_stride   (mvector,  local_stride); 
  error += PLA_Obj_local_ldim      (mvector,  local_lld); 
  return (error);
}


int PLA_Mvector_global_all (PLA_Obj          mvector,
			   int *             global_length,
			   int *             global_width,
			   int  *            global_align,
			   PLA_Template *     template)
     
{
  int error=0;

  error += PLA_Obj_global_length            (mvector,  global_length);
  error += PLA_Obj_global_width             (mvector,  global_width);
  error += PLA_Obj_global_align_row  (mvector,  global_align); 
  error += PLA_Obj_template          (mvector,  template); 
  return (error);
}

  
     
int PLA_Mvector_view_shift (PLA_Obj mvector,
			    int length_top,
			    int width_left,
			    int width_right,
			    int length_bottom)
{
  
  PLA_Template   template;
  int           global_align_row, global_align_col;
  int           local_len, global_length;
  int           local_align_col, local_width, global_width;
  int           local_align, offset;
  int           error=0;
  int           data_size;
  PLA_Start_num  zero_or_one;


  if ( mvector->obj_type != PLA_MVECTOR && 
       mvector->obj_type != PLA_VECTOR) {
    printf ("Invalid object passed to PLA_Mvector_subvector: NonMvector\n");
    return (1);
  }  


  template = mvector->template;

  global_align_row = mvector->global_align_row + length_top;
  global_align_col = mvector->global_align_col + width_left;
  global_length = mvector->global_length + length_bottom - length_top;
  global_width = mvector->global_width + width_right - width_left;


  /* 
    Set the local width and length.  There is no check that the 
    submultivector is actually contained within the original
    multivector.  It could be that the calling program messed up.
    For now assume that the calling program knows what it is doing.
    */


  if ( mvector->master_align_row > global_align_row || 
       mvector->master_align_col > global_align_col) {
    printf
      ("Mvector_view_shift shifted before the beginning of the mvector\n");
    printf
      ("Original alignment (%d,%d) plus offset (%d,%d)\n",
       mvector->global_align_row,mvector->global_align_col,
       length_top, width_left);
    printf
      ("gives offset (%d,%d)\n",
       global_align_row, global_align_col);
    printf
      ("Original creation alignment is (%d,%d)\n",
       mvector->master_align_row, mvector->master_align_col);
  }
  
  if (global_align_row+global_length >
         mvector->master_align_row+mvector->master_length   ||
      global_align_col+global_width  >
         mvector->master_align_col+mvector->master_width) {
    printf
      ("PLA_Mvector_view_shift shifted past the end of the Mvector\n");
    printf
      ("New alignment (%d,%d) plus length (%d,%d)\n",
       global_align_row,global_align_col,
       global_length, global_width);
    printf
      ("Exceeds original alignment plus length (%d,%d),(%d,%d)\n",
       mvector->master_align_row, mvector->master_align_col,
       mvector->master_length, mvector->master_width);
  }



  mvector->global_align_row = global_align_row;
  mvector->global_align_col = global_align_col;
  mvector->global_length = global_length;
  mvector->global_width = global_width;
  mvector->local_width = global_width;



  error += PLA_Temp_vector_local_length (template, 
					global_length, 
					global_align_row, 
					&local_len);

  mvector->local_length = local_len;



  error += PLA_Temp_vector_first_row_block   (template,
					     global_length,
					     global_align_row,
					     &(mvector->owner_top_block),
					     &(mvector->length_top_block));
  error += PLA_Temp_vector_last_row_block    (template,
					     global_length,
					     global_align_row,
					     &(mvector->owner_bottom_block),
					     &(mvector->length_bottom_block));
  mvector->width_left_block = 0;
  mvector->width_right_block  = 0;
  mvector->owner_left_block = 0;
  mvector->owner_right_block  = 0;


  /*
     Determine how much of the local vector to skip at the beginning.
     There is a detailed description of this in PLA_Matrix_submatrix.
     */

  if (local_len != 0) {
    error += PLA_Temp_vector_local_length (template, 
					   global_align_row -
					   mvector->master_align_row,
					   mvector->master_align_row,
					   &local_align);
    /* 
      Since multivectors are always created with global_align_col =0
      global_align_col should be the number of colums to skip at the
      beginning of our local buffer also.
    */

    offset = global_align_col * mvector->local_lld + local_align;
    
    MPI_Type_size (mvector->datatype, &data_size);
    offset *= (int) data_size;

    mvector->buffer = (void *) ( (char *)(mvector->master_buffer) + offset );

    if (mvector->buffer < mvector->master_buffer) {
      printf ("Error in PLA_Mvector_view_shift.\n");
      printf ("Computed beginning of buffer, %p, is less than\n",
	      mvector->buffer);
      printf ("actual beginning of buffer, %p\n", 
	      mvector->master_buffer);
    }
  }
  else 
    mvector->buffer = NULL;


  return(error);
}





















