

/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/



#include "PLA.h"


int PLA_Obj_objtype_cast(PLA_Obj                 laobject,
                           PLA_Obj_obj_type         object_type)
{
   int return_value = 0;
/* NOTE : NO check on object_type legitimacy due to the fact that */
/*        there may be PLA types not covered with PLA -- use at    */
/*        your own risk.                                          */

   laobject->obj_type = object_type;

/* Now the inevitable is starting to happen.  Of course you can't just
   change A to B and expect things to work.  If the block size of the
   top or bottom block is equal to the size of the object itself,
   then the object exists on only one row or column.  I would check
   the orientation to make sure which one I should really be checking
   but the caller might change orientation.  I think that the caller
   should have to set the duplication flag also, after all they are
   supposed to be checking orientation.
*/
   if (PLA_PVECTOR==object_type || PLA_PMVECTOR==object_type) {
      int error=0,
          length_top_block,
          owner_top_block,
          width_left_block,
          owner_left_block,
          global_length,
          global_width,
          global_align_row,
          global_align_col;
      error=PLA_Obj_left      (laobject,
                               &width_left_block,
                               &owner_left_block);
      if (0!=error&&error>return_value) return_value=error;
      error=PLA_Obj_top       (laobject,
                               &length_top_block,
                               &owner_top_block);
      if (0!=error&&error>return_value) return_value=error;
      error=PLA_Obj_global_all (laobject,
                                &global_length,
                                &global_width,
                                &global_align_row,
                                &global_align_col);
      if (0!=error&&error>return_value) return_value=error;
      if (length_top_block==global_length ||
          width_left_block==global_width) 
         laobject->annotate_flags.duplicate = PLA_NO_DUPLICATE;
   }

   return(return_value);
}


int PLA_Obj_swap (PLA_Obj *              aobj1,
		   PLA_Obj *              aobj2)
{
  PLA_Obj temp;
  temp=*aobj1; *aobj1=*aobj2; *aobj2=temp;
  return(0);
}

int PLA_Obj_init (PLA_Obj                aobj)
{
  aobj->obj_type  = -1;
  aobj->annotate_flags.panel_rowcol = PLA_PROJ_ONTO_COL;
  aobj->annotate_flags.duplicate    = PLA_NO_DUPLICATE;
  aobj->datatype                    = MPI_DOUBLE;
  aobj->views                       = PLA_NULL;
  aobj->master_buffer               = PLA_NULL;
  aobj->length_master_buffer_byte   = 0;
  aobj->buffer                      = PLA_NULL;
  aobj->master_length               = 0;
  aobj->master_width                = 0;
  aobj->master_align_row            = 0;
  aobj->master_align_col            = 0;
  aobj->local_stride                = 0;
  aobj->local_lld                   = 0;
  aobj->local_length                = 0;
  aobj->local_width                 = 0;
  aobj->global_length               = 0;
  aobj->global_width                = 0;
  aobj->template                    = PLA_NULL;
  aobj->global_align_row            = 0;
  aobj->global_align_col            = 0;
  aobj->length_top_block            = 0;
  aobj->owner_top_block             = 0;
  aobj->length_bottom_block         = 0;
  aobj->owner_bottom_block          = 0;
  aobj->width_left_block            = 0;
  aobj->owner_left_block            = 0;
  aobj->width_right_block           = 0;
  aobj->owner_right_block           = 0;
  aobj->api                         = PLA_NULL;
  return(0);
}

int PLA_Obj_check_type (PLA_Obj          aobj)
{
  char routine_name [ 32 ] = "PLA_Obj_check_type";
  if (aobj->obj_type != PLA_MSCALAR  &&
      aobj->obj_type != PLA_VECTOR  &&
      aobj->obj_type != PLA_MVECTOR &&
      aobj->obj_type != PLA_MATRIX  &&
      aobj->obj_type != PLA_PVECTOR &&
      aobj->obj_type != PLA_PMVECTOR)
  {
    PLA_Log_message ( routine_name, "Invalid object" );
    return ( PLA_RETURN_LOG_MESSAGE );
  }
  return ( PLA_SUCCESS );
}

int PLA_Obj_free (PLA_Obj * aobj )

{
  PLA_Obj object = NULL;
  char routine_name [ 32 ] = "PLA_Obj_free";
  extern struct PLA_Obj_def *PLA_GLOBAL_OBJ_STRUCT;
  extern int                *PLA_GLOBAL_OBJ_INDEX;
  extern int                *PLA_GLOBAL_VIEWS_STRUCT;
  extern int                *PLA_GLOBAL_VIEWS_INDEX;
  int obj_index;
  int view_index;
  int alloc_type, data_space_state;
  object = * aobj;

  if ( object == NULL ) return ( PLA_SUCCESS );

  if ( PLA_Obj_check_type ( object ) != PLA_SUCCESS )
  {
    PLA_Log_message ( routine_name, "Invalid object" );
    return ( PLA_RETURN_LOG_MESSAGE );
  }
    
/*  Since objects can have subobjects that point to the same buffer, we
    have to be careful about releasing the data buffer.  There is a
    counter to keep track of how many times this data buffer is pointeed
    to and we should only release this when the count gets to zero.
    
    There is a more bullet proof way described in the template object
    header.  It has the same problem that multiple things can point
    to the same template.
    */

  /* printf("Object views count is %5d\n", object->views[0]); */
  object->views[0]--;

  if ( object->views[0] == 0 )
  {
    PLA_Obj_alloc_type(*aobj, &alloc_type);
    if ( object->master_buffer )
       if(PLA_HEAP == alloc_type) /* If it was USER_SUPPLIED -- don't free it */
       {
           pla_free ( object->master_buffer );
       }
    view_index = (int*)(object->views)-(int*)(&PLA_GLOBAL_VIEWS_STRUCT[0]); /* index of view */
    PLA_GLOBAL_VIEWS_INDEX[view_index] = 0; /* mark unused */
    /* if ( object->views         ) pla_free ( object->views ); JAG 05/11/97 */
    object->master_buffer = NULL;
    object->views         = NULL;
  }
/* printf("Aobj is at %ld and the base is at %ld 1 at %ld\n", *aobj, &PLA_GLOBAL_OBJ_STRUCT[0], &PLA_GLOBAL_OBJ_STRUCT[1]); */
  obj_index = (PLA_Obj)(*aobj)-(PLA_Obj)(&PLA_GLOBAL_OBJ_STRUCT[0]); /* index of object */
  /* printf("Free Obj index is %5d\n", obj_index); */
  PLA_GLOBAL_OBJ_INDEX[obj_index] = 0; /* mark unused */
/*  pla_free ( object );  JAG 05/11/97 This stuff is now all in user-setaside -- reused */
  * aobj = NULL;
  return ( PLA_SUCCESS );
  
}


int PLA_Obj_objtype (PLA_Obj                aobj, 
		      PLA_Obj_obj_type *     type)

{
  if (0 != PLA_Obj_check_type (aobj)) 
    *type = PLA_UNDEFINED;
  else 
    *type = aobj->obj_type;
  return (0);
}




int PLA_Obj_datatype (PLA_Obj            laobject, 
		       MPI_Datatype *      datatype) 
{
  *datatype = laobject->datatype;
  return(0);
}

int PLA_Obj_global_length (PLA_Obj              laobject, 
			    int *                 global_length) 
{
  *global_length = laobject->global_length;
  return(0);
}

int PLA_Obj_global_width  (PLA_Obj              laobject, 
			    int *                 global_width) 
{
  *global_width = laobject->global_width;
  return(0);
}

int PLA_Obj_global_size   (PLA_Obj              laobject, 
			    int *                 global_size) 
{
  int   panel_type;
  int   error=0;

  error=PLA_Obj_project_onto (laobject, &panel_type);
  if (PLA_PROJ_ONTO_ROW != panel_type)
    PLA_Obj_global_width (laobject, global_size);
  else
    PLA_Obj_global_length(laobject, global_size);
  return(error);
}

int PLA_Obj_global_numvecs(PLA_Obj              laobject, 
			    int *                 global_numvecs) 
{
  int   panel_type;
  int   error=0;

  error=PLA_Obj_project_onto (laobject, &panel_type);
  if (PLA_PROJ_ONTO_ROW != panel_type)
    PLA_Obj_global_length(laobject, global_numvecs);
  else
    PLA_Obj_global_width (laobject, global_numvecs);
  return(error);
}


int PLA_Obj_local_buffer (PLA_Obj              laobject, 
		     void **               buffer) 
{
  *buffer = laobject->buffer;
  return(0);
}

int PLA_Obj_local_ldim (PLA_Obj           laobject,        
		  int *                    local_lld) 
{
  *local_lld = laobject->local_lld;
  return(0);
}

int PLA_Obj_local_stride (PLA_Obj        laobject,    
			   int *           local_stride) 
{
  *local_stride = laobject->local_stride;
  return(0);
}

int PLA_Obj_template (PLA_Obj            laobject, 
		       PLA_Template *       template) 
{
  *template = laobject->template;
  return(0);
}

int PLA_Obj_global_align (PLA_Obj    laobject,   
			  int *       global_align_row) 
{
  *global_align_row = laobject->global_align_row;
  return(0);
}

int PLA_Obj_global_align_row (PLA_Obj    laobject,   
			       int *       global_align_row) 
{
  *global_align_row = laobject->global_align_row;
  return(0);
}

int PLA_Obj_global_align_col (PLA_Obj    laobject,
			       int *       global_align_col) 
{
  *global_align_col = laobject->global_align_col;
  return(0);
}

int PLA_Obj_align_rowcol (PLA_Obj           laobject,   
			   int *              global_align_row,
			   int *              global_align_col) 
{
  *global_align_row = laobject->global_align_row;
  *global_align_col = laobject->global_align_col;
  return(0);
}

int PLA_Obj_local_length (PLA_Obj        laobject,   
			   int *           local_length) 
{
  *local_length = laobject->local_length;
  return(0);
}

int PLA_Obj_local_width (PLA_Obj        laobject,    
			  int *           local_width) 
{
  *local_width = laobject->local_width;
  return(0);
}



int PLA_Obj_local_info(PLA_Obj          obj,
                         int               *local_length,
                         int               *local_width,
                         void             **local_buffer,
                         int               *local_stride,
                         int               *local_ldim)
{
   MPI_Datatype      datatype;

   int return_value = 0;

   return_value += PLA_Obj_local_all(obj, &datatype, local_length, local_width,
                                      local_buffer, local_ldim);
   return_value += PLA_Obj_local_stride(obj, local_stride);

   return(return_value);
}


int PLA_Obj_local_all   (PLA_Obj        laobject,
			  MPI_Datatype *  datatype,
			  int *           local_length,
			  int *           local_width,
			  void **         buffer,
			  int *           local_lld)
{
  int error = 0;
  error += PLA_Obj_datatype       (laobject, datatype); 
  error += PLA_Obj_local_length   (laobject, local_length);
  error += PLA_Obj_local_width    (laobject, local_width);
  error += PLA_Obj_local_buffer         (laobject, buffer); 
  error += PLA_Obj_local_ldim      (laobject, local_lld); 
  return(error);
}


int PLA_Obj_global_info(PLA_Obj         obj,
                         int               *global_length,
                         int               *global_width,
                         int               *project_onto,
                         int               *global_owner_row,
                         int               *global_owner_col,
                         int               *global_align_row,
                         int               *global_align_col)
{
   int return_value = 0;

   return_value += PLA_Obj_project_onto (obj,  project_onto);
   return_value += PLA_Obj_owner_row       (obj,  global_owner_row);
   return_value += PLA_Obj_owner_col       (obj,  global_owner_col);
   return_value += PLA_Obj_global_all(obj, global_length, global_width,
                                      global_align_row, global_align_col);

   return(return_value);
}

int PLA_Obj_global_all   (PLA_Obj       laobject,
			   int *          global_length,
			   int *          global_width,
			   int *          global_align_row,
			   int *          global_align_col)
{
  int error = 0;
  PLA_Obj_global_length            (laobject, global_length); 
  PLA_Obj_global_width             (laobject, global_width);
  PLA_Obj_global_align_row         (laobject, global_align_row);
  PLA_Obj_global_align_col         (laobject, global_align_col);
  return(error);
}

int PLA_Obj_top  (PLA_Obj   laobject,
		   int *      length_top_block,
		   int *      owner_top_block)
{
  *length_top_block = laobject->length_top_block;
  *owner_top_block  = laobject->owner_top_block;
  return(0);
}

int PLA_Obj_bottom  (PLA_Obj   laobject,
		      int *      length_bottom_block,
		      int *      owner_bottom_block)
{
  *length_bottom_block = laobject->length_bottom_block;
  *owner_bottom_block  = laobject->owner_bottom_block;
  return(0);
}

int PLA_Obj_left  (PLA_Obj   laobject,
		    int *      width_left_block,
		    int *      owner_left_block)
{
  *width_left_block  = laobject->width_left_block;
  *owner_left_block  = laobject->owner_left_block;
  return(0);
}

int PLA_Obj_right  (PLA_Obj   laobject,
		     int *      width_right_block,
		     int *      owner_right_block)
{
  *width_right_block = laobject->width_right_block;
  *owner_right_block  = laobject->owner_right_block;
  return(0);
}

int PLA_Obj_top_left  (PLA_Obj   laobject,
			int *      length_top_left_block,
			int *      owner_row_top_left_block,
			int *      owner_col_top_left_block)
{
  int length_top_block, length_left_block, error=0;
  error += PLA_Obj_top   (laobject,
			   &length_top_block,
			   owner_row_top_left_block);
  error += PLA_Obj_left  (laobject,
			   &length_left_block,
			   owner_col_top_left_block);
  *length_top_left_block = 
    ( length_top_block < length_left_block ) ?
      length_top_block : length_left_block ;
  return(error);
}

int PLA_Obj_top_right  (PLA_Obj   laobject,
			 int *      length_top_right_block,
			 int *      owner_row_top_right_block,
			 int *      owner_col_top_right_block)
{
  int length_top_block, length_right_block, error=0;
  error += PLA_Obj_top   (laobject,
			   &length_top_block,
			   owner_row_top_right_block);
  error += PLA_Obj_right (laobject,
			   &length_right_block,
			   owner_col_top_right_block);
  *length_top_right_block = 
    ( length_top_block < length_right_block ) ?
      length_top_block : length_right_block ;
  return(error);
}


int PLA_Obj_bottom_left  (PLA_Obj   laobject,
			   int *      length_bottom_left_block,
			   int *      owner_row_bottom_left_block,
			   int *      owner_col_bottom_left_block)
{
  int length_bottom_block, length_left_block, error=0;
  error += PLA_Obj_bottom (laobject,
			    &length_bottom_block,
			    owner_row_bottom_left_block);
  error += PLA_Obj_left   (laobject,
			    &length_left_block,
			    owner_col_bottom_left_block);
  *length_bottom_left_block = 
    ( length_bottom_block < length_left_block ) ?
      length_bottom_block : length_left_block ;
  return(error);
}

int PLA_Obj_bottom_right  (PLA_Obj   laobject,
			    int *      length_bottom_right_block,
			    int *      owner_row_bottom_right_block,
			    int *      owner_col_bottom_right_block)
{
  int length_bottom_block, length_right_block, error=0;
  error += PLA_Obj_bottom (laobject,
			    &length_bottom_block,
			    owner_row_bottom_right_block);
  error += PLA_Obj_right  (laobject,
			    &length_right_block,
			    owner_col_bottom_right_block);
  *length_bottom_right_block = 
    ( length_bottom_block < length_right_block ) ?
      length_bottom_block : length_right_block ;
  return(error);
}


int PLA_Obj_owner_row     (PLA_Obj           laobject,
			   int              *owner_row)
{
  int objtype, proj_onto;
  PLA_Obj_objtype ( laobject, & objtype );
  PLA_Obj_project_onto ( laobject, & proj_onto );
  if ( objtype == PLA_PMVECTOR && proj_onto == PLA_PROJ_ONTO_ROW )
    * owner_row = laobject->annotate_flags.duplicate;
  else if ( objtype == PLA_MSCALAR )
    * owner_row = laobject->owner_top_block; 
  else
    * owner_row = PLA_ALL_ROWS;

  return ( PLA_SUCCESS );
}

int PLA_Obj_owner_col     (PLA_Obj           laobject,
			   int              *owner_col)
{
  int objtype, proj_onto;
  PLA_Obj_objtype ( laobject, & objtype );
  PLA_Obj_project_onto ( laobject, & proj_onto );
  if ( objtype == PLA_PMVECTOR && proj_onto == PLA_PROJ_ONTO_COL )
    * owner_col = laobject->annotate_flags.duplicate;
  else if ( objtype == PLA_MSCALAR )
    * owner_col = laobject->owner_left_block;
  else
    * owner_col = PLA_ALL_COLS;
  return ( PLA_SUCCESS );
}

int PLA_Obj_project_onto     (PLA_Obj           laobject,
			          PLA_Panel_rowcol_flag *panel_type)
{
  *panel_type = laobject->annotate_flags.panel_rowcol;
  return (0);
}


int PLA_Obj_set_orientation      (PLA_Obj   laobject,
				       int  panel_rowcol)
{
  if (panel_rowcol == PLA_UNSPECIFIED_PANEL ||
      panel_rowcol == PLA_PROJ_ONTO_ROW         ||
      panel_rowcol == PLA_PROJ_ONTO_COL)
    laobject->annotate_flags.panel_rowcol = panel_rowcol;
  else {
    printf 
      ("Invalid PLA_Panel_rowcol_flag passed to PLA_Obj_set_orientation\n");
    return (1);
  }
  return (0);
}

int PLA_Obj_get_orientation      (PLA_Obj   laobject,
				  int      *panel_rowcol)
{
  *panel_rowcol = laobject->annotate_flags.panel_rowcol;
  return (0);
}

int PLA_Obj_annotate_duplicate    (PLA_Obj         laobject, 
                                    PLA_Duplicate_flag *duplicate)
{
  *duplicate = laobject->annotate_flags.duplicate;
   return(0);
}


