

/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/



#include "PLA.h"


int PLA_Obj_view_swap (PLA_Obj  *obj1,  PLA_Obj  *obj2)
{
PLA_Obj temp;
temp=*obj1; *obj1=*obj2; *obj2=temp;
return(0);
}

int PLA_Obj_view_all(PLA_Obj   old_obj,
                   PLA_Obj   *new_obj) /* new view */
{
int return_value = 0;

return_value = PLA_Obj_view(old_obj,
                            PLA_DIM_ALL,
                            PLA_DIM_ALL,
                            PLA_ALIGN_FIRST,
                            PLA_ALIGN_FIRST,
                            new_obj);
return(return_value);
}

/* And now . . . for a really sloppy wrapper routine . . . */
/* I am not sure what has been done yet as far as replacing the type */
/* PLA_PVECTOR with PLA_PVECTOR etc.  This routine is built to handle    */
/* the old PLA_PVECTOR type but will have to be modified a bit to handle*/
/* the PROJECTED stuff.     John Gunnels                             */
/* Naming wise, I mean -- I think the group needs to talk about this.*/

int PLA_Obj_view(PLA_Obj   old_obj,
                   int         global_length,
                   int         global_width,
                   int         align_row,
                   int         align_col,
                   PLA_Obj   *new_obj) /* new view */
{
   int return_value = 0;
   PLA_Obj_obj_type Object_type;

   return_value += PLA_Obj_objtype(old_obj, &Object_type);
   if(Object_type == PLA_PVECTOR || Object_type == PLA_PMVECTOR)
   {
      return_value += PLA_Panel_subpanel(old_obj, global_length, global_width,
                                        align_row, align_col, new_obj);
   }
   else if(Object_type == PLA_MATRIX)
   {
      return_value += 
              PLA_Matrix_submatrix(old_obj, global_length, global_width,
                                          align_row, align_col, new_obj);
   }
   else if (Object_type == PLA_MVECTOR)
   {
      return_value += 
             PLA_Mvector_submvector(old_obj, global_length, global_width,
                                            align_row, align_col, new_obj);
   }
   else if (Object_type == PLA_VECTOR)
   {
      return_value += PLA_Vector_subvector(old_obj, global_length,
                                          align_row, new_obj);
   }
   else if (Object_type == PLA_MSCALAR)
   {
      return_value += 
            PLA_Mscalar_submscalar(old_obj, global_length, global_width,
                                            align_row, align_col, new_obj);
   }
   else
      return_value += UNKNOWN_OBJECT_TYPE;
   return(return_value);
}


int PLA_Obj_horz_split_2           (PLA_Obj         laobject,
				    int             length,
				    PLA_Obj        *upper_obj,
				    PLA_Obj        *lower_obj)
     
{
  int           error = 0;
  int           size_bottom, size_top;

  /*
     Split the object in two.  Lengths do not start counting with the
     zero_or_one stuff.  So size_bottom + size_top = total.
     But because I want to move the bottom up for the upper_obj
     and the shift routines need a negative number to move the 
     bottom up, size_top=total+size_bottom.
     */
  if (0<=length) {
    size_bottom  = -(laobject->global_length - length);
    size_top     = length;
  }
  else {
    size_top     = laobject->global_length + length;
    size_bottom  = length;
  }
  if (0<size_bottom) {
    char message[128];
    sprintf (message,
	     "Error in PLA_Obj_hor_split_2: requested split size %d",
	     size_top);
    MMPI_Log_message ("PLA_Obj_horz_split_2",message);
    sprintf (message,
	     "exceeds length of object:%d",laobject->global_length );
    MMPI_Log_message ("PLA_Obj_horz_split_2",message);
    error = 200101;
    return(error);
  }

  /* 
     Split in half with the upper half being size size_top. 
     This will create two subthings.  It is quite possible that
     the upper part of the split is so small that it will exist
     only on one processor row.  This is fine.  The other processors
     will still get an object allocated, the buffer will just 
     point to NULL
     */
  
  if (PLA_DUMMY !=  upper_obj &&
      laobject  != *upper_obj  ) 
    error+=PLA_Obj_view_all(laobject, upper_obj);

  if (PLA_DUMMY !=  lower_obj &&
      laobject  != *lower_obj  )
    error+=PLA_Obj_view_all(laobject, lower_obj);

  if (PLA_DUMMY !=  upper_obj &&
      PLA_DUMMY !=  lower_obj &&
      *upper_obj == *lower_obj ) {
    char *message1=
      "Output upper and lower objects point to the same object.";
    char *message2=
      "Two seperate objects or PLA_DUMMY must be specified.";
    MMPI_Log_message ("PLA_Obj_horz_split_2", message1);
    MMPI_Log_message ("PLA_Obj_horz_split_2", message2);
    return (200120);
  }

  if (PLA_DUMMY != upper_obj)
    error+= PLA_Obj_view_shift(*upper_obj,
			       0,
			       0,  0,
			       size_bottom);
  if (PLA_DUMMY != lower_obj)
    error+= PLA_Obj_view_shift(*lower_obj,
			       size_top,
			       0,   0,
			       0);
  return(error);
}


int PLA_Obj_vert_split_2           (PLA_Obj         laobject,
				    int             width,
				    PLA_Obj        *left_obj,
				    PLA_Obj        *right_obj)
     
{
  int           error = 0;
  int           right_width, left_width;

   /*
     Split the object in two.  Widths do not start counting with the
     zero_or_one stuff.  So right + left = total.
     But because I want to move the right boundry of the left object
     to the left, I need to make that a negative number for the
     shift routines.
     */ 
  if (0<=width) {
    right_width      = -(laobject->global_width - width);
    left_width       = width;
  }
  else {
    left_width       = laobject->global_width + width;
    right_width      = width;
  }

  if (0<right_width) {
    char message[128];
    sprintf (message, "requested split size %d", left_width);
    MMPI_Log_message ("PLA_Obj_vert_split_2", message);
    sprintf (message,
	     "exceeds width of object:%d",laobject->global_width );
    MMPI_Log_message ("PLA_Obj_vert_split_2", message);
    error = 200102;
    return(error);
  }

  /* 
     Split in half with the left half being size left_width. 
     This will create two subthings.  It is quite possible that
     the part of the split is so small that it will exist
     only on one processor row/col.  This is fine.  The other processors
     will still get an object allocated, the buffer will just 
     point to NULL
     */
  
  if (PLA_DUMMY !=  left_obj &&
      laobject  != *left_obj  )
    error+=PLA_Obj_view_all(laobject, left_obj);
  
  if (PLA_DUMMY !=  right_obj &&
      laobject  != *right_obj  )
    error+=PLA_Obj_view_all(laobject, right_obj);
  
  if (PLA_DUMMY != left_obj &&
      PLA_DUMMY != right_obj &&
      *right_obj == *left_obj ) {
    char *message1=
      "Output left and right objects point to the same object.";
    char *message2=
      "Two seperate objects or PLA_DUMMY must be specified.";
    MMPI_Log_message ("PLA_Obj_vert_split_2", message1);
    MMPI_Log_message ("PLA_Obj_vert_split_2", message2);
    return (200121);
  }

  if (PLA_DUMMY != left_obj)
    error+= PLA_Obj_view_shift(*left_obj,
                               0,
                               0,  right_width,
                               0);
  if (PLA_DUMMY != right_obj)
    error+= PLA_Obj_view_shift(*right_obj,
                               0,
                               left_width,   0,
                               0);
  return(error);
}




int PLA_Obj_split_4           (PLA_Obj         laobject,
			       int              length,
			       int              width,
			       PLA_Obj        *upper_left_obj,
			       PLA_Obj        *upper_right_obj,
			       PLA_Obj        *lower_left_obj,
			       PLA_Obj        *lower_right_obj)
     
{
  char         *first_op=NULL, *second_op=NULL;
  int           error = 0;
  int           lower_length, right_width;
  int           upper_length, left_width;

  /*
     Split the object in two.  Lengths do not start counting with the
     zero_or_one stuff.  So lower + upper = total.
     We have to be careful because the shift routines sometimes need
     a negative and sometimes a positive.  This is complicated
     by the fact that the left edge of the right block is shifted
     right.  See the shift_2 routines for details.
     */  
  if (0<=length) {
    lower_length     = -(laobject->global_length - length);
    upper_length     = length;
  }
  else {
    upper_length     = laobject->global_length + length;
    lower_length     = length;
  }
  if (0<=width) {
    right_width      = -(laobject->global_width - width);
    left_width       = width;
  }
  else {
    left_width      = laobject->global_width + width;
    right_width     = width;
  }

  if (0<lower_length) {
    char message[128];
    sprintf (message, "Requested split size %d", upper_length);
    MMPI_Log_message ("PLA_Obj_split_4",message);
    sprintf (message,
             "exceeds length of object:%d",laobject->global_length );
    MMPI_Log_message ("PLA_Obj_split_4",message);
    error = 200101;
    return(error);
  }
  if (0<right_width) {
    char message[128];
    sprintf (message, "Requested split size %d", left_width);
    MMPI_Log_message ("PLA_Obj_split_4",message);
    sprintf (message,
             "exceeds length of object:%d",laobject->global_length );
    MMPI_Log_message ("PLA_Obj_split_4",message);
    error = 200102;
    return(error);
  }

  
  if (PLA_DUMMY !=  upper_left_obj &&
      laobject  != *upper_left_obj  )
    error+=PLA_Obj_view_all(laobject, upper_left_obj);
  
  if (PLA_DUMMY !=  upper_right_obj &&
      laobject  != *upper_right_obj  )
    error+=PLA_Obj_view_all(laobject, upper_right_obj);
  
  if (PLA_DUMMY !=  lower_left_obj &&
      laobject  != *lower_left_obj  )
    error+=PLA_Obj_view_all(laobject, lower_left_obj);
  
  if (PLA_DUMMY !=  lower_right_obj &&
      laobject  != *lower_right_obj  )
    error+=PLA_Obj_view_all(laobject, lower_right_obj);

  if (error) return(error);


  /* Do some long error checking.  I don't want any two of the
     output matrixies to be the same.  It would be a lot worse
     if there were nine output parameters.
     */
  if (PLA_DUMMY != upper_left_obj) { 
    first_op = "upper left object";
    second_op= NULL;
    if (PLA_DUMMY != upper_right_obj &&
	*upper_left_obj == *upper_right_obj)
      second_op = "upper_right_obj";
    else
      if (PLA_DUMMY != lower_left_obj &&
	  *upper_left_obj == *lower_left_obj) 
	second_op = "lower left obj";
      else
	if (PLA_DUMMY != lower_right_obj &&
	    *upper_left_obj == *lower_right_obj)
	  second_op = "lower_right_obj";
    if (NULL!=second_op) error=200121;
  }
  if (PLA_DUMMY != upper_right_obj) {
    first_op = "upper right object";
    second_op= NULL;
    if (PLA_DUMMY != lower_left_obj &&
	*upper_right_obj == *lower_left_obj)
      second_op = "lower left obj";
    else
      if (PLA_DUMMY != lower_right_obj &&
	  *upper_right_obj == *lower_right_obj)
	second_op = "lower_right_obj";
    if (NULL!=second_op) error = 200121;
  }
  if (PLA_DUMMY != lower_left_obj &&
      PLA_DUMMY != lower_right_obj &&
      *lower_left_obj == *lower_right_obj) {
    first_op = "lower left object";
    error    = 200121;
    second_op = "lower_right_obj";
  }
  if (error) {
    char *message1=
      "Output left and right objects point to the same object.";
    char *message2=
      "Two seperate objects or PLA_DUMMY must be specified.";
    char  message[128];
    sprintf (message, "%s is the same as the %s object",
	     first_op, second_op);
    MMPI_Log_message ("PLA_Obj_vert_split_2", message1);
    MMPI_Log_message ("PLA_Obj_vert_split_2", message2);
    MMPI_Log_message ("PLA_Obj_vert_split_2", message);
    return (error);
  }


  /* I don't call the split_2 routines because I don't know which 
     of the output objects are just PLA_DUMMY and I don't want to
     have to go through a big case statement to find out.
     */

  if (PLA_DUMMY != upper_left_obj)
    error+= PLA_Obj_view_shift(*upper_left_obj,
                               0,
                               0,  right_width,
                               lower_length);
  if (PLA_DUMMY != upper_right_obj)
    error+= PLA_Obj_view_shift(*upper_right_obj,
                               0,
                               left_width,   0,
                               lower_length);   
  if (PLA_DUMMY != lower_left_obj)
    error+= PLA_Obj_view_shift(*lower_left_obj,
                               upper_length,
                               0,  right_width,
                               0);
  if (PLA_DUMMY != lower_right_obj)
    error+= PLA_Obj_view_shift(*lower_right_obj,
                               upper_length,
                               left_width,   0,
                               0);
  return(error);
}


/* I believe that the following routine is fairly simple to write -- it just */
/* isn't around -- (that I can find) at present.                             */
int PLA_Obj_set_local_contents(int           transpose,
                                 int            rows_in_buffer,
                                 int            cols_in_buffer,
                                 void          *buffer,
                                 int            leading_dim_buffer,
                                 int            stride_buffer,
                                 PLA_Obj       laobject)
{
   int return_value = 0;

   printf("PLA_Obj_set_local_contents is not yet implemented\n");
   printf("Please e-mail gunnels@cs.utexas.edu if you see this message.\n");
   return_value += UNIMPLIMENTED_FUNCTION;
   return(return_value);
}


int PLA_Obj_view_shift(PLA_Obj       laobject,
                         int             length_top,
                         int             width_left,
                         int             width_right,
                         int             length_bottom)
{
   int return_value = 0;
   
   switch (laobject->obj_type)
   {
      case PLA_MATRIX :
         return_value = PLA_Matrix_view_shift (laobject,
                                              length_top,
                                              width_left,
                                              width_right,
                                              length_bottom);
      break;
      case PLA_MSCALAR :
         return_value = PLA_Mscalar_view_shift (laobject,
                                                length_top,
                                                width_left,
                                                width_right,
                                                length_bottom);
      break;
      case PLA_VECTOR :
         return_value = PLA_Vector_view_shift  (laobject,
                                                length_top,
                                                width_left,
                                                width_right,
                                                length_bottom);
      break;
      case PLA_MVECTOR :
         return_value = PLA_Mvector_view_shift (laobject,
                                                length_top,
                                                width_left,
                                                width_right,
                                                length_bottom);
      break;
      case PLA_PVECTOR :
      case PLA_PMVECTOR :
         return_value = PLA_Panel_view_shift   (laobject,
                                                length_top,
                                                width_left,
                                                width_right,
                                                length_bottom);
      break;
      default: {
         char message[128];
         sprintf (message,"Unrecognized object type: %d", laobject->obj_type);
         MMPI_Log_message ("PLA_Obj_view_shift",message);
         return_value = 200100;
       }
   }
   return(return_value);
}


int PLA_Obj_split_size(PLA_Obj      laobject,
                         int            side,
                         int           *size,
                         int           *owner)
{
   int return_value = 0;

   switch(side)
   {
      case PLA_SIDE_TOP :    return_value += PLA_Obj_top(laobject, size, owner);
                             break;
      case PLA_SIDE_LEFT :   return_value += PLA_Obj_left(laobject, size, owner);
                             break;
      case PLA_SIDE_RIGHT  : return_value += PLA_Obj_right(laobject, size, owner);
                             break;
      case PLA_SIDE_BOTTOM : return_value += PLA_Obj_bottom(laobject, size, owner);
                             break;
      default : return_value += UNKNOWN_SIDE;
                printf("PLA_Obj_split_size received an unknown 'side' as a\n");
                printf("parameter in a call.\n");
   }
   return(return_value);
}


int PLA_Obj_get_local_contents(PLA_Obj      laobject,
                                 int            transpose,
                                 int           *rows_in_buffer,
                                 int           *cols_in_buffer,
                                 void          *buffer,
                                 int            leading_dim_buffer,
                                 int            stride_buffer)
{
   int return_value = 0;

   if(transpose == PLA_NO_TRANS)
   {
      return_value += PLA_Obj_extract_contents(laobject, leading_dim_buffer,
                               stride_buffer, rows_in_buffer, cols_in_buffer,
                               buffer);
   }
   else /* Assume PLA_TRANS */
      return_value += 
            PLA_Obj_extract_transpose_contents(laobject, leading_dim_buffer,
                               stride_buffer, rows_in_buffer, cols_in_buffer,
                               buffer);

   return(return_value);
}


int PLA_Obj_extract_contents      (PLA_Obj         laobject,
				    int              leading_dim_ret_buf,
				    int              stride_ret_buf,
				    int  *           rows_in_ret_buf,
				    int  *           cols_in_ret_buf,
				    void *           return_buffer)
{
  int           error = 0, one=1;
  int           length, width, stride, datasize, number_bytes, lda;
  int           i,j,k, inc_old_buffer, inc_return_buffer;
  void         *buffer, *temp_buf1, *temp_buf2;
  PLA_Template   template;
  PLA_Start_num  zero_or_one;
  MPI_Datatype  datatype;
  PLA_Obj_obj_type object_type;
  
  error +=  PLA_Obj_objtype (laobject, 
                              &object_type);
  if (error!=0) return(error);
  
  template = laobject->template;
  error += PLA_Temp_zero_or_one (template, &zero_or_one);
  
  datatype    = laobject->datatype;
  buffer      = laobject->buffer;
  length      = laobject->local_length;
  width       = laobject->local_width;
  stride      = laobject->local_stride;
  lda         = laobject->local_lld;
/* Note that the compiler on eureka.cc -- and I guess a lot of      */
/* Intel Paragons (?) will flag the following as a compiler warning */
/* IF you want to make it happy, you can declare datasize to be of  */
/* type MPI_Aint.  If you trust that eventually the MPI standard    */
/* will be observed, leave it as is.                                */
  MPI_Type_size (datatype, &datasize);

  /* Now comes the hard part.  There might be other ways to do this.  One
     would be to make the main swithch statement dependent on if there
     is a stride or not a stride.  Even a matrix could be copied as 
     one array if the leading dimention were the same as the length.
     But splitting it up as object type I think is more logical.  
     The code is long but that is partly because I put each element
     of a calling sequence on a seperate line and that I try to 
     do the exceptional case when the datasize is unusual, say 
     this was character data.
     */
  switch (object_type) {
  case PLA_VECTOR: {

    *rows_in_ret_buf = length;
    *cols_in_ret_buf = 1;
    
    if (datatype == MPI_DOUBLE           || 
	datasize == sizeof(double))
      PLA_dcopy (&length, 
	      buffer, 
	      &stride, 
	      return_buffer, 
	      &stride_ret_buf);
    else if (datatype == MPI_FLOAT       || 
	     datasize == sizeof(float))
      PLA_scopy (&length, 
	      buffer, 
	      &stride, 
	      return_buffer, 
	      &stride_ret_buf);

    /* 2*sizeof(float) would be single precision complex, FORTRAN
       */
    else if (datasize == 2*sizeof(float))
      PLA_ccopy (&length, 
	      buffer, 
	      &stride, 
	      return_buffer, 
	      &stride_ret_buf);

    /* 2*sizeof(double) is double precision complex.
       */
    else if (datasize == 2*sizeof(double))
      PLA_zcopy (&length, 
	      buffer, 
	      &stride, 
	      return_buffer, 
	      &stride_ret_buf);
    
    /* Else things get complicated.  This case will surly never happen
       since all numberic data is taken care of in the first 4 cases
       in fact I could issue an error message at this point. 
       But instead I try to muddle on the best I can.
       */
    else {
      if (stride == 1) {
	number_bytes = datasize * length;
	memcpy (return_buffer, buffer, number_bytes);
      }
      else {
	inc_return_buffer = stride*datasize;
	for (i=0; i<length; i++) {
	  memcpy (return_buffer, buffer, datasize);
	  return_buffer = (char *) return_buffer + datasize;
	  buffer        = (char *) buffer + inc_return_buffer;
	}
      }
    }
    break;
  }
  case PLA_MVECTOR: {
    
    /* Mutlivectors are a real pain.  That is because they have strides
       like vectors and leading dimentions like matrixies. They have the
       worse of both worlds and are the hardest case to take care of. 
       In general they need to be copied one column at a time.  
       
       I am assuming multivectors will not happen often.  There are special
       cases where the lacopy routines should be used in stead of what
       is done now.  Something to add later.
       */
    *rows_in_ret_buf = length;
    *cols_in_ret_buf = width;
    
    /* There is a special flag that just says copy it into the 
       return buffer as a continuous block of data.
       */
    if (leading_dim_ret_buf == PLA_LEADING_DIM_TIGHT_FIT)
      leading_dim_ret_buf = length;

    if (datatype == MPI_DOUBLE           || 
	datasize == sizeof(double))
      if (1 == stride_ret_buf  &&
	  1 == stride) {
	PLA_dlacpy ("General",
		 &length, 
		 &width,
		 buffer, 
		 &lda, 
		 return_buffer, 
		 &leading_dim_ret_buf);
      }
      else { 
	for (i=0; i<width; i++) {
	  PLA_dcopy (&length, 
		  buffer, 
		  &stride, 
		  return_buffer, 
		  &stride_ret_buf);
	  return_buffer = (double *) return_buffer + leading_dim_ret_buf;
	  buffer        = (double *) buffer + lda;
	}
      }
    else if (datatype == MPI_FLOAT       || 
	     datasize == sizeof(float))
      if (1 == stride_ret_buf  &&
	  1 == stride) {
	PLA_slacpy ("General",
		 &length, 
		 &width,
		 buffer, 
		 &lda, 
		 return_buffer, 
		 &leading_dim_ret_buf);
      }
      else {
	for (i=0; i<width; i++) {
	  PLA_scopy (&length, 
		  buffer, 
		  &stride, 
		  return_buffer, 
		  &stride_ret_buf);
	  return_buffer = (float *) return_buffer + leading_dim_ret_buf;
	  buffer        = (float *) buffer + lda;
	}
      }
    /* 2*sizeof(float) would be single precision complex, FORTRAN
       */
    else if (datasize == 2*sizeof(float))
      if (1 == stride_ret_buf  &&
	  1 == stride) {
	PLA_clacpy ("General",
		 &length, 
		 &width,
		 buffer, 
		 &lda, 
		 return_buffer, 
		 &leading_dim_ret_buf);
      }
      else  {		 
	for (i=0; i<width; i++) {
	  PLA_ccopy (&length, 
		  buffer, 
		  &stride, 
		  return_buffer, 
		  &stride_ret_buf);
	  return_buffer = (float *) return_buffer + 2*leading_dim_ret_buf;
	  buffer        = (float *) buffer + 2*lda;
	}
      }
    /* 2*sizeof(double) is double precision complex.
     */
    else if (datasize == 2*sizeof(double))
      if (1 == stride_ret_buf  &&
	  1 == stride) {
	PLA_zlacpy ("General",
		 &length, 
		 &width,
		 buffer, 
		 &lda, 
		 return_buffer, 
		 &leading_dim_ret_buf);
      }
      else {
	for (i=0; i<width; i++) {
	  PLA_zcopy (&length, 
		  buffer, 
		  &stride, 
		  return_buffer, 
		  &stride_ret_buf);
	  return_buffer = (double *) return_buffer + 2*leading_dim_ret_buf;
	  buffer        = (double *) buffer + 2*lda;
	}
      }
    /*  Highly suspect code.  I am pretty sure the above works, but
	not about this.
	*/
    else {
      if (stride == 1) {
	number_bytes = datasize * length;
	inc_return_buffer =  leading_dim_ret_buf * datasize;
	inc_old_buffer    =  lda * datasize;
	for (i=0; i<width; i++) {
	  memcpy (return_buffer, buffer, number_bytes);
	  return_buffer = (char *) return_buffer + inc_return_buffer;
	  buffer        = (char *) buffer        + inc_old_buffer;
	}
      }
      else {
	inc_return_buffer = stride*datasize;
	for (j=0; j<width; j++) {
	  temp_buf1 = buffer;
	  temp_buf2 = return_buffer;
	  for (i=0; i<length; i++) {
	    memcpy (temp_buf2, temp_buf1, datasize);
	    temp_buf2 = (char *) temp_buf2 + datasize;
	    temp_buf1 = (char *) temp_buf1 + inc_return_buffer;
	  }
	  return_buffer = (char *) return_buffer+leading_dim_ret_buf*datasize;
	  buffer = (char *) buffer + lda * datasize;
	}
      }
    }
    break;
  }
  case PLA_MSCALAR:
  case PLA_MATRIX:
  case PLA_PVECTOR:  
  case PLA_PMVECTOR: {

    *rows_in_ret_buf = length;
    *cols_in_ret_buf = width;
    
    if (leading_dim_ret_buf == PLA_LEADING_DIM_TIGHT_FIT)
      leading_dim_ret_buf = length;

    if (datatype == MPI_DOUBLE ||
	datasize == sizeof(double))
      PLA_dlacpy ("General",
	       &length, 
	       &width,
	       buffer, 
	       &lda, 
	       return_buffer, 
	       &leading_dim_ret_buf);
    else if (datatype == MPI_FLOAT       || 
	     datasize == sizeof(float))
      PLA_slacpy ("General",
	       &length, 
	       &width,
	       buffer, 
	       &lda, 
	       return_buffer, 
	       &leading_dim_ret_buf);
    else if (datasize == 2*sizeof(float))
      PLA_clacpy ("General",
	       &length, 
	       &width,
	       buffer, 
	       &lda, 
	       return_buffer, 
	       &leading_dim_ret_buf);
    else if (datasize == 2*sizeof(double))
      PLA_zlacpy ("General",
	       &length, 
	       &width,
	       buffer, 
	       &lda, 
	       return_buffer, 
	       &leading_dim_ret_buf);
    else {
      number_bytes = datasize * length;
      inc_return_buffer =  leading_dim_ret_buf * datasize;
      inc_old_buffer    =  lda * datasize;
      for (i=0; i<width; i++) {
	memcpy (return_buffer, buffer, number_bytes);
	return_buffer = (char *) return_buffer + inc_return_buffer;
	buffer        = (char *) buffer        + inc_old_buffer;
      }
    }
    break;
  }
  default: {
    printf (" Invalid PLA_Object passed to PLA_Obj_extract_contents\n");
    printf (" Object must be a Matrix, vector, multivector, multiscalar\n");
    error = 1;
    return(error);
  }
  }
  return(error);
}


int PLA_Obj_extract_transpose_contents      (PLA_Obj   laobject,
					      int        leading_dim_ret_buf,
					      int        stride_ret_buf,
					      int  *     rows_in_ret_buf,
					      int  *     cols_in_ret_buf,
					      void *     return_buffer)
{
  int           error = 0, one=1;
  int           length, width, stride, datasize, number_bytes, lda;
  int           i,j,k, inc_old_buffer, inc_return_buffer;
  void         *buffer, *temp_buf1, *temp_buf2;
  PLA_Template   template;
  PLA_Start_num  zero_or_one;
  MPI_Datatype  datatype;
  PLA_Obj_obj_type object_type;
  
  error +=  PLA_Obj_objtype (laobject, 
                              &object_type);
  if (error!=0) return(error);
  
  length      = laobject->local_length;
  if (leading_dim_ret_buf == PLA_LEADING_DIM_TIGHT_FIT)
    leading_dim_ret_buf = length;

  /*  Now this is getting tricky.  Vectors are easy since the transpose
      is not allowed.  
      The rest are harder.
      What I do is call the regular case with the stride and leading
      dimention exchanged.  This is the same as doing a transpose. 
      It means that I do not have to duplicate the entire code for the
      Multivector code from the nontransposd case and copy it here.  

      It also means that the code is nonobvious.  
      */

  switch (object_type) {
  case PLA_MVECTOR: {
    error += PLA_Obj_extract_contents  (laobject,
					 stride_ret_buf,
					 leading_dim_ret_buf,
					 rows_in_ret_buf,
					 cols_in_ret_buf,
					 return_buffer);
    break;
  }
  case PLA_MSCALAR: {
    stride_ret_buf = 1;
    laobject->obj_type = PLA_MVECTOR;
    error += PLA_Obj_extract_contents  (laobject,
					 stride_ret_buf,
					 leading_dim_ret_buf,
					 rows_in_ret_buf,
					 cols_in_ret_buf,
					 return_buffer);
    laobject->obj_type = PLA_MSCALAR;
    break;
  }
  case PLA_MATRIX: {
    stride_ret_buf = 1;
    laobject->obj_type = PLA_MVECTOR;
    error += PLA_Obj_extract_contents  (laobject,
					 stride_ret_buf,
					 leading_dim_ret_buf,
					 rows_in_ret_buf,
					 cols_in_ret_buf,
					 return_buffer);
    laobject->obj_type = PLA_MATRIX;
    break;
  }
  case PLA_PVECTOR: {
    stride_ret_buf = 1;
    laobject->obj_type = PLA_MVECTOR;
    error += PLA_Obj_extract_contents  (laobject,
					 stride_ret_buf,
					 leading_dim_ret_buf,
					 rows_in_ret_buf,
					 cols_in_ret_buf,
					 return_buffer);
    laobject->obj_type = PLA_PVECTOR;
    break;
  }
  case PLA_PMVECTOR: {
    stride_ret_buf = 1;
    laobject->obj_type = PLA_MVECTOR;
    error += PLA_Obj_extract_contents  (laobject,
					 stride_ret_buf,
					 leading_dim_ret_buf,
					 rows_in_ret_buf,
					 cols_in_ret_buf,
					 return_buffer);
    laobject->obj_type = PLA_PMVECTOR;
    break;
  }
  default: {
    printf 
      (" Invalid PLA_Object passed to PLA_Obj_extract_transpose_contents\n");
    printf (" Object must be a Matrix, multivector, multiscalar\n");
    error = 1;
    return(error);
  }
  }
  return (error);
}



