/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/
#include "PLA.h"
int PLA_Panel_buf_size(MPI_Datatype        datatype,
                       int               proj_onto,
                       int               owner,
                       int               length,
                       int               width,
                       PLA_Template      template,
                       int               global_align,
                       int              *bytes_required)
{
   int local_length, local_width, data_size, zero_or_one;
   int my_rank;
 
  MPI_Type_size(datatype, &data_size); 
  PLA_Temp_zero_or_one (template, &zero_or_one);
  if ( PLA_ALIGN_FIRST == global_align) global_align = zero_or_one;
  if ( proj_onto == PLA_PROJ_ONTO_ROW )
  {
    PLA_Temp_comm_col_rank ( template, & my_rank );
    local_length = ( ( owner == PLA_ALL_ROWS || owner == my_rank      ) ? length : 0 );
    PLA_Temp_matrix_local_width ( template, width, global_align, &local_width  );
  }
  else if ( proj_onto == PLA_PROJ_ONTO_COL )
  {
    PLA_Temp_comm_row_rank(template, & my_rank);
    PLA_Temp_matrix_local_length ( template, length, global_align, &local_length );
    local_width = ( ( owner == PLA_ALL_COLS || owner == my_rank      ) ? width : 0 );
  }
  else
  {
    printf("ILLEGAL projection passed to PLA_Panel_buf_size\n");
    return(1);
  }
  *bytes_required = data_size * local_length * local_width;
  return(PLA_SUCCESS);
}

int PLA_Panel_create(MPI_Datatype        datatype,
                       int               project_onto,
                       int               owner,
                       int               global_length,
                       int               global_width,
                       PLA_Template      template,
                       int               global_align,
                       PLA_Obj          *new_panel)
{
   int return_value = 0;
   void *data_buffer = NULL;
   int buffer_flag = 0;
   int buffer_size = 0;

   if(project_onto == PLA_PROJ_ONTO_ROW)
   {
      return_value += PLA_Panel_create_base(datatype, PLA_PROJ_ONTO_ROW, owner,
                                      global_length, global_width, template, global_align,
                                      buffer_flag, buffer_size, data_buffer,
                                      new_panel);
   }
   else if(project_onto == PLA_PROJ_ONTO_COL)
   {
      return_value += PLA_Panel_create_base(datatype, PLA_PROJ_ONTO_COL, owner,
                                      global_length, global_width, template, global_align,
                                      buffer_flag, buffer_size, data_buffer,
                                      new_panel);
   }
   else
   {
      return_value += UNKNOWN_PROJECTION;
      return(return_value);
   }       /* One of my new fields */
   (*new_panel)->allocation_type = PLA_HEAP;       /* vs. PLA_USER_BUFFER */
   (*new_panel)->has_data_space  = PLA_HAS_SPACE;  /* vs. PLA_NO_SPACE */
   return(return_value);
}

int PLA_Panel_create_base (MPI_Datatype           datatype,
                     PLA_Panel_rowcol_flag   proj_onto,
                     PLA_Duplicate_flag      owner,
      		     int                     length,
		     int                     width,
		     PLA_Template            template,
		     int                     global_align_idx,
                     int                     buffer_flag,
                     int                     external_buffer_size,
                     void                  * external_buffer,
		     PLA_Obj               * new_panel )
{

  int buf_size, error=0, data_size, my_rank;
  static int messagenum=0;
  PLA_Obj      panel;
  PLA_Start_num  zero_or_one;
  char routine_name [ 32 ] = "PLA_Panel_create";
  int index;
  int local_buffer_size_required;


  PLA_Stack_push_current ( routine_name );
  error += PLA_Obj_free(new_panel);

  if(! PLA_Initialized_object_structs())
     PLA_Init_object_structs(PLA_DEFAULT_OBJ_COUNT); /* some reasonable number of objects */
  index = PLA_Get_free_object_struct();
    if(-1 == index)
  {
     printf("Could not get an allocated object structure in Panel/P(m)vector create\n");
     printf("either the user did not set aside any object structures\n");
     printf("or they didn't set aside any and the system number is low\n");
     /* Note : in the future this will call for doing a mark-and-copy  */
     /*        type deal -- IF the user did NOT do their own set-aside */
     /*        and just FAILING if they did.                           */
  }
/*printf("Panel : index is %5d and place is %ld\n", index, &PLA_GLOBAL_OBJ_STRUCT[index]); */
  panel = (PLA_Obj) (&PLA_GLOBAL_OBJ_STRUCT[index]);

  PLA_Temp_zero_or_one (template, &zero_or_one);
  if ( PLA_ALIGN_FIRST == global_align_idx) global_align_idx = zero_or_one;
  PLA_Obj_init (panel);

/*  if (1==width)
    panel->obj_type         = PLA_PVECTOR;
  else
    panel->obj_type         = PLA_PMVECTOR;
*/

/* considering problems with gemv and trsm, I'm defaulting all panels
   to type PMVECTOR regardless of width.  Will print out message for
   possible vector case to inform users.  gab 3/8/97

   Just print once so we can still do timing tests.  jro 3/28/97
*/

  panel->obj_type         = PLA_PMVECTOR;
  panel->datatype         = datatype;
  panel->master_length    = panel->global_length    = length;
  panel->master_width     = panel->global_width     = width;
  panel->local_stride     = 1;
  panel->template         = template;

  panel->annotate_flags.panel_rowcol      = proj_onto;
  panel->annotate_flags.duplicate         = owner;

  if ( proj_onto == PLA_PROJ_ONTO_ROW )
  {
    PLA_Temp_comm_col_rank ( template, & my_rank );

    panel->global_align_col = panel->master_align_col = global_align_idx;
    panel->global_align_row = panel->master_align_row = (int)zero_or_one;

    panel->local_length = ( ( owner == PLA_ALL_ROWS ||
                              owner == my_rank      ) ? length : 0 );
    PLA_Temp_matrix_local_width ( template, width, global_align_idx,
                                  & ( panel->local_width ) );

    panel->owner_top_block = owner;
    panel->length_top_block = ( ( owner == PLA_ALL_ROWS ||
                                  owner == my_rank      ) ? length : 0 );
    panel->owner_bottom_block = owner;
    panel->length_bottom_block = ( ( owner == PLA_ALL_ROWS ||
                                     owner == my_rank      ) ? length : 0 );

    PLA_Temp_matrix_first_col_block ( template, width, global_align_idx,
                                      & ( panel->owner_left_block ),
                                      & ( panel->width_left_block ) );
    PLA_Temp_matrix_last_col_block ( template, width, global_align_idx,
                                     & ( panel->owner_right_block ),
                                     & ( panel->width_right_block ) );
  }
  else if ( proj_onto == PLA_PROJ_ONTO_COL )
  {
    PLA_Temp_comm_row_rank(template, & my_rank);

    panel->global_align_row = panel->master_align_row = global_align_idx;
    panel->global_align_col = panel->master_align_col = (int)zero_or_one;

    PLA_Temp_matrix_local_length ( template, length, global_align_idx,
                                   & ( panel->local_length ) );
    panel->local_width = ( ( owner == PLA_ALL_COLS ||
                             owner == my_rank      ) ? width : 0 );


    PLA_Temp_matrix_first_row_block ( template, length, global_align_idx,
                                      & ( panel->owner_top_block ),
                                      & ( panel->length_top_block ) );
    PLA_Temp_matrix_last_row_block ( template, length, global_align_idx,
                                     & ( panel->owner_bottom_block ),
                                     & ( panel->length_bottom_block ) );

    panel->owner_left_block = owner;
    panel->width_left_block = ( ( owner == PLA_ALL_COLS ||
                                  owner == my_rank      ) ? width : 0 );
    panel->owner_right_block = owner;
    panel->width_right_block = ( ( owner == PLA_ALL_COLS ||
                                   owner == my_rank      ) ? width : 0 );

  }
  else
  {
    PLA_Obj_free ( & panel );
    PLA_Log_message ( routine_name, "Invalid Projection - NULL object returned" );
    PLA_Stack_pop_current ( routine_name );
    return ( PLA_RETURN_LOG_MESSAGE );
  }

  panel->local_lld        = panel->local_length;

  MPI_Type_extent ( datatype, & data_size );

  panel->length_master_buffer_byte = panel->local_length *
                                     panel->local_width * data_size;
  if(-1 == buffer_flag) /* Create an object with NO!!! buffer */
     buf_size = -1;
  else buf_size = panel->local_length * panel->local_width;

  if(0 == buffer_flag)
  {
     if (panel->length_master_buffer_byte)
     {
        panel->master_buffer =  pla_calloc (panel->length_master_buffer_byte,1);
        if (panel->master_buffer == NULL)
        {
           printf ("Could not allocate buffer of %d bytes in PLA_Panel_create\n",
	                 buf_size*(int)data_size);
           return(1);
        }
     }
     else
        panel->master_buffer = NULL;
  }
  else if(1 == buffer_flag)
  {
     MPI_Type_size (datatype, &data_size);
     local_buffer_size_required = data_size * buf_size;
     if(external_buffer_size < local_buffer_size_required)
     {
        printf("PLA_Panel_create_buf was called with an attached buffer that was too small\n");
        return(1);
     }
     panel->master_buffer = external_buffer;
  }
  panel->buffer = panel->master_buffer;

/* The views will also have to be taken from some set-aside space */
  index = PLA_Get_free_views_struct();
  if(-1 == index)
  {
     printf("Could not get an allocated views structure\n");
     printf("either the user did not set aside any object structures\n");
     printf("or they didn't set aside any and the system number is low\n");
     /* Note : in the future this will call for doing a mark-and-copy  */
     /*        type deal -- IF the user did NOT do their own set-aside */
     /*        and just FAILING if they did.                           */
  }
  panel->views  = (int *) (&PLA_GLOBAL_VIEWS_STRUCT[index]);
  panel->views[0] = 1;

  * new_panel = panel;

  PLA_Stack_pop_current ( routine_name );
  return ( PLA_SUCCESS );
}



int PLA_Panel_subpanel (PLA_Obj   old_panel, 
                        int       length, 
                        int       width,
                        int       align_row,
                        int       align_col,
                        PLA_Obj * new_panel )
{
  int zero_or_one, top_shift, left_shift, right_shift, bottom_shift;
  char routine_name [ 32 ] = "PLA_Panel_subpanel";
  PLA_Template template = NULL;
  int  index;

  PLA_Stack_push_current ( routine_name );

  if ( old_panel->obj_type != PLA_PVECTOR && old_panel->obj_type != PLA_PMVECTOR )
  {
    PLA_Log_message ( routine_name, "Invalid Object - No action performed" );
    PLA_Stack_pop_current ( routine_name );
    return ( PLA_RETURN_LOG_MESSAGE );
  }  

  if ( * new_panel )
  {
    if ( ( * new_panel )->views[0] == 1 )
      PLA_Obj_free ( new_panel );
    else
      ( * new_panel )->views[0]--;
  }
  
  if ( NULL == * new_panel )
  {
    index = PLA_Get_free_object_struct();
    if(-1 == index)
    {
     printf("Could not get an allocated object structure in Panel/P(m)vector create\n");
     printf("either the user did not set aside any object structures\n");
     printf("or they didn't set aside any and the system number is low\n");
     /* Note : in the future this will call for doing a mark-and-copy  */
     /*        type deal -- IF the user did NOT do their own set-aside */
     /*        and just FAILING if they did.                           */
    }
/* printf("subPanel : index is %5d and place is %ld\n", index, &PLA_GLOBAL_OBJ_STRUCT[index]); */
    * new_panel = (PLA_Obj) (&PLA_GLOBAL_OBJ_STRUCT[index]);
    PLA_Obj_init ( *new_panel );
  }
  * ( * new_panel ) = * old_panel;

  ( * new_panel )->views[0]++;

  PLA_Obj_template ( * new_panel, & template );
  PLA_Temp_zero_or_one ( template, & zero_or_one );

  if ( length == PLA_DIM_ALL )        length = ( * new_panel )->global_length;
  if ( width == PLA_DIM_ALL )         width  = ( * new_panel )->global_width;
  if ( align_row == PLA_ALIGN_FIRST ) align_row = zero_or_one;
  if ( align_col == PLA_ALIGN_FIRST ) align_col = zero_or_one;

  top_shift  = align_row - zero_or_one;
  left_shift = align_col - zero_or_one;
  right_shift  = width  + align_col - zero_or_one - ( * new_panel )->global_width;
  bottom_shift = length + align_row - zero_or_one - ( * new_panel )->global_length;

  if ( PLA_Panel_view_shift ( * new_panel,
                                top_shift,
                                left_shift,             
                                right_shift,
                                bottom_shift ) != PLA_SUCCESS )
  {
    PLA_Obj_free ( new_panel );
    PLA_Log_message ( routine_name, "View shift failed - NULL object returned" );
    PLA_Stack_pop_current ( routine_name );
    return ( PLA_RETURN_LOG_MESSAGE );
  }

  PLA_Stack_pop_current ( routine_name );
  return ( PLA_SUCCESS );
}


int PLA_Panel_local_all  (PLA_Obj              panel,
			 MPI_Datatype *        datatype,
			 int *                 local_length,
			 int *                 local_width,
			 void **               buffer,
			 int  *                local_lld)
{
  int error=0;

  error += PLA_Obj_datatype       (panel,  datatype);
  error += PLA_Obj_local_length   (panel,  local_length);
  error += PLA_Obj_local_width    (panel,  local_width);
  error += PLA_Obj_local_buffer         (panel,  buffer); 
  error += PLA_Obj_local_ldim      (panel,  local_lld); 
  return (error);
}


int PLA_Panel_global_all  (PLA_Obj          panel,
                          PLA_Panel_rowcol_flag* proj_onto,
			  int *             global_length,
			  int *             global_width,
			  int  *            global_align_row,
			  int  *            global_align_col,
			  PLA_Template *     template)
     
{
  int error=0;
  error += PLA_Obj_project_onto   (panel,  proj_onto);
  error += PLA_Obj_global_length     (panel,  global_length);
  error += PLA_Obj_global_width      (panel,  global_width);
  error += PLA_Obj_global_align_row  (panel,  global_align_row); 
  error += PLA_Obj_global_align_col  (panel,  global_align_col); 
  error += PLA_Obj_template          (panel,  template); 
  return (error);
}


int PLA_Panel_view_shift (PLA_Obj panel, 
			 int      length_top, 
			 int      width_left,
			 int      width_right,
			 int      length_bottom)
{
  
  
  PLA_Template template = NULL;
  int length, width, align_row, align_col,
      offset, error=0, data_size, my_rank; 
  PLA_Panel_rowcol_flag proj_onto;
  PLA_Duplicate_flag owner;
  char routine_name [ 32 ] = "PLA_Panel_view_shift";

  PLA_Stack_push_current ( routine_name );

  if ( length_top  == 0 && width_left    == 0 &&
       width_right == 0 && length_bottom == 0 )
  {
    PLA_Stack_pop_current ( routine_name );
    return ( PLA_SUCCESS );
  }

  if ( panel->obj_type != PLA_PVECTOR && panel->obj_type != PLA_PMVECTOR )
  {
    PLA_Log_message ( routine_name, "Invalid Object - No action performed" );
    PLA_Stack_pop_current ( routine_name );
    return ( PLA_RETURN_LOG_MESSAGE );
  }  

  align_row = panel->global_align_row + length_top;
  align_col = panel->global_align_col + width_left;
  length = panel->global_length + length_bottom - length_top;
  width = panel->global_width + width_right - width_left;


  if ( align_row < panel->master_align_row ||
       align_col < panel->master_align_col ||
       length + align_row > panel->master_align_row + panel->master_length ||
       width + align_col > panel->master_align_col + panel->master_width )
  {
    PLA_Log_message ( routine_name, "Shift outside master buffer - No action taken" );
    PLA_Stack_pop_current ( routine_name );
    return ( PLA_RETURN_LOG_MESSAGE );
  }


  PLA_Obj_template ( panel, & template );
  PLA_Obj_project_onto ( panel, & proj_onto );
  MPI_Type_extent ( panel->datatype, & data_size );

  if ( proj_onto == PLA_PROJ_ONTO_ROW )
  {
    PLA_Temp_comm_col_rank ( template, & my_rank );
    PLA_Obj_owner_row ( panel, & owner );

    panel->local_length = ( ( owner == PLA_ALL_ROWS ||
                              owner == my_rank      ) ? length : 0 );
    PLA_Temp_matrix_local_width ( template, width, align_col,
                                  & ( panel->local_width ) );

    panel->owner_top_block = owner;
    panel->length_top_block = ( ( owner == PLA_ALL_ROWS ||
                                  owner == my_rank      ) ? length : 0 );
    panel->owner_bottom_block = owner;
    panel->length_bottom_block = ( ( owner == PLA_ALL_ROWS ||
                                     owner == my_rank      ) ? length : 0 );

    PLA_Temp_matrix_first_col_block ( template, width, align_col,
                                      & ( panel->owner_left_block ),
                                      & ( panel->width_left_block ) );
    PLA_Temp_matrix_last_col_block ( template, width, align_col,
                                     & ( panel->owner_right_block ),
                                     & ( panel->width_right_block ) );

    PLA_Temp_matrix_local_width ( template,
                                  align_col - panel->master_align_col,
                                  panel->master_align_col, & offset );
    offset *= panel->local_lld;
    offset += align_row;
    offset *= data_size;

    if ( panel->master_buffer != NULL )
       panel->buffer = ( char * ) panel->master_buffer + offset;

  }
  else if ( proj_onto == PLA_PROJ_ONTO_COL )
  {
    PLA_Temp_comm_row_rank ( template, & my_rank );
    PLA_Obj_owner_col ( panel, & owner );

    PLA_Temp_matrix_local_length ( template, length, align_row,
                                   & ( panel->local_length ) );
    panel->local_width = ( ( owner == PLA_ALL_COLS ||
                             owner == my_rank      ) ? width : 0 );

    PLA_Temp_matrix_first_row_block ( template, length, align_row,
                                      & ( panel->owner_top_block ),
                                      & ( panel->length_top_block ) );
    PLA_Temp_matrix_last_row_block ( template, length, align_row,
                                     & ( panel->owner_bottom_block ),
                                     & ( panel->length_bottom_block ) );

    panel->owner_left_block = owner;
    panel->width_left_block = ( ( owner == PLA_ALL_COLS ||
                                  owner == my_rank      ) ? width : 0 );
    panel->owner_right_block = owner;
    panel->width_right_block = ( ( owner == PLA_ALL_COLS ||
                                   owner == my_rank      ) ? width : 0 );

    PLA_Temp_matrix_local_length ( template,
                                   align_row - panel->master_align_row,
                                   panel->master_align_row,
                                   & offset );

    offset += align_col * panel->local_lld;
    offset *= data_size;

    if ( panel->master_buffer != NULL )
       panel->buffer = ( char * ) panel->master_buffer + offset;

  }
  else
  {
    PLA_Log_message ( routine_name, "Invalid Projection - No action performed" );
    PLA_Stack_pop_current ( routine_name );
    return ( PLA_RETURN_LOG_MESSAGE );
  }

  panel->global_length = length;
  panel->global_width = width;
  panel->global_align_col = align_col;
  panel->global_align_row = align_row;

  PLA_Stack_pop_current ( routine_name );
  return ( PLA_SUCCESS );
}

  
