/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/


#include "PLA_Temp_local.h"



int PLA_Temp_local_compute (int  global_length, 
			   int  global_offset, 
			   int  processor_num,
			   int  nb,
			   int  num_processors,
			   int  z_or_1,
			   int *local_length, 
			   int *starting_proc )

/********************************************************************
Scalable BLAS utility routine: PLA_Temp_local_compute

Assume template vector 
         / x0  \    where each subvector is of length nb
     x = | x1  |    is distributed so that xi is assigned to
         | x2  |    the ith processor in communicator comm.
         \ ... /                
Given a vector of length n, with the first element alighned
with the ith element of x, this function computes the length
of the local vector (returned in *local_length) and the processor
number of the processor that owns the first element (returned
in *icur)

Parameters

n                 INPUT: int
                  length of vector

global_offset                 INPUT: int
                  element of template vector to which vector is aligned

*local_length              OUTPUT: pointer to int
                  length this node's part of vector


*starting_proc             OUTPUT: pointer to int
                  index of node that owns first element of vector
 
Copyright 1995  The University of Texas at Austin
This code is distributed under gnu license.

Programmed by Robert van de Geijn
              rvdg@cs.utexas.edu
********************************************************************/
{
  int adj_global_offset,
      itemp, ipre;      /* temporary variables                     */

  
  adj_global_offset = global_offset - z_or_1;
                /* Compute proc. that holds element (n+i) of a vector
                   aligned with processor 0 */
  *starting_proc = 
    ( ( global_length+adj_global_offset )/nb ) % num_processors ;
                /* Compute how much of a vector of length i+global_length-1
                   aligned at proc. 0 this proc. would own         */
  itemp = (global_length+adj_global_offset) / nb;
  *local_length = 
    itemp / num_processors * nb + 
      ( processor_num < ( itemp%num_processors ) ? nb : 0 ) +
	( processor_num == *starting_proc ? 
	 ( global_length + adj_global_offset ) % nb : 0 );
                /* Compute proc. that holds first element          */
  *starting_proc = ( ( adj_global_offset )/nb ) % num_processors ;
                /* Compute how much of a vector of length i-1
                   aligned at proc. 0 this proc. would own         */
  itemp = (adj_global_offset) / nb;
  ipre = itemp / num_processors * nb + 
    ( processor_num < ( itemp%num_processors ) ? nb : 0 ) +
      ( processor_num == *starting_proc ? ( adj_global_offset ) % nb : 0 );
                 /* subtract to compute output local_length        */
  *local_length -= ipre;
  
  return (0);
}

