#include "PLA.h"
#include "POOCLA.h"


void PLA_No_test_ooc_dsyrk ( int my_pe, PLA_Template template )
{
   int length, width, ldim;
   double time, mflops;

   PLA_Obj alpha_obj = NULL, beta_obj = NULL; 
   POOCLA_Obj a_ooc=NULL, c_ooc=NULL;
   MPI_Comm comm1d;
   int      n, k, np, int_uplo, int_trans;

   PLA_Temp_comm_all ( template, & comm1d );
   MPI_Comm_size( comm1d, &np );

   if ( my_pe == 0 )
   {
      printf ( "n k\n" );
      scanf ( "%d%d", & n, & k );
   }

   MPI_Bcast ( & n, 1, MPI_INT, 0, comm1d );
   MPI_Bcast ( & k, 1, MPI_INT, 0, comm1d );

/*
 * create ooc matrices
 */
   make_really_big_matrix ( n, k, template, "/usr/tmp/u20001/Mat_A", & a_ooc );
   make_really_big_matrix ( n, n, template, "/usr/tmp/u20001/Mat_C", & c_ooc );

   POOCLA_free ( & a_ooc );
   POOCLA_free ( & c_ooc );

   POOCLA_create ( template, MPI_DOUBLE, n, k, 0, 0,
                    PLA_MATRIX, PLA_UNDEFINED, PLA_UNDEFINED,
                    "/usr/tmp/u20001/Mat_A", & a_ooc );

   POOCLA_create ( template, MPI_DOUBLE, n, n, 0, 0,
                    PLA_MATRIX, PLA_UNDEFINED, PLA_UNDEFINED,
                    "/usr/tmp/u20001/Mat_C", & c_ooc );

   MPI_Barrier ( MPI_COMM_WORLD );
   if ( my_pe == 0 ) printf ( "Matrices filled\n" );

/*
 * initialize scaling parameters and constants
 */
  PLA_Mscalar_create ( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                       template, & alpha_obj );
  PLA_Obj_set_to_one ( alpha_obj );
  PLA_Mscalar_create ( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                       template, & beta_obj );
  PLA_Obj_set_to_one ( beta_obj );

  int_uplo =  PLA_LOWER_TRIANGULAR;
  int_trans = PLA_NO_TRANS;


/*
 * calculation ooc syrk, results on disk
 */

  POOCLA_activate_instrumentation ();

  MPI_Barrier( MPI_COMM_WORLD );
  time = MPI_Wtime();

  POOCLA_Syrk ( int_uplo, int_trans, alpha_obj, a_ooc, beta_obj, c_ooc );

  MPI_Barrier( MPI_COMM_WORLD );
  time = MPI_Wtime() - time;
  mflops = 1.0e-6 * n * n * k / ( time * np );
  if ( my_pe == 0 )
     printf ( "Out-of-Core : %f seconds at %f mflops per processor\n", time, mflops );

  POOCLA_read_time ( & time );
  MPI_Reduce ( & time, & mflops, 1, MPI_DOUBLE, MPI_SUM, 0, comm1d );
  if ( my_pe == 0 ) printf ( "Average Read Time : %f seconds\n", mflops / np );
  
  POOCLA_write_time ( & time );
  MPI_Reduce ( & time, & mflops, 1, MPI_DOUBLE, MPI_SUM, 0, comm1d );
  if ( my_pe == 0 ) printf ( "Average Write Time : %f seconds\n", mflops / np );

  PLA_Obj_free ( & alpha_obj );
  PLA_Obj_free ( & beta_obj );
  POOCLA_free ( & a_ooc );
  POOCLA_free ( & c_ooc );
}

int make_really_big_matrix ( int m, int n, PLA_Template template,
                             char * filename, POOCLA_Obj * ooc )
{
   int top_size, left_size, length, width, align_row, align_col,
       filestatus;
   PLA_Obj obj = NULL;
   POOCLA_Obj top = NULL, bottom = NULL, right = NULL;

   POOCLA_create ( template, MPI_DOUBLE, m, n, 0, 0,
                    PLA_MATRIX, PLA_UNDEFINED, PLA_UNDEFINED, filename, ooc );

   POOCLA_filestatus ( * ooc, & filestatus );

   if ( filestatus == 0 ) return ( 0 );

   POOCLA_view_all ( * ooc, & right );


   POOCLA_split_size ( right, PLA_SIDE_LEFT, & left_size );

   while ( left_size > 0 )
   {
      POOCLA_vert_split_2 ( right, left_size, & bottom, & right );

      POOCLA_split_size ( bottom, PLA_SIDE_TOP, & top_size );

      while ( top_size > 0 )
      {
         POOCLA_horz_split_2 ( bottom, top_size, & top, & bottom );

         POOCLA_global_length ( top, & length );
         POOCLA_global_width ( top, & width );
         POOCLA_global_align_row ( top, & align_row );
         POOCLA_global_align_col ( top, & align_col );

         PLA_Matrix_create ( MPI_DOUBLE, length, width, template,
                             align_row, align_col, & obj );
         PLA_Obj_fill ( "G", 1.0e+0, & obj );
         POOCLA_write ( top, obj );

         POOCLA_split_size ( bottom, PLA_SIDE_TOP, & top_size );
      }

      POOCLA_split_size ( right, PLA_SIDE_LEFT, & left_size );
   }

   POOCLA_free ( & top );
   POOCLA_free ( & bottom );
   POOCLA_free ( & right );
   PLA_Obj_free ( & obj );

   return ( 0 );
}
