#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include "PLA.h"
#include "POOCLA.h"

int main ( int argc, char **argv )
{
   int my_pe, n_pes, nb, nb_alg, nb_incore, option;
   MPI_Comm base_comm;
   PLA_Template template = PLA_NULL;

   MPI_Init ( & argc, & argv );
   PLA_Init ( MPI_COMM_NULL );

   PLA_Base_comm ( & base_comm );
   MPI_Comm_rank ( base_comm, & my_pe );

   if ( my_pe == 0 )
   {
      printf ( "Enter nb, nb_alg, nb_incore : \n" );
      fflush ( stdout );
      scanf ( "%d%d%d", & nb, & nb_alg, & nb_incore );
   }
   MPI_Bcast ( &nb, 1, MPI_INT, 0, base_comm );
   MPI_Bcast ( &nb_alg, 1, MPI_INT, 0, base_comm );
   MPI_Bcast ( &nb_incore, 1, MPI_INT, 0, base_comm );

   PLA_Temp_create ( nb, 0, & template );
   PLA_Temp_comm_all_info ( template, & base_comm, & my_pe, & n_pes );

   pla_Environ_set_nb_alg ( PLA_OP_ALL_ALG, nb_alg );
   POOCLA_set_nb_incore ( nb_incore, nb_incore );

   POOCLA_activate_instrumentation ();

   if ( my_pe == 0 )
   {
      printf ( "0. Time\n1. Test Block\n2. OOC Syrk\n3. OOC Chol\n4. OOC Trsm\n5. Distribute\n6. Checkpoints\nelse exit :\n" );
      fflush ( stdout );
      scanf ( "%d", & option );
   }
   MPI_Bcast ( & option, 1, MPI_INT, 0, base_comm );

   switch ( option )
   {
      case 0  : PLA_test_ooc_time ( my_pe, template );
                break;
      case 1  : PLA_test_ooc_one_to_all ( my_pe, template );
                break;
      case 2  : PLA_test_ooc_dsyrk ( my_pe, template );
                break;
      case 3  : PLA_test_ooc_dchol ( my_pe, template );
                break;
      case 4  : PLA_test_ooc_dtrsm ( my_pe, template );
                break;
      case 5  : test_matrix ( my_pe, template );
                break;
      case 6  : test_chkpt ( my_pe, template );
      default : break;
   }

   PLA_Temp_free ( & template );
   PLA_Finalize ();
   MPI_Finalize ();
}

int PLA_test_ooc_time ( int my_pe, PLA_Template template )
{
   int universal_length, universal_width, universal_align_row, universal_align_col;
   int i, local_length, local_width, n_pes;
   double * incore_read_buffer, * incore_object_buffer;
   double start, elapsed, bytes;
   PLA_Obj incore_object, incore_read = NULL;
   POOCLA_Obj ooc_object = NULL;
   MPI_Comm comm_all = NULL;

   PLA_Temp_comm_all ( template, & comm_all );
   if ( my_pe == 0 )
   {
      printf ( "Universal Length    : \n" );
      scanf ( "%d", & universal_length );
      printf ( "Universal Width     : \n" );
      scanf ( "%d", & universal_width );
      printf ( "Universal Align Row : \n" );
      scanf ( "%d", & universal_align_row );
      printf ( "Universal Align Col : \n" );
      scanf ( "%d", & universal_align_col );
   }
   MPI_Bcast ( & universal_length,    1, MPI_INT, 0, comm_all );
   MPI_Bcast ( & universal_width,     1, MPI_INT, 0, comm_all );
   MPI_Bcast ( & universal_align_row, 1, MPI_INT, 0, comm_all );
   MPI_Bcast ( & universal_align_col, 1, MPI_INT, 0, comm_all );

   POOCLA_create ( template, MPI_DOUBLE,
                    universal_length, universal_width,
                    universal_align_row, universal_align_col,
                    PLA_PMVECTOR, PLA_PROJ_ONTO_COL, 0,
                    "/workspace/baker/time", & ooc_object );

   if ( ooc_object == NULL ) return ( -1 );

   PLA_Matrix_create ( MPI_DOUBLE, universal_length, universal_width, template,
                       universal_align_row, universal_align_col, & incore_object );

   PLA_Obj_fill ( "C", 9.0e+0, & incore_object );

   bytes = 8.0e-6 * universal_length * universal_width;
   MPI_Comm_size ( comm_all, & n_pes );


   MPI_Barrier ( MPI_COMM_WORLD );
   start = MPI_Wtime ();

   POOCLA_write ( ooc_object, incore_object );

   elapsed = MPI_Wtime () - start;

   printf ( "PE%2d  Write : Elapsed = %f  Rate = %f\n",
             my_pe, elapsed, bytes / ( n_pes * elapsed ) );

   MPI_Allreduce ( & elapsed, & start, 1, MPI_DOUBLE, MPI_MAX, comm_all );
   if ( my_pe == 0 )
      printf ( "Write : Elapsed = %f  Rate = %f\n", start, bytes / start );

   PLA_Obj_free ( & incore_object );


   MPI_Barrier ( MPI_COMM_WORLD );
   start = MPI_Wtime ();

   POOCLA_read ( ooc_object, PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & incore_read );

   elapsed = MPI_Wtime () - start;

   printf ( "PE%2d  READ : Elapsed = %f  Rate = %f\n",
             my_pe, elapsed, bytes / ( n_pes * elapsed ) );

   MPI_Allreduce ( & elapsed, & start, 1, MPI_DOUBLE, MPI_MAX, comm_all );
   if ( my_pe == 0 )
      printf ( "Read : Elapsed = %f  Rate = %f\n", start, bytes / start );

   PLA_Obj_free ( & incore_read );
   POOCLA_free ( & ooc_object );

   return ( 0 );
}


int PLA_test_ooc_one_to_all ( int my_pe, PLA_Template template )
{
   int universal_length, universal_width, universal_align_row, universal_align_col,
       num_row_blocks, top_size, left_size;
   double * buffer, * current_buffer, d_one = 1.0e+0;
   PLA_Obj incore_object = NULL;
   POOCLA_Obj ooc_object = NULL, topleft_block = NULL, bottomright_block = NULL,
              dummy = NULL;
   MPI_Comm comm_all = NULL;

   PLA_Temp_comm_all ( template, & comm_all );
   if ( my_pe == 0 )
   {
      printf ( "Universal Length, Width, Align Row, Align col : \n" );
      scanf ( "%d%d%d%d", & universal_length, & universal_width,
                          & universal_align_row, & universal_align_col );
   }
   MPI_Bcast ( & universal_length,    1, MPI_INT, 0, comm_all );
   MPI_Bcast ( & universal_width,     1, MPI_INT, 0, comm_all );
   MPI_Bcast ( & universal_align_row, 1, MPI_INT, 0, comm_all );
   MPI_Bcast ( & universal_align_col, 1, MPI_INT, 0, comm_all );

   buffer = pla_calloc ( universal_length * universal_width, sizeof ( double ) );

   if ( my_pe == 0 )
      PLA_test_dgenmatrix ( "G", 5.0e+0, universal_length, universal_width,
                             buffer, universal_length );

   MPI_Bcast ( buffer, universal_length * universal_width, MPI_DOUBLE, 0, comm_all );

   POOCLA_create ( template, MPI_DOUBLE,
                   universal_length, universal_width,
                   universal_align_row, universal_align_col,
                   PLA_PMVECTOR, PLA_PROJ_ONTO_COL, 0,
                   "data", & ooc_object );

   POOCLA_create_PLA_Obj_conf_to ( ooc_object, PLA_MATRIX, & incore_object );

   PLA_Obj_set_to_zero ( incore_object );
   PLA_Test_gdist_to_local ( buffer, universal_length, incore_object );

   POOCLA_write ( ooc_object, incore_object );
   POOCLA_validate_filestatus ( ooc_object );

   PLA_Obj_free ( & incore_object );

   current_buffer = buffer;

   POOCLA_view_all ( ooc_object, & bottomright_block );
   POOCLA_split_size ( bottomright_block, PLA_SIDE_TOP, & top_size );
   POOCLA_split_size ( bottomright_block, PLA_SIDE_LEFT, & left_size );

   while ( top_size > 0 && left_size > 0 )
   {
      POOCLA_split_4 ( bottomright_block, top_size, left_size,
                        & topleft_block, & dummy, & dummy, & bottomright_block );

      POOCLA_read ( topleft_block,
                     PLA_MATRIX, PLA_UNDEFINED, PLA_UNDEFINED, & incore_object );

      if ( PLA_test_compare ( current_buffer, universal_length, incore_object )) break;

      POOCLA_write ( topleft_block, incore_object );

      current_buffer += top_size + left_size * universal_length;

      POOCLA_split_size ( bottomright_block, PLA_SIDE_TOP, & top_size );
      POOCLA_split_size ( bottomright_block, PLA_SIDE_LEFT, & left_size );

   }

   PLA_Obj_free ( & incore_object );
   POOCLA_free ( & ooc_object );
   POOCLA_free ( & topleft_block );
   POOCLA_free ( & bottomright_block );
   POOCLA_free ( & dummy );
   POOCLA_free ( & dummy );
   pla_free ( buffer ); buffer = NULL; current_buffer = NULL;

   return ( 0 );
}

