#include "POOCLA.h"

int POOCLA_Chol ( int uplo, POOCLA_Obj a )
{
   POOCLA_Chol_l ( a );
   return ( 0 );
}

int POOCLA_Chol_l ( POOCLA_Obj ooc_object )
{
   int length, block_size, top_size, left_size;
   MPI_Datatype datatype = MPI_DATATYPE_NULL;
   PLA_Template template = NULL;
   POOCLA_Obj relevent = NULL,
           a_top = NULL, a_bottom = NULL, c_top = NULL, c_bottom = NULL,
           a_top_row = NULL, a_top_block = NULL, a_block = NULL,
           a_diag = NULL, c_diag = NULL;
   PLA_Obj AObj = NULL, BObj = NULL, CObj = NULL, DObj = NULL,
           one = NULL, min_one = NULL;

   POOCLA_template ( ooc_object, & template );
   POOCLA_datatype ( ooc_object, & datatype );

   PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &one );
   PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &min_one );
   PLA_Obj_set_to_one ( one );
   PLA_Obj_set_to_minus_one ( min_one );

   POOCLA_view_all ( ooc_object, & relevent );
   POOCLA_global_length ( relevent, & length );

   while ( length > 0 )
   {
      POOCLA_vert_split_2 ( relevent, -length, & a_bottom, & c_bottom );

      POOCLA_split_size ( c_bottom, PLA_SIDE_TOP, & top_size );
      POOCLA_split_size ( c_bottom, PLA_SIDE_LEFT, & left_size );
      block_size = min ( top_size, left_size );

      POOCLA_view_shift ( c_bottom, 0, 0, block_size - length, 0 );

      POOCLA_horz_split_2 ( a_bottom, block_size, & a_diag, & a_bottom );
      POOCLA_horz_split_2 ( c_bottom, block_size, & c_diag, & c_bottom );

      POOCLA_split_size ( c_bottom, PLA_SIDE_TOP, & top_size );
      POOCLA_horz_split_2 ( a_bottom, top_size, & a_top, & a_bottom );
      POOCLA_horz_split_2 ( c_bottom, top_size, & c_top, & c_bottom );

      POOCLA_read ( c_diag, PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & CObj );
      POOCLA_read ( c_top,  PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & DObj );

      POOCLA_view_all ( a_diag, & a_top_row );
      POOCLA_split_size ( a_top_row, PLA_SIDE_LEFT, & left_size );

      while ( left_size > 0 )
      {
         POOCLA_vert_split_2 ( a_top_row, left_size, & a_top_block, & a_top_row );
         POOCLA_vert_split_2 ( a_top,     left_size, & a_block,     & a_top );

         POOCLA_read ( a_top_block, PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & AObj );
         POOCLA_read ( a_block,     PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & BObj );

         PLA_Syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS, min_one, AObj, one, CObj );
         PLA_Gemm ( PLA_NO_TRANS, PLA_TRANS, min_one, BObj, AObj, one, DObj );

         POOCLA_split_size ( a_top_row, PLA_SIDE_LEFT, & left_size );
      }

      PLA_Chol ( PLA_LOW_TRIAN, CObj );
      PLA_Trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                 one, CObj, DObj );

      POOCLA_write ( c_diag, CObj );
      POOCLA_write ( c_top,  DObj );

      POOCLA_split_size ( c_bottom, PLA_SIDE_TOP, & top_size );

      while ( top_size > 0 )
      {
         POOCLA_horz_split_2 ( a_bottom, top_size, & a_top, & a_bottom );
         POOCLA_horz_split_2 ( c_bottom, top_size, & c_top, & c_bottom );

         POOCLA_read ( c_top,  PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & DObj );

         POOCLA_view_all ( a_diag, & a_top_row );
         POOCLA_split_size ( a_top_row, PLA_SIDE_LEFT, & left_size );

         while ( left_size > 0 )
         {
            POOCLA_vert_split_2 ( a_top_row, left_size, & a_top_block, & a_top_row );
            POOCLA_vert_split_2 ( a_top,     left_size, & a_block,     & a_top );

            POOCLA_read ( a_top_block, PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & AObj );
            POOCLA_read ( a_block,     PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & BObj );

            PLA_Gemm ( PLA_NO_TRANS, PLA_TRANS, min_one, BObj, AObj, one, DObj );

            POOCLA_split_size ( a_top_row, PLA_SIDE_LEFT, & left_size );
         }

         PLA_Trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                    one, CObj, DObj );

         POOCLA_write ( c_top,  DObj );

         POOCLA_split_size ( c_bottom, PLA_SIDE_TOP, & top_size );
      }

      POOCLA_view_shift ( relevent, block_size, 0, 0, 0 );
      POOCLA_global_length ( relevent, & length );
   }

   POOCLA_free ( & relevent );
   POOCLA_free ( & a_top );
   POOCLA_free ( & a_bottom );
   POOCLA_free ( & c_top );
   POOCLA_free ( & c_bottom );
   POOCLA_free ( & a_top_row );
   POOCLA_free ( & a_top_block );
   POOCLA_free ( & a_block );
   POOCLA_free ( & a_diag );
   POOCLA_free ( & c_diag );

   PLA_Obj_free ( & AObj );
   PLA_Obj_free ( & BObj );
   PLA_Obj_free ( & CObj );
   PLA_Obj_free ( & DObj );
   PLA_Obj_free ( & one );
   PLA_Obj_free ( & min_one );

   return ( 0 );
}

