#include "POOCLA.h"

int POOCLA_Trsm ( int side, int uplo, int transa, int diag,
                   PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b )
{
   if ( side == PLA_SIDE_LEFT )
   {
      if ( uplo == PLA_UP_TRIAN )
      {
         if ( transa == PLA_NO_TRANS ) POOCLA_Trsm_lun ( diag, alpha, a, b );
         if ( transa == PLA_TRANS    ) POOCLA_Trsm_lut ( diag, alpha, a, b );
      }

      if ( uplo == PLA_LOW_TRIAN )
      {
         if ( transa == PLA_NO_TRANS ) POOCLA_Trsm_lln ( diag, alpha, a, b );
         if ( transa == PLA_TRANS    ) POOCLA_Trsm_llt ( diag, alpha, a, b );
      }
   }

   if ( side == PLA_SIDE_RIGHT )
   {
      if ( uplo == PLA_UP_TRIAN )
      {
         if ( transa == PLA_NO_TRANS ) POOCLA_Trsm_run ( diag, alpha, a, b );
         if ( transa == PLA_TRANS    ) POOCLA_Trsm_rut ( diag, alpha, a, b );
      }

      if ( uplo == PLA_LOW_TRIAN )
      {
         if ( transa == PLA_NO_TRANS ) POOCLA_Trsm_rln ( diag, alpha, a, b );
         if ( transa == PLA_TRANS    ) POOCLA_Trsm_rlt ( diag, alpha, a, b );
      }
   }

   return ( PLA_SUCCESS );
}

int POOCLA_Trsm_lun ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b )
{
   return ( PLA_SUCCESS );
}

int POOCLA_Trsm_lut ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b )
{
   return ( PLA_SUCCESS );
}

int POOCLA_Trsm_lln ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b )
{
   int length, top_size, left_size, size;
   PLA_Obj AObj = NULL, BObj = NULL, CObj = NULL, DObj = NULL,
           one = NULL, min_one = NULL;
   POOCLA_Obj a_relevent = NULL, a_left = NULL, a_diag = NULL,
           b_left = NULL, b_right = NULL, b_top = NULL,  b_diag = NULL,
           a_block = NULL, b_block = NULL;

   PLA_Mscalar_create_conf_to ( alpha, PLA_ALL_ROWS, PLA_ALL_COLS, & one );
   PLA_Mscalar_create_conf_to ( alpha, PLA_ALL_ROWS, PLA_ALL_COLS, & min_one );

   PLA_Obj_set_to_one ( one );
   PLA_Obj_set_to_minus_one ( min_one );

   POOCLA_view_all ( b, & b_right );

   POOCLA_split_size ( b_right, PLA_SIDE_LEFT, & left_size );

   while ( left_size > 0 )
   {
      POOCLA_vert_split_2 ( b_right, left_size,  & b_left, & b_right );


      POOCLA_view_all ( a, & a_relevent );
      POOCLA_global_length ( a_relevent, & length );

      while ( length > 0 )
      {
         POOCLA_vert_split_2 ( a_relevent, -length, & a_left, & a_diag );
         POOCLA_horz_split_2 ( b_left, -length, & b_top, & b_diag );

         POOCLA_split_size ( a_diag, PLA_SIDE_TOP, & top_size );
         POOCLA_split_size ( a_diag, PLA_SIDE_LEFT, & left_size );

         size = min ( top_size, left_size );

         POOCLA_view_shift ( a_relevent, size, 0, 0, 0 );

         POOCLA_view_shift ( a_left, 0, 0, 0, size - length );
         POOCLA_view_shift ( a_diag, 0, 0, size - length, size - length );
         POOCLA_view_shift ( b_diag, 0, 0, 0, size - length );


         POOCLA_read ( b_diag, PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & BObj );
         PLA_Scal ( alpha, BObj );


         POOCLA_split_size ( b_top, PLA_SIDE_TOP, & top_size );
         POOCLA_split_size ( a_left, PLA_SIDE_LEFT, & left_size );

         while ( ( size = min ( top_size, left_size ) ) > 0 )
         {
            POOCLA_vert_split_2 ( a_left, size, & a_block, & a_left );
            POOCLA_horz_split_2 ( b_top, size, & b_block, & b_top );

            POOCLA_read ( a_block, PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & CObj );
            POOCLA_read ( b_block, PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & DObj );

            PLA_Gemm ( PLA_NO_TRANS, PLA_NO_TRANS, min_one, CObj, DObj, one, BObj );

            POOCLA_split_size ( b_top, PLA_SIDE_TOP, & top_size );
            POOCLA_split_size ( a_left, PLA_SIDE_LEFT, & left_size );
         }

         PLA_Obj_free ( & CObj );
         PLA_Obj_free ( & DObj );

         POOCLA_read ( a_diag, PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & AObj );

         PLA_Trsm ( PLA_SIDE_LEFT, PLA_LOW_TRIAN, PLA_NO_TRANS, diag, one, AObj, BObj );

         POOCLA_write ( b_diag, BObj );

         PLA_Obj_free ( & AObj );

         POOCLA_global_length ( a_relevent, & length );
      }

      POOCLA_split_size ( b_right, PLA_SIDE_LEFT, & left_size );
   }

   return ( PLA_SUCCESS );
}

int POOCLA_Trsm_llt ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b )
{
   int width, top_size, bottom_size, left_size, right_size, size;
   PLA_Obj AObj = NULL, BObj = NULL, CObj = NULL, DObj = NULL,
           one = NULL, min_one = NULL;
   POOCLA_Obj a_relevent = NULL, a_bottom = NULL, a_diag = NULL,
           b_left = NULL, b_right = NULL, b_bottom = NULL, b_diag = NULL,
           a_block = NULL, b_block = NULL;

   PLA_Mscalar_create_conf_to ( alpha, PLA_ALL_ROWS, PLA_ALL_COLS, & one );
   PLA_Mscalar_create_conf_to ( alpha, PLA_ALL_ROWS, PLA_ALL_COLS, & min_one );

   PLA_Obj_set_to_one ( one );
   PLA_Obj_set_to_minus_one ( min_one );

   POOCLA_view_all ( b, & b_right );
   POOCLA_split_size ( b_right, PLA_SIDE_LEFT, & left_size );

   while ( left_size > 0 )
   {
      POOCLA_vert_split_2 ( b_right, left_size, & b_left, & b_right );

      POOCLA_view_all ( a, & a_relevent );
      POOCLA_global_width ( a_relevent, & width );

      while ( width > 0 )
      {

         POOCLA_horz_split_2 ( a_relevent, width, & a_diag, & a_bottom );
         POOCLA_horz_split_2 ( b_left, width, & b_diag, & b_bottom );

         POOCLA_split_size ( a_diag, PLA_SIDE_BOTTOM, & bottom_size );
         POOCLA_split_size ( a_diag, PLA_SIDE_RIGHT, & right_size );

         size = min ( bottom_size, right_size );

         POOCLA_view_shift ( a_relevent, 0, 0, -size, 0 );

         POOCLA_view_shift ( a_bottom, 0, width - size, 0, 0 );
         POOCLA_view_shift ( a_diag, width - size, width - size, 0, 0 );
         POOCLA_view_shift ( b_diag, width - size, 0, 0, 0 );


         POOCLA_read ( b_diag, PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & BObj );
         PLA_Scal ( alpha, BObj );


         POOCLA_split_size ( a_bottom, PLA_SIDE_TOP, & top_size );

         while ( top_size > 0 )
         {
            POOCLA_horz_split_2 ( a_bottom, top_size, & a_block, & a_bottom );
            POOCLA_horz_split_2 ( b_bottom, top_size, & b_block, & b_bottom );

            POOCLA_read ( a_block, PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & CObj );
            POOCLA_read ( b_block, PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & DObj );

            PLA_Gemm ( PLA_TRANS, PLA_NO_TRANS, min_one, CObj, DObj, one, BObj );

            POOCLA_split_size ( a_bottom, PLA_SIDE_TOP, & top_size );
         }

         PLA_Obj_free ( & CObj );
         PLA_Obj_free ( & DObj );

         POOCLA_read ( a_diag, PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & AObj );

         PLA_Trsm ( PLA_SIDE_LEFT, PLA_LOW_TRIAN, PLA_TRANS, diag, one, AObj, BObj );

         POOCLA_write ( b_diag, BObj );

         PLA_Obj_free ( & AObj );

         POOCLA_global_width ( a_relevent, & width );
      }

      POOCLA_split_size ( b_right, PLA_SIDE_LEFT, & left_size );
   }
 
   return ( PLA_SUCCESS );
}

int POOCLA_Trsm_run ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b )
{
   return ( PLA_SUCCESS );
}

int POOCLA_Trsm_rut ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b )
{
   return ( PLA_SUCCESS );
}

int POOCLA_Trsm_rln ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b )
{
   return ( PLA_SUCCESS );
}

int POOCLA_Trsm_rlt ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b )
{
   return ( PLA_SUCCESS );
}

