#include "PLA.h"

extern int PLA_ALGORYTHMIC_BLOCKSIZE;

void PLA_test_opt_dchol( 	int int_alg,
				PLA_Template template, char *uplo,
			 	int n,
				int ia, int ja,
				double *time, double *mflops, int *info )

{
  int mym, myn,         /* this node's row and column dimension    */
	 lda_all,    /* leading dimensions of global matrices   */
	 llda,   /* leading dimensions of local matrices    */
	 idummy,             /* dummy variable                          */
	 ii;                 /* loop counter -- dummy                   */
  double
	 *a_all,     /* arrays holding the global matrices      */
	 *a;         /* arrays holding the local matrices       */

  PLA_Obj a_obj=NULL;
  MPI_Comm comm1d;
  int      np, int_uplo;

	/* Create distributed matrix objects for A, B, C */
  a_all = (double *) pla_malloc( n * n * sizeof( double ) );
  PLA_Matrix_create ( MPI_DOUBLE, n, n, template, ia, ja, &a_obj );
  lda_all = n;
  PLA_test_dgenmatrix ( "D", 10.0e+0, n, n, a_all, lda_all );
  PLA_Test_gdist_to_local ( a_all, lda_all, a_obj );

		/* compute correct sequential answer             */
  PLA_dchol( n, a_all, n );

	/* compute and time parallel answer              */
  int_uplo =  (*uplo == 'L') ? PLA_LOWER_TRIANGULAR : PLA_UPPER_TRIANGULAR;

  MPI_Barrier( MPI_COMM_WORLD );
  *time = MPI_Wtime();

  switch ( int_alg )
  {
     case 1 : PLA_opt_CHOLxL1 ( int_uplo, a_obj );
              break;
     case 2 : PLA_opt_CHOLxL2 ( int_uplo, a_obj );
              break;
     case 3 : PLA_opt_CHOLxL3 ( int_uplo, a_obj );
              break;
     case 4 : PLA_opt_CHOLxL4 ( int_uplo, a_obj );
              break;
     case 5 : PLA_opt_CHOLxL5 ( int_uplo, a_obj );
              break;
     case 6 : PLA_opt_CHOLxL6 ( int_uplo, a_obj );
              break;
     case 11 : PLA_opt_CHOLxR1 ( int_uplo, a_obj );
              break;
     case 12 : PLA_opt_CHOLxR2 ( int_uplo, a_obj );
              break;
     case 13 : PLA_opt_CHOLxR3 ( int_uplo, a_obj );
              break;
     case 14 : PLA_opt_CHOLxR4 ( int_uplo, a_obj );
              break;
     case 15 : PLA_opt_CHOLxR5 ( int_uplo, a_obj );
              break;
     case 16 : PLA_opt_CHOLxR6 ( int_uplo, a_obj );
              break;
     case 17 : PLA_opt_CHOLxR7 ( int_uplo, a_obj );
              break;
  }

  MPI_Barrier( MPI_COMM_WORLD );

  PLA_Temp_comm1D( template, &comm1d );
  MPI_Comm_size( comm1d, &np );

  *time = MPI_Wtime() - *time;
						/* compute MFLOPS/node                           */
  *mflops = 1.0e-6 * n  * n * n / ( 3 * *time * np );
						/* check answer                                  */


  *info = PLA_test_compare( a_all, lda_all, a_obj);
						/* free space                                    */
  if ( a_all != NULL ) pla_free( a_all );
  PLA_Obj_free( &a_obj );
}

void PLA_no_test_opt_dchol(	int int_alg,
				PLA_Template template, char *uplo,
			 	int n,
				int ia, int ja,
				double *time, double *mflops, int *info )

{
  PLA_Obj a_obj=NULL;
  MPI_Comm comm1d;
  int      np, int_uplo;

  PLA_Matrix_create ( MPI_DOUBLE, n, n, template, ia, ja, &a_obj );
  PLA_Obj_fill ( "D", 1.0e+0, & a_obj );

  int_uplo =  (*uplo == 'L') ? PLA_LOWER_TRIANGULAR : PLA_UPPER_TRIANGULAR;

  MPI_Barrier( MPI_COMM_WORLD );
  *time = MPI_Wtime();

  switch ( int_alg )
  {
     case 1 : PLA_opt_CHOLxL1 ( int_uplo, a_obj );
              break;
     case 2 : PLA_opt_CHOLxL2 ( int_uplo, a_obj );
              break;
     case 3 : PLA_opt_CHOLxL3 ( int_uplo, a_obj );
              break;
     case 4 : PLA_opt_CHOLxL4 ( int_uplo, a_obj );
              break;
     case 5 : PLA_opt_CHOLxL5 ( int_uplo, a_obj );
              break;
     case 6 : PLA_opt_CHOLxL6 ( int_uplo, a_obj );
              break;
     case 11 : PLA_opt_CHOLxR1 ( int_uplo, a_obj );
              break;
     case 12 : PLA_opt_CHOLxR2 ( int_uplo, a_obj );
              break;
     case 13 : PLA_opt_CHOLxR3 ( int_uplo, a_obj );
              break;
     case 14 : PLA_opt_CHOLxR4 ( int_uplo, a_obj );
              break;
     case 15 : PLA_opt_CHOLxR5 ( int_uplo, a_obj );
              break;
     case 16 : PLA_opt_CHOLxR6 ( int_uplo, a_obj );
              break;
     case 17 : PLA_opt_CHOLxR7 ( int_uplo, a_obj );
              break;
  }

  MPI_Barrier( MPI_COMM_WORLD );

  PLA_Temp_comm1D( template, &comm1d );
  MPI_Comm_size( comm1d, &np );

  *time = MPI_Wtime() - *time;
  *mflops = 1.0e-6 * n  * n * n / ( 3.0e+0 * *time * np );
  PLA_Obj_free( &a_obj );
}
