#include "PLA.h"

/*
 * PLAPACK Level-2 BLAS Implementation
 */
int PLA_opt_CHOLxL1 ( int int_uplo, PLA_Obj a )
{
   int size;
   PLA_Template template = NULL;
   PLA_Obj adone = NULL, anext = NULL,
           a21 = NULL, a22 = NULL, a23 = NULL,
           a31 = NULL, a32 = NULL,
           min_one = NULL, one = NULL;


   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & one );

   PLA_Obj_set_to_one ( one );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Obj_vert_split_2 ( a, 0, & adone, & anext );

   while ( TRUE )
   {
      PLA_Obj_global_length ( anext, & size );

      if ( size == 0 ) break;


      PLA_Obj_split_4 ( anext, 1, 1, & a22, & a23,
                                     & a32, & anext );
      PLA_Obj_horz_split_2 ( adone, 1, & a21,
                                       & a31 );

      PLA_Obj_set_orientation ( a21, PLA_PROJ_ONTO_ROW );
      PLA_Obj_objtype_cast ( a22, PLA_MSCALAR );

      PLA_Dot_x ( min_one, a21, a21, one, a22 );
      PLA_Gemv ( PLA_NO_TRANS, min_one, a31, a21, one, a32 );

      PLA_Take_sqrt ( a22 );

      PLA_Obj_objtype_cast ( a22, PLA_MSCALAR );
      PLA_Inv_scal ( a22, a32 );

      PLA_Obj_view_shift ( adone,    1,
                                  0,    1,
                                     0 );
   }

   PLA_Obj_free ( & adone );
   PLA_Obj_free ( & anext );
   PLA_Obj_free ( & a21 );
   PLA_Obj_free ( & a22 );
   PLA_Obj_free ( & a23 );
   PLA_Obj_free ( & a31 );
   PLA_Obj_free ( & a32 );
   PLA_Obj_free ( & min_one );
   PLA_Obj_free ( & one );

   return ( PLA_SUCCESS );
}


/*
 * PLAPACK Level-3 PBLAS Implementation
 */
int PLA_opt_CHOLxL2 ( int int_uplo, PLA_Obj a )
{
   int nb_alg, size;
   PLA_Template template = NULL;
   PLA_Obj adone = NULL, anext = NULL,
           a21 = NULL, a22 = NULL, a23 = NULL,
           a31 = NULL, a32 = NULL,
           min_one = NULL, one = NULL;


   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & one );

   PLA_Obj_set_to_one ( one );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Environ_nb_alg ( PLA_OP_MAT_PAN, template, & nb_alg );

   PLA_Obj_vert_split_2 ( a, 0, & adone, & anext );

   while ( TRUE )
   {
      PLA_Obj_global_length ( anext, & size );

      if ( ( size = min ( size, nb_alg ) ) == 0 ) break;

      PLA_Obj_split_4 ( anext, size, size, & a22, & a23,
                                           & a32, & anext );
      PLA_Obj_horz_split_2 ( adone, size, & a21,
                                          & a31 );

      PLA_Syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS, min_one, a21, one, a22 );

      PLA_Gemm ( PLA_NO_TRANS, PLA_TRANS, min_one, a31, a21, one, a32 );

      PLA_opt_CHOLxL1 ( PLA_LOW_TRIAN, a22 );

      PLA_Trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                 one, a22, a32 );

      PLA_Obj_view_shift ( adone,    size,
                                  0,       size,
                                        0 );
   }

   PLA_Obj_free ( & adone );
   PLA_Obj_free ( & anext );
   PLA_Obj_free ( & a21 );
   PLA_Obj_free ( & a22 );
   PLA_Obj_free ( & a23 );
   PLA_Obj_free ( & a31 );
   PLA_Obj_free ( & a32 );
   PLA_Obj_free ( & min_one );
   PLA_Obj_free ( & one );

   return ( PLA_SUCCESS );
}


/*
 * PLAPACK Level-3 BLAS Implementation (calls single processor cholesky)
 */
int PLA_opt_CHOLxL3 ( int int_uplo, PLA_Obj a )
{
   int size, size_top, owner_top, size_left, owner_left;
   PLA_Template template = NULL;
   PLA_Obj adone = NULL, anext = NULL,
           a21 = NULL, a22 = NULL, a23 = NULL,
           a31 = NULL, a32 = NULL,
           min_one = NULL, one = NULL;


   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & one );

   PLA_Obj_set_to_one ( one );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Obj_vert_split_2 ( a, 0, & adone, & anext );

   while ( TRUE )
   {
      PLA_Obj_split_size ( anext, PLA_SIDE_TOP, & size_top, & owner_top );

      PLA_Obj_split_size ( anext, PLA_SIDE_LEFT, & size_left, & owner_left );

      if ( ( size = min ( size_top, size_left ) ) == 0 ) break;

      PLA_Obj_split_4 ( anext, size, size, & a22, & a23,
                                           & a32, & anext );
      PLA_Obj_horz_split_2 ( adone, size, & a21,
                                          & a31 );

      PLA_Syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS, min_one, a21, one, a22 );

      PLA_Gemm ( PLA_NO_TRANS, PLA_TRANS, min_one, a31, a21, one, a32 );

      PLA_Local_chol ( a22 );

      PLA_Trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                 one, a22, a32 );

      PLA_Obj_view_shift ( adone,    size,
                                  0,       size,
                                        0 );
   }

   PLA_Obj_free ( & adone );
   PLA_Obj_free ( & anext );
   PLA_Obj_free ( & a21 );
   PLA_Obj_free ( & a22 );
   PLA_Obj_free ( & a23 );
   PLA_Obj_free ( & a31 );
   PLA_Obj_free ( & a32 );
   PLA_Obj_free ( & min_one );
   PLA_Obj_free ( & one );

   return ( PLA_SUCCESS );
}


/*
 * PLAPACK Level-3 BLAS Implementation (calls Local_syrk)
 */
int PLA_opt_CHOLxL4 ( int int_uplo, PLA_Obj a )
{
   int size, size_top, owner_top, size_left, owner_left;
   PLA_Template template = NULL;
   PLA_Obj adone = NULL, anext = NULL,
           a21 = NULL, a22 = NULL, a23 = NULL,
           a31 = NULL, a32 = NULL,
           min_one = NULL, zero = NULL, one = NULL,
           a22_dup = NULL;

   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & one );

   PLA_Obj_set_to_one ( one );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & zero );

   PLA_Obj_set_to_zero ( zero );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Obj_vert_split_2 ( a, 0, & adone, & anext );

   while ( TRUE )
   {
      PLA_Obj_split_size ( anext, PLA_SIDE_TOP, & size_top, & owner_top );

      PLA_Obj_split_size ( anext, PLA_SIDE_LEFT, & size_left, & owner_left );

      if ( ( size = min ( size_top, size_left ) ) == 0 ) break;

      PLA_Obj_split_4 ( anext, size, size, & a22, & a23,
                                           & a32, & anext );
      PLA_Obj_horz_split_2 ( adone, size, & a21,
                                          & a31 );

      /* SYRK Optimization */

/*PLA_Mscalar_create_conf_to ( a22, PLA_INHERIT, PLA_ALL_COLS, & a22_dup );*/
      PLA_Mscalar_create_conf_to ( a22, PLA_ALL_ROWS, PLA_ALL_COLS, & a22_dup );

      PLA_Local_syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS,
                       min_one, a21, zero, a22_dup );

      PLA_Reduce_x ( PLA_SHAPE_LOW_TRAP, a22_dup, one, a22 );

      PLA_Gemm ( PLA_NO_TRANS, PLA_TRANS, min_one, a31, a21, one, a32 );

      PLA_Local_chol ( a22 );

      PLA_Trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                 one, a22, a32 );

      PLA_Obj_view_shift ( adone,    size,
                                  0,       size,
                                        0 );
   }

   PLA_Obj_free ( & adone );
   PLA_Obj_free ( & anext );
   PLA_Obj_free ( & a21 );
   PLA_Obj_free ( & a22 );
   PLA_Obj_free ( & a23 );
   PLA_Obj_free ( & a31 );
   PLA_Obj_free ( & a32 );
   PLA_Obj_free ( & a22_dup );
   PLA_Obj_free ( & min_one );
   PLA_Obj_free ( & zero );
   PLA_Obj_free ( & one );

   return ( PLA_SUCCESS );
}


/*
 * PLAPACK Level-3 BLAS Implementation (calls Local_syrk and Local_gemm)
 */
int PLA_opt_CHOLxL5 ( int int_uplo, PLA_Obj a )
{
   int size, size_top, owner_top, size_left, owner_left,
       length;
   PLA_Template template = NULL;
   PLA_Obj adone = NULL, anext = NULL,
           a21     = NULL, a22     = NULL, a23 = NULL,
           a31     = NULL, a32     = NULL,
           a21_dup = NULL, a22_dup = NULL,
                           a32_dup = NULL,
           min_one = NULL, zero = NULL, one = NULL;

   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & one );

   PLA_Obj_set_to_one ( one );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & zero );

   PLA_Obj_set_to_zero ( zero );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Obj_vert_split_2 ( a, 0, & adone, & anext );

   while ( TRUE )
   {
      PLA_Obj_split_size ( anext, PLA_SIDE_TOP, & size_top, & owner_top );

      PLA_Obj_split_size ( anext, PLA_SIDE_LEFT, & size_left, & owner_left );

      if ( ( size = min ( size_top, size_left ) ) == 0 ) break;

      PLA_Obj_split_4 ( anext, size, size, & a22, & a23,
                                           & a32, & anext );
      PLA_Obj_horz_split_2 ( adone, size, & a21,
                                          & a31 );

      /* SYRK Optimization */

/*      PLA_Mscalar_create_conf_to ( a22, PLA_INHERIT, PLA_ALL_COLS, & a22_dup );*/
      PLA_Mscalar_create_conf_to ( a22, PLA_ALL_ROWS, PLA_ALL_COLS, & a22_dup );

      PLA_Local_syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS,
                       min_one, a21, zero, a22_dup );

      PLA_Reduce_x ( PLA_SHAPE_LOW_TRAP, a22_dup, one, a22 );

      /* GEMM Optimization */

      PLA_Obj_global_length ( a22, & length );
      PLA_Pmvector_create_conf_to ( a31, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                    length, & a21_dup );
      PLA_Obj_set_orientation ( a21, PLA_PROJ_ONTO_ROW );
      PLA_Copy ( a21, a21_dup );

      PLA_Pmvector_create_conf_to ( a31, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                    length, & a32_dup );
      PLA_Obj_set_orientation ( a32, PLA_PROJ_ONTO_COL );

      PLA_Local_gemm ( PLA_NO_TRANS, PLA_TRANS,
                       min_one, a31, a21_dup, zero, a32_dup );

      PLA_Reduce_x ( PLA_SHAPE_GENERAL, a32_dup, one, a32 );

      PLA_Local_chol ( a22 );

      PLA_Trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                 one, a22, a32 );

      PLA_Obj_view_shift ( adone,    size,
                                  0,       size,
                                        0 );
   }

   PLA_Obj_free ( & adone );
   PLA_Obj_free ( & anext );
   PLA_Obj_free ( & a21 );      PLA_Obj_free ( & a21_dup );
   PLA_Obj_free ( & a22 );      PLA_Obj_free ( & a22_dup );
   PLA_Obj_free ( & a23 );
   PLA_Obj_free ( & a31 );
   PLA_Obj_free ( & a32 );      PLA_Obj_free ( & a32_dup );
   PLA_Obj_free ( & min_one );
   PLA_Obj_free ( & zero );
   PLA_Obj_free ( & one );

   return ( PLA_SUCCESS );
}

/*
 * PLAPACK Level-3 BLAS Implementation (calls Local_syrk, Local_gemm, Local_trsm)
 */
int PLA_opt_CHOLxL6 ( int int_uplo, PLA_Obj a )
{
   int size, size_top, owner_top, size_left, owner_left,
       length;
   PLA_Template template = NULL;
   PLA_Obj adone = NULL, anext = NULL,
           a21     = NULL, a22     = NULL, a23 = NULL,
           a31     = NULL, a32     = NULL,
           a21_dup = NULL, a22_dup = NULL,
                           a32_dup = NULL,
           min_one = NULL, zero = NULL, one = NULL;

   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & one );

   PLA_Obj_set_to_one ( one );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & zero );

   PLA_Obj_set_to_zero ( zero );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Obj_vert_split_2 ( a, 0, & adone, & anext );

   while ( TRUE )
   {
      PLA_Obj_split_size ( anext, PLA_SIDE_TOP, & size_top, & owner_top );

      PLA_Obj_split_size ( anext, PLA_SIDE_LEFT, & size_left, & owner_left );

      if ( ( size = min ( size_top, size_left ) ) == 0 ) break;

      PLA_Obj_split_4 ( anext, size, size, & a22, & a23,
                                           & a32, & anext );
      PLA_Obj_horz_split_2 ( adone, size, & a21,
                                          & a31 );

/*      PLA_Mscalar_create_conf_to ( a22, PLA_INHERIT, PLA_ALL_COLS, & a22_dup );*/
      PLA_Mscalar_create_conf_to ( a22, PLA_ALL_ROWS, PLA_ALL_COLS, & a22_dup );

      PLA_Local_syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS,
                       min_one, a21, zero, a22_dup );

      PLA_Reduce_x ( PLA_SHAPE_LOW_TRAP, a22_dup, one, a22 );

      PLA_Obj_global_length ( a22, & length );
      PLA_Pmvector_create_conf_to ( a31, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                    length, & a21_dup );
      PLA_Obj_set_orientation ( a21, PLA_PROJ_ONTO_ROW );
      PLA_Copy ( a21, a21_dup );

      PLA_Pmvector_create_conf_to ( a31, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                    length, & a32_dup );
      PLA_Obj_set_orientation ( a32, PLA_PROJ_ONTO_COL );

      PLA_Local_gemm ( PLA_NO_TRANS, PLA_TRANS,
                       min_one, a31, a21_dup, zero, a32_dup );

      PLA_Reduce_x ( PLA_SHAPE_GENERAL, a32_dup, one, a32 );

      PLA_Local_chol ( a22 );

      /* TRSM Optimization */

/*      PLA_Mscalar_create_conf_to ( a22, PLA_ALL_ROWS, PLA_INHERIT, & a22_dup );*/

      PLA_Copy ( a22, a22_dup );

      PLA_Local_trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                       one, a22_dup, a32 );

      PLA_Obj_view_shift ( adone,    size,
                                  0,       size,
                                        0 );
   }

   PLA_Obj_free ( & adone );
   PLA_Obj_free ( & anext );
   PLA_Obj_free ( & a21 );      PLA_Obj_free ( & a21_dup );
   PLA_Obj_free ( & a22 );      PLA_Obj_free ( & a22_dup );
   PLA_Obj_free ( & a23 );
   PLA_Obj_free ( & a31 );
   PLA_Obj_free ( & a32 );      PLA_Obj_free ( & a32_dup );
   PLA_Obj_free ( & min_one );
   PLA_Obj_free ( & zero );
   PLA_Obj_free ( & one );

   return ( PLA_SUCCESS );
}

/*
 * PLAPACK Level-3 BLAS Implementation (Reduce to multivector)
 */
int PLA_opt_CHOLxL7 ( int int_uplo, PLA_Obj a )
{
   int size, size_top, owner_top, size_left, owner_left,
       length;
   PLA_Template template = NULL;
   PLA_Obj adone = NULL, anext = NULL,
           a21     = NULL, a22     = NULL, a23 = NULL,
           a31     = NULL, a32     = NULL,
           a21_dup = NULL, a22_dup = NULL,
                           a32_dup = NULL,
           a21_mv  = NULL, a32_mv  = NULL,
           min_one = NULL, zero = NULL, one = NULL;

   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & one );

   PLA_Obj_set_to_one ( one );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & zero );

   PLA_Obj_set_to_zero ( zero );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Obj_vert_split_2 ( a, 0, & adone, & anext );

   while ( TRUE )
   {
      PLA_Obj_split_size ( anext, PLA_SIDE_TOP, & size_top, & owner_top );

      PLA_Obj_split_size ( anext, PLA_SIDE_LEFT, & size_left, & owner_left );

      if ( ( size = min ( size_top, size_left ) ) == 0 ) break;

      PLA_Obj_split_4 ( anext, size, size, & a22, & a23,
                                           & a32, & anext );
      PLA_Obj_horz_split_2 ( adone, size, & a21,
                                          & a31 );

      PLA_Obj_set_orientation ( a21, PLA_PROJ_ONTO_ROW );

      PLA_Mvector_create_conf_to ( a21, PLA_INHERIT, & a21_mv );
      PLA_Copy ( a21, a21_mv );

      PLA_Mvector_create_conf_to ( a32, PLA_INHERIT, & a32_mv );
      PLA_Copy ( a32, a32_mv );

      PLA_Mscalar_create_conf_to ( a22, PLA_ALL_ROWS, PLA_ALL_COLS, & a22_dup );

      PLA_Local_syrk ( PLA_LOW_TRIAN, PLA_TRANS,  /* Transpose required after copy */
                       min_one, a21_mv, zero, a22_dup );

      PLA_Reduce_x ( PLA_SHAPE_LOW_TRAP, a22_dup, one, a22 );

      PLA_Obj_global_length ( a22, & length );
      PLA_Pmvector_create_conf_to ( a31, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                    length, & a21_dup );
      /*PLA_Copy ( a21_mv, a21_dup ); ??? */
      PLA_Copy ( a21, a21_dup );

      PLA_Pmvector_create_conf_to ( a31, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                    length, & a32_dup );
      PLA_Obj_set_orientation ( a32, PLA_PROJ_ONTO_COL );

      PLA_Local_gemm ( PLA_NO_TRANS, PLA_TRANS,
                       min_one, a31, a21_dup, zero, a32_dup );

      PLA_Reduce_x ( PLA_SHAPE_GENERAL, a32_dup, one, a32_mv );

      PLA_Local_chol ( a22 );

      PLA_Copy ( a22, a22_dup );

      PLA_Local_trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                       one, a22_dup, a32_mv );

      PLA_Copy ( a32_mv, a32 );

      PLA_Obj_view_shift ( adone,    size,
                                  0,       size,
                                        0 );
   }

   PLA_Obj_free ( & adone );
   PLA_Obj_free ( & anext );
   PLA_Obj_free ( & a21 );      PLA_Obj_free ( & a21_dup );  PLA_Obj_free ( & a21_mv );
   PLA_Obj_free ( & a22 );      PLA_Obj_free ( & a22_dup );
   PLA_Obj_free ( & a23 );
   PLA_Obj_free ( & a31 );
   PLA_Obj_free ( & a32 );      PLA_Obj_free ( & a32_dup );   PLA_Obj_free ( & a32_mv );
   PLA_Obj_free ( & min_one );
   PLA_Obj_free ( & zero );
   PLA_Obj_free ( & one );

   return ( PLA_SUCCESS );
}

/*
 * PLAPACK Level-3 BLAS Implementation (nb_alg > nb_temp)
 */
int PLA_opt_CHOLxL8 ( int int_uplo, PLA_Obj a )
{
   int nb_alg, size, size_top, owner_top, size_left, owner_left,
       length;
   PLA_Template template = NULL;
   PLA_Obj adone = NULL, anext = NULL,
           a21     = NULL, a22     = NULL, a23 = NULL,
           a31     = NULL, a32     = NULL,
           a21_dup = NULL, a22_dup = NULL,
                           a32_dup = NULL,
           a21_mv  = NULL, a32_mv  = NULL,
           min_one = NULL, zero = NULL, one = NULL;

   PLA_Obj_template ( a, & template );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & one );

   PLA_Obj_set_to_one ( one );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & zero );

   PLA_Obj_set_to_zero ( zero );

   PLA_Mscalar_create ( MPI_DOUBLE,
                        PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
                        template, & min_one );

   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Environ_nb_alg ( PLA_OP_MAT_PAN, template, & nb_alg );

   PLA_Obj_vert_split_2 ( a, 0, & adone, & anext );

   while ( TRUE )
   {
      PLA_Obj_global_length ( anext, & size );

      if ( ( size = min ( size, nb_alg ) ) == 0 ) break;

      PLA_Obj_split_4 ( anext, size, size, & a22, & a23,
                                           & a32, & anext );
      PLA_Obj_horz_split_2 ( adone, size, & a21,
                                          & a31 );

      PLA_Obj_set_orientation ( a21, PLA_PROJ_ONTO_ROW );
      PLA_Obj_set_orientation ( a23, PLA_PROJ_ONTO_COL );

      PLA_Mvector_create_conf_to ( a21, PLA_INHERIT, & a21_mv );
      PLA_Copy ( a21, a21_mv );

      PLA_Mvector_create_conf_to ( a32, PLA_INHERIT, & a32_mv );
      PLA_Copy ( a32, a32_mv );

      PLA_Mscalar_create_conf_to ( a22, PLA_ALL_ROWS, PLA_ALL_COLS, & a22_dup );

      PLA_Local_syrk ( PLA_LOW_TRIAN, PLA_TRANS,  /* Transpose required after copy */
                       min_one, a21_mv, zero, a22_dup );
printf ( "reduce x\n" );
      PLA_Reduce_x ( PLA_SHAPE_LOW_TRAP, a22_dup, one, a22 );

printf ( "create pmvector\n" );
      PLA_Pmvector_create_conf_to ( a31, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS,
                                    size, & a21_dup );
      /*PLA_Copy ( a21_mv, a21_dup ); ??? */
      PLA_Copy ( a21, a21_dup );

      PLA_Pmvector_create_conf_to ( a31, PLA_PROJ_ONTO_COL, PLA_ALL_COLS,
                                    size, & a32_dup );

      PLA_Local_gemm ( PLA_NO_TRANS, PLA_TRANS,
                       min_one, a31, a21_dup, zero, a32_dup );

      PLA_Reduce_x ( PLA_SHAPE_GENERAL, a32_dup, one, a32_mv );

printf ( "cholx6\n" );
      PLA_opt_CHOLxL6 ( int_uplo, a22 );

printf ( "copy mscalar\n" );
      PLA_Copy ( a22, a22_dup );

      PLA_Local_trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                       one, a22_dup, a32_mv );

      PLA_Copy ( a32_mv, a32 );

      PLA_Obj_view_shift ( adone,    size,
                                  0,       size,
                                        0 );
   }

   PLA_Obj_free ( & adone );
   PLA_Obj_free ( & anext );
   PLA_Obj_free ( & a21 );      PLA_Obj_free ( & a21_dup );  PLA_Obj_free ( & a21_mv );
   PLA_Obj_free ( & a22 );      PLA_Obj_free ( & a22_dup );
   PLA_Obj_free ( & a23 );
   PLA_Obj_free ( & a31 );
   PLA_Obj_free ( & a32 );      PLA_Obj_free ( & a32_dup );   PLA_Obj_free ( & a32_mv );
   PLA_Obj_free ( & min_one );
   PLA_Obj_free ( & zero );
   PLA_Obj_free ( & one );

   return ( PLA_SUCCESS );
}

