
#include "PLA.h"

void PLA_test_dtrmm( PLA_Template template, char *side, char *uplo, char *transa, char *diag,
          int m, int n, 
          double alpha, 
          int ia, int ja, int ib, int jb,
          double *time, double *mflops, int *info )

{
  int mym, myn,         /* this node's row and column dimension    */
    lda_all, ldb_all,   /* leading dimensions of global matrices   */
    ldc_all,
    llda, lldb, lldc,   /* leading dimensions of local matrices    */
    idummy,             /* dummy variable                          */
    ii;                 /* loop counter -- dummy                   */
  double
    *a_all = NULL, *b_all = NULL, *c_all = NULL, /* arrays holding the global matrices      */
    *a, *b, *c;         /* arrays holding the local matrices       */

  PLA_Obj a_obj=NULL, b_obj=NULL, alpha_obj = NULL; 
  MPI_Comm comm1d;
  int      np;
  int      size1, size2, testrow, testcol;
  int    extra;
  double *extra_p, d_one;
  int side_int, uplo_int, transa_int, diag_int;

   /* Create distributed matrix objects for A, B */
  if ( *transa == 'N' ) {
         if(*side == 'L')
         {
             a_all = (double *) pla_malloc( m * m * sizeof( double ) );
        lda_all = m;
        PLA_Matrix_create ( MPI_DOUBLE, m, m, template, ia, ja, &a_obj );
        PLA_test_dgenmatrix( "S", 10.0, m, m, a_all, lda_all );
        PLA_Test_gdist_to_local( a_all, lda_all, a_obj );
         }
         else /* (*side == 'R') */
         {
             a_all = (double *) pla_malloc( n * n * sizeof( double ) );
        lda_all = n;
        PLA_Matrix_create ( MPI_DOUBLE, n, n, template, ia, ja, &a_obj );
        PLA_test_dgenmatrix( "S", 10.0, n, n, a_all, lda_all );
        PLA_Test_gdist_to_local( a_all, lda_all, a_obj );
         }
  }
  else if ( *transa == 'T' ) {
         if(*side == 'R')
         {
             a_all = (double *) pla_malloc( n * n * sizeof( double ) );
        lda_all = n;
        PLA_Matrix_create ( MPI_DOUBLE, n, n, template, ia, ja, &a_obj );
        PLA_test_dgenmatrix( "S", 10.0, n, n, a_all, lda_all );
        PLA_Test_gdist_to_local( a_all, lda_all, a_obj );
         }
         else /* (*side == 'L') */
         {
             a_all = (double *) pla_malloc( m * m * sizeof( double ) );
        lda_all = m;
        PLA_Matrix_create ( MPI_DOUBLE, m, m, template, ia, ja, &a_obj );
        PLA_test_dgenmatrix( "S", 10.0, m, m, a_all, lda_all );
        PLA_Test_gdist_to_local( a_all, lda_all, a_obj );
         }
  }
  b_all = (double *) pla_malloc( m * n * sizeof( double ) );
    PLA_Matrix_create ( MPI_DOUBLE, m, n, template, ib, jb, &b_obj );
    ldb_all = m;
/* The next two lines must put the same numbers in b_all and c_all */
         PLA_test_dgenmatrix( "S", 1000.0000, m, n, b_all, ldb_all );
    PLA_Test_gdist_to_local( b_all, ldb_all, b_obj );
      /* compute correct sequential answer             */
  PLA_dtrmm( side, uplo, transa, diag , &m, &n, &alpha, a_all, &lda_all,
       b_all, &ldb_all); 
   /* compute and time parallel answer              */
  MPI_Barrier( MPI_COMM_WORLD );
  *time = MPI_Wtime();
  PLA_Mscalar_create(MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template,&alpha_obj);
  PLA_Obj_set_to_one(alpha_obj);
  if((*side == 'L') || (*side == 'l'))
     side_int = PLA_SIDE_LEFT;
  else side_int = PLA_SIDE_RIGHT;
  if((*transa == 'N') || (*transa == 'n'))
     transa_int = PLA_NO_TRANSPOSE;
  else transa_int = PLA_TRANSPOSE;
  if((*uplo == 'U') || (*uplo == 'u'))
     uplo_int = PLA_UPPER_TRIANGULAR;
  else uplo_int = PLA_LOWER_TRIANGULAR;
  if((*diag == 'N') || (*diag == 'n'))
     diag_int = PLA_NONUNIT_DIAG;
  else diag_int = PLA_UNIT_DIAG;
  PLA_Trmm ( side_int, uplo_int, transa_int, diag_int, alpha_obj, a_obj, b_obj );
  MPI_Barrier( MPI_COMM_WORLD );
  PLA_Temp_comm1D( template, &comm1d );
  MPI_Comm_size( comm1d, &np );

  *time = MPI_Wtime() - *time;
                  /* compute MFLOPS/node                           */
  if( side_int == PLA_SIDE_LEFT )
     *mflops = 0.5 * m * n * m / *time * 1.0e-6 / np;
  else
     *mflops = 0.5 * n * n * m / *time * 1.0e-6 / np;
                  /* check answer                                  */
/* Note how b_all was changed -- first multiply -- then reverse solve */
/* to get the same answer                                             */
  *info = PLA_test_compare( b_all, ldb_all, b_obj);
                  /* free space                                    */
  if ( a_all != NULL ) pla_free( a_all );
  if ( b_all != NULL ) pla_free( b_all );
  if ( c_all != NULL ) pla_free( c_all );
  PLA_Obj_free( &a_obj );
  PLA_Obj_free( &b_obj);
  PLA_Obj_free( &alpha_obj);
}




