#include <mpi.h>
#include <PLA.h>
#include <MISC_DEFS.h>
#include "PRISM.h"

/* link list methods for the final list */

/* PRISM_subproblem is a node in the link list */

void PRISM_subproblem_create( PRISM_subproblem *prob ) {
   PRISM_subproblem temp;
   temp = (PRISM_subproblem) malloc(sizeof(struct prism_subproblem));
   if (temp == NULL) {
      printf("malloc error in PRISM_subproblem_create\n");
      exit(1);
   }
   temp->next = NULL;
   temp->id = 0;
   temp->node = 0;
   temp->A = NULL;
   temp->Z = NULL;
   temp->D = NULL;
   temp->sub = NULL;
   temp->diag = NULL;
   *prob = temp;
}

void PRISM_subproblem_delete( PRISM_subproblem *prob ) {
   PRISM_subproblem temp = *prob;
   PLA_Obj_free(&temp->A);
   PLA_Obj_free(&temp->Z);
   PLA_Obj_free(&temp->D);
   PLA_Obj_free(&temp->sub);
   PLA_Obj_free(&temp->diag);
   free(temp);
}


/* PRISM_final_list is a link list of subproblems */

void PRISM_final_list_create( PRISM_final_list *list ) {
   PRISM_final_list temp;
   temp = (PRISM_final_list) malloc(sizeof(struct prism_final_list));
   if (temp == NULL) {
      printf("malloc error in PRISM_final_list_create\n");
      exit(1);
   }
   temp->head = NULL;
   temp->tail = NULL;
   *list = temp;
}

void PRISM_final_list_delete( PRISM_final_list *list ) {
   PRISM_final_list temp = *list;
   PRISM_subproblem next;

   while ( temp->head ) {
      next = temp->head->next;
      PRISM_subproblem_delete( &temp->head );
      temp->head = next;
   }
   free(temp);
}

void PRISM_final_list_insert( PRISM_subproblem prob, PRISM_final_list list ) {
   if ( list->head ) {
      prob->id = list->tail->id + 1;
      list->tail->next = prob; 
      list->tail = prob;
      prob->next = NULL;
   } else {
      prob->id = 1;
      list->head = prob;
      list->tail = prob;
      prob->next = NULL;
   }
}

void PRISM_final_list_get_first( PRISM_final_list list, PRISM_subproblem *prob) {
   *prob = list->head;
}

void PRISM_final_list_get_next( PRISM_final_list list, PRISM_subproblem *prob) {
   *prob = (*prob)->next;
}



/* MISC utilites */

void PRISM_subproblem_print( PRISM_subproblem prob ) {
   if (me==ROOT) printf("subproblem %d\n",prob->id);
   PRISM_print( prob->A );
   PRISM_print( prob->Z ); 
}

void PRISM_final_list_print( PRISM_final_list list ) {
   PRISM_subproblem temp = list->head;
   while (temp) {
      PRISM_subproblem_print( temp );
      /* MPI_Barrier fixes an occassional crash - I'm not sure why */
      MPI_Barrier(MPI_COMM_WORLD); 
      temp = temp->next;
   }
}
