#include <mpi.h>
#include <PLA.h>
#include <MISC_DEFS.h>
#include "PRISM.h"

void PRISM_range(PLA_Obj X, PLA_Obj lambda_min, PLA_Obj lambda_max) {
/*
    Compute estimates for the range of eigenvalues of X such that 
    lambda_min <= lambda( X ) <= lambda_max.  
    
    Method: Gerschgorin's theorem.
*/
    int align, length;
    MPI_Datatype datatype;
    PLA_Template template;
    PLA_Obj norm=NULL, diag=NULL, work=NULL;

    /* Extract info from matrix X */
    PLA_Obj_global_align	( X, &align );
    PLA_Obj_global_length	( X, &length );
    PLA_Obj_datatype		( X, &datatype );
    PLA_Obj_template		( X, &template );

    /* Generate workspace multivectors conformal to columns of X */
    PLA_Mvector_create( datatype, length, 1, template, align, &norm );
    PLA_Mvector_create( datatype, length, 1, template, align, &diag );
    PLA_Mvector_create( datatype, length, 1, template, align, &work );

    /* Fill multivector norm with column asum values */
    PRISM_asum( X, norm );

    /* Fill multivector diag with diagonal elements */
    PRISM_get_diag( X, diag );

    /* Remove diagonal contribitions from norm */
    PRISM_abs( diag, work );
    PLA_Local_axpy( PRISM_mone, work, norm );

    /* Compute lambda_min */
    PLA_Local_copy( diag, work );
    PLA_Local_axpy( PRISM_mone, norm, work );
    PRISM_min( work, lambda_min );

    /* Compute lambda_max */
    PLA_Local_copy( diag, work );
    PLA_Local_axpy( PRISM_one, norm, work );
    PRISM_max( work, lambda_max );

    /* free objects */
    PLA_Obj_free( &diag );
    PLA_Obj_free( &work );
    PLA_Obj_free( &norm );

#ifdef DB_GLOBAL_
    if (me==ROOT) {
       printf(">> PRISM_range\n");
       printf("lambda_min="); prism_pla_prt_m(lambda_min);
       printf("lambda_max="); prism_pla_prt_m(lambda_max);
    }
#endif
}

